#include "elasticityTensor.hh"
#include "basics/debug.hh"

template <int dim>
concepts::Mapping<Real, dim> ElasticityTensor<dim>::operator()
  (uint i, uint j) const {
  conceptsAssert(i<dim, concepts::Assertion());
  conceptsAssert(j<dim, concepts::Assertion());
  Real temp[dim*dim];
  for (uint r=0; r<dim; ++r) {
    for (uint s=0; s<dim; ++s) {
      temp[dim*r+s] = this->entry(i, r, j, s);
    }
  }
  concepts::Mapping<Real, dim> res = rp2map_(temp);
  return res;
}

template <int dim>
concepts::Mapping<Real, dim> ElasticityTensor<dim>::operator()
  (concepts::Mapping<Real, dim>& strain) const {
  Real temp[dim*dim];
  for (uint r=0; r<dim; ++r) {
    for (uint s=0; s<dim; ++s) {
      temp[dim*r+s] = 0;
      for (uint i=0; i<dim; ++i) {
        for (uint j=0; j<dim; ++j) {
          temp[dim*r+s] += strain(i,j)*this->entry(i, j, r, s);
        }
      }
    }
  }
  concepts::Mapping<Real, dim> res = rp2map_(temp);
  return res;
}

template <>
concepts::Mapping<Real, 2>
ElasticityTensor<2>::rp2map_(Real* temp) const {
  return concepts::Mapping<Real, 2>(temp[0], temp[1],
                                    temp[2], temp[3]);
}

template <>
concepts::Mapping<Real, 3>
ElasticityTensor<3>::rp2map_(Real* temp) const {
  return concepts::Mapping<Real, 3>(temp[0], temp[1], temp[2],
                                    temp[3], temp[4], temp[5],
                                    temp[6], temp[7], temp[8]);
}

template <int dim>
ElasticityTensorIsotropic<dim>::ElasticityTensorIsotropic(Real mu,
                                                          Real lambda)
  : mu_(mu), lambda_(lambda) { }

template <int dim>
Real ElasticityTensorIsotropic<dim>::entry(uint i, uint j,
                                           uint r, uint s) const {
  conceptsAssert(i<dim, concepts::Assertion());
  conceptsAssert(j<dim, concepts::Assertion());
  conceptsAssert(r<dim, concepts::Assertion());
  conceptsAssert(s<dim, concepts::Assertion());
  Real res = 0;
  res += (i==r)&&(j==s) ? mu_ : 0;
  res += (i==s)&&(j==r) ? mu_ : 0;
  res += (i==j)&&(r==s) ? lambda_ : 0;
  return res;
}

template class ElasticityTensor<2>;
template class ElasticityTensorIsotropic<2>;
template class ElasticityTensor<3>;
template class ElasticityTensorIsotropic<3>;
