#include <cmath>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <unistd.h>
#include "basics.hh"
#include "toolbox.hh"

int main(int argc, char** argv) {

  uint l = 0;
  uint p = 1;
  bool debug = false;

  try {
    // ********************************************************** input data **
    concepts::InputParser input(true);
    concepts::InOutParameters& inputParameters = input.inputParameters();
    inputParameters.addInt("level", l);
    inputParameters.addInt("polynomial", p);
    inputParameters.addBool("debug", debug);
    inputParameters.addString("parameterout", "inputoutput.out");

    concepts::InOutParameters& outputParameters = input.outputParameters();

    outputParameters.addArrayDouble("error");
    concepts::ResultsTable table;
    table.addMap(concepts::ResultsTable::DOUBLE, "error", outputParameters);

    std::string inputfile;
    int opt;
    while ((opt = getopt(argc, argv, "-f:l:p:d")) != EOF)
      switch(opt) {
      case 'l': inputParameters.addInt("level", std::atoi(optarg)); break;
      case 'p': inputParameters.addInt("polynomial", std::atoi(optarg)); 
	break;
      case 'f': inputfile = std::string(optarg); 
	input.parse(inputfile);
	break;
      case 'd': inputParameters.addBool("debug", true); break;
      default:
	std::cout << "Usage: " << argv[0] 
		  << " [-l LEVEL] [-p DEGREE] [-f FILE] [-d]" 
		  << std::endl
		  << "where" << std::endl
		  << "  FILE: name of the input file" << std::endl
		  << "  LEVEL: level of refinement" << std::endl
		  << "  DEGREE: polynomial degree" << std::endl
		  << "  -d: print the matrices" << std::endl
		  << "Options given after the input file override the values "
		  << "read from the"
		  << std::endl << "input file." << std::endl;
	std::exit(1);
	break;
      }

    // ***************************************************** show parameters **
    std::cout << '[' << argv[0] << "]" << std::endl;
    std::cout << "--" << std::endl;
    std::cout << "Parameters:" << std::endl
	 << "  input file = " << inputfile << std::endl << inputParameters;
    std::cout << "--" << std::endl;

    l = inputParameters.getInt("level");
    p = inputParameters.getInt("polynomial");

    // ******************************************************** computations **
    outputParameters.addInt("nelm", 10);
    for (int i = 0; i < 10; ++i)
      outputParameters.addArrayDouble("error", i, 1.0/(1<<i));

    // ************************************** output of input data and other **
    std::cout << "--" << std::endl
	 << "Writing gathered data to disk: " 
	 << inputParameters.getString("parameterout") << std::endl;

    std::ofstream *ofs = new std::ofstream
      (inputParameters.getString("parameterout").c_str());

    *ofs << "/* program:\t" << argv[0] << std::endl
	 << " * command:\t";
    for (int i = 0; i < argc; ++i)
      *ofs << argv[i] << " ";
    *ofs << std::endl
	 << " * input file:\t" << inputfile << std::endl;
    *ofs << " */" << std::endl << input;
    delete ofs;

    std::cout << table << std::endl;
    ofs = new std::ofstream("inputoutput.gnuplot");
    *ofs << std::setprecision(20);
    table.print<concepts::ResultsTable::GNUPLOT>(*ofs);
    delete ofs;
  }
  catch (concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
