/* exceptions
 */

#include <cstdio>
#include <unistd.h>
#include <cstdlib>

#include "exceptions.hh"
#include "debug.hh"

// debugging
#define ExceptionBaseConstr_D 0
#define ExceptionBaseDestr_D 0
#define stacktrace_D 0

namespace concepts {

  // ********************************************************* ExceptionBase **

  ExceptionBase::ExceptionBase() throw() :
    file_(std::string("")), line_(0), function_(std::string("")), 
    excName_(std::string("")) {
    DEBUGL(ExceptionBaseConstr_D, "done.");
  }

  ExceptionBase::ExceptionBase(const std::string& file,
			       const unsigned int line, 
			       const std::string& function,
			       const std::string& excName) 
    throw() : file_(file), line_(line), function_(function), 
    excName_(excName) {
    DEBUGL(ExceptionBaseConstr_D, "done.");
  }

  ExceptionBase::~ExceptionBase() throw() {
    DEBUGL(ExceptionBaseDestr_D, "done.");
  }

  void ExceptionBase::setFields(const std::string& file,
				const unsigned int line, 
				const std::string& function,
				const std::string& excName) throw() {

    file_ = file;
    line_ = line;
    function_ = function;
    excName_ = excName;
  }

  std::ostream& ExceptionBase::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- error " << excName_;
  }

  // ******************************************************** MissingFeature **

  MissingFeature::MissingFeature(const std::string& feature) throw() : 
    ExceptionBase("", 0, "", ""), feature_(feature) {
  }

  MissingFeature::MissingFeature(const std::string& file,
				 const unsigned int line, 
				 const std::string& function,
				 const std::string& excName,
				 const std::string& feature) throw() : 
    ExceptionBase(file, line, function, excName),
    feature_(feature) {
  }

  MissingFeature::~MissingFeature() throw() {}

  std::ostream& MissingFeature::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- missing feature: " << feature_;
  }

  // ************************************************************* Assertion **

  Assertion::Assertion() throw(): 
    MissingFeature("", 0, "", "", "") {
  }

  Assertion::Assertion(const std::string& file, const unsigned int line, 
		       const std::string& function,
		       const std::string& excName, 
		       const std::string& cond) throw() : 
    MissingFeature(file, line, function, excName, cond) {
  }

  Assertion::~Assertion() throw() {}

  std::ostream& Assertion::info(std::ostream& os) const throw() {
    return os << '[' << file_ << ", line " << line_ << "] " << function_ 
	      << " -- error " << excName_ << " " << feature_;
  }

  void Assertion::setFieldsAssert(const std::string& file,
				  const unsigned int line, 
				  const std::string& function, 
				  const std::string& excName, 
				  const std::string& cond) throw() {

    ExceptionBase::setFields(file, line, function, excName);
    feature_ = cond;
  }

  // ************************************************************ Stacktrace **

  bool Stacktrace::doit_ = true;

  void Stacktrace::print() {
#if defined (__GNUC__) && ! defined (__arch64__)
    if (doit_) {
      const uint pid = getpid();
      DEBUGL(stacktrace_D, "pid = " << pid);
      char cmd[256], cmdName[256], cmds[256];
      // get name of this application from ps
      tmpnam(cmdName);
      DEBUGL(stacktrace_D, "cmdName = " << cmdName);
      snprintf(cmd, sizeof(cmd),
	       "ps -A -o pid -o comm | grep '^[ ]*%d' | sed 's/^.* //' > %s",
	       pid, cmdName);
      DEBUGL(stacktrace_D, "cmd = " << cmd);
      system(cmd);
      tmpnam(cmds);
      snprintf(cmd, sizeof(cmd), "echo 'bt\nquit\ny' > %s", cmds);
      DEBUGL(stacktrace_D, "cmd = " << cmd);
      system(cmd);
      DEBUGL(stacktrace_D, "cmds = " << cmds);
      DEBUGL(stacktrace_D, "cmdName = " << cmdName);
      snprintf(cmd, sizeof(cmd), "gdb -x %s `cat %s` %d", cmds, cmdName, pid);
      DEBUGL(stacktrace_D, "cmd = " << cmd);
      std::cout << "--- [35mStack trace follows[m ---" << std::endl
		<< "[35m" << std::flush;
      system(cmd);
      std::cout << "[m--- [35mStack trace ends[m ---" << std::endl;
      DEBUGL(stacktrace_D, "unlinking " << cmds << " and " << cmdName);
#  if stacktrace_D == 0
      unlink(cmds);
      unlink(cmdName);
#  endif
    }
#endif
  }

} // namespace
