/* level handling
 */

#ifndef level_hh
#define level_hh

#include <iostream>
#include "typedefs.hh"

namespace concepts {

  // ***************************************************************** Level **

  /** Level information used for multidimensionl hp FEM.
      @param dim Dimension of the method using this level
      and the size of the array.
      @author Philipp Frauenfelder, 2002
  */
  template<uint dim>
  struct Level {
    /// Level
    short l_[dim];

    /// Constructor
    Level(const Level<dim>& l) {
      for (uint i = 0; i < dim; ++i) l_[i] = l.l_[i]; }
    /// Constructor
    Level(short l) { for (uint i = 0; i < dim; ++i) l_[i] = l; }
    /// Default constructor
    Level() { for (uint i = 0; i < dim; ++i) l_[i] = 0; }

    /// Assignement operator
    Level<dim>& operator=(const Level<dim>& c);

    /** Comparison operator <.
	@return true if at least one component of \c l_ fulfils <
    */
    bool operator<(const Level<dim>& c) const;
    /** Comparison operator >.
	@return true if at least one component of \c l_ fulfils >
    */
    bool operator>(const Level<dim>& c) const;
    /** Comparison operator ==.
	@return true if all components of \c l_ are equal
    */
    bool operator==(const Level<dim>& c) const;

    /// Addition operator
    Level<dim> operator+(const Level<dim>& c) const;

    /// Subtraction operator
    Level<dim> operator-(const Level<dim>& c) const;

    /// Componentwise minimization
    void minimize(const Level<dim>& c);
    /// Componentwise maximazation
    void maximize(const Level<dim>& c);
  };

  template<uint dim>
  std::ostream& operator<<(std::ostream& os, const Level<dim>& c);

} // namespace concepts

#endif // level_hh
