// Test from Chuck Allison in C/C++ Users Journal, September 2000:
// "The Simplest Automated Unit Test Framework That Could Possibly Work"

#include "testcase.hh"
#include <iostream>
#include <typeinfo>     // Visual Studio requires /GR""
#include <cmath>
#include <limits>

namespace test {

  using namespace std;

  void TestCase::do_test(bool cond, const std::string& lbl,
			 const char* fname, long lineno) {
    if (!cond)
      do_fail(lbl, fname, lineno);
    else
      _succeed();
  }

  void TestCase::do_numtest(double num, double orig, const std::string& lbl,
			    const string& lbl2, const char* fname,
			    long lineno, const double tol) {
    string lblTOT(lbl);
    lblTOT.append(" <> ");
    lblTOT.append(lbl2);
    double res = 0;
    if (std::fabs(orig) > std::numeric_limits<double>::epsilon()*10)
      res = std::fabs( (num-orig) / orig );
    else
      res = std::fabs(num-orig);
    if (res > tol)
      do_fail(lblTOT, fname, lineno);
    else
      _succeed();
  }

  void TestCase::do_fail(const std::string& lbl,
			 const char* fname, long lineno) {
    ++m_nFail;
    if (m_osptr) {
      *m_osptr << typeid(*this).name()
	       << "failure: (" << lbl << ") , "
	       << fname << " (line " << lineno << ")\n";
    }
  }

  long TestCase::report() const {
    if (m_osptr) {
      *m_osptr << "Test \"" 
	       << typeid(*this).name() << "\":\n"
	       << "\tPassed: " << m_nPass
	       << "\tFailed: " << m_nFail
	       << endl;
    }
    return m_nFail;
  }

} // namespace test
