/* testing exceptions
 */

#include "basics/exceptions.hh"
#include <cstdlib>
#include <iostream>
#include <unistd.h>
#include "excTest.hh"

namespace test {

  void ExceptionTest::run() {
    testAssert(); testBase(); testMissingFeature();
  }

  void ExceptionTest::testAssert() {
    try {
      conceptsAssert(false, concepts::Assertion());
#if DEBUG
      _fail("Should have thrown Assertion!");
#else
      _succeed();
#endif
    }
    catch (concepts::Assertion& e) {
#if DEBUG
      _succeed();
#else
      _fail("Should not have thrown Assertion!");
#endif
    }

    try {
      conceptsAssert(true, concepts::Assertion());
      _succeed();
    }
    catch (concepts::Assertion& e) {
      _fail("Should not have thrown Assertion!");
    }

    try {
      conceptsAssert3(false, concepts::Assertion(), "add " << "a message"
		      << " here");
#if DEBUG
      _fail("Should have thrown Assertion!");
#else
      _succeed();
#endif
    }
    catch (concepts::Assertion& e) {
#if DEBUG
      _succeed();
#else
      _fail("Should not have thrown Assertion!");
#endif
    }
  }

  void ExceptionTest::testBase() {
    try {
      throw conceptsException(concepts::ExceptionBase());
      _fail("Should have thrown ExceptionBase!");
    }
    catch (concepts::ExceptionBase& e) {
      _succeed();
    }
  }

  void ExceptionTest::testMissingFeature() {
    try {
      throw conceptsException(concepts::MissingFeature("test"));
      _fail("Should have thrown MissingFeature!");
    }
    catch (concepts::MissingFeature& e) {
      _succeed();
    }
  }

} // namespace test

int main(int argc, char** argv) {
  concepts::Stacktrace::doit() = false;
  test::ExceptionTest t;
  t.run();
  return t.report();
}
