// space of constraints

#include <memory>
#include "space.hh"
#include "basics/typedefs.hh"
#include "basics/debug.hh"
#include "space/element.hh"
#include "space/tmatrix.hh"
#include "element.hh"
#include "analytical.hh"

// debugging

#define spaceConstr_D 0

namespace constraints {

  // ***************************************************************** Space **

  template<typename F>
  Space<F>::Space(concepts::Space<F>& spc, ConstraintsList<F>& constr) :
    dim_(0), nelm_(0), elm_(0), compSpc_(spc), constr_(constr),
    elmPairList_(0) {
    rebuild();
  }

  template<typename F>
  Space<F>::~Space() {
    concepts::Joiner<Element<F>*, 1>::destructor(elm_);
  }

  template<typename F>
  typename Space<F>::Scan* Space<F>::scan() const {
    return new concepts::PListScan<Element<F> >(*elm_);
  }

  template<typename F>
  void Space<F>::rebuild() {
    dim_ = 0;
    nelm_ = 0;

    concepts::Joiner<Element<F>*, 1>* endElm = 0;

    std::auto_ptr<typename concepts::Space<F>::Scanner> sc(compSpc_.scan());
    while ( *(sc.get()) ) {
      concepts::Element<F>& elm = (*sc)++;
      DEBUGL(spaceConstr_D, elm);
      concepts::TColumn<F>* t = 0;
      constr_.assembly(elm, t, dim_);

      Element<F>* constrElm = new Element<F>(elm, t);
      DEBUGL(spaceConstr_D, "new element = " << *constrElm);

      if (elm_ == 0) {
	endElm = elm_ = new concepts::Joiner<Element<F>*, 1>(constrElm, elm_);
      } else {
	// create new container (to be added to end of list), pointing to 0
	concepts::Joiner<Element<F>*, 1>* newElm =
	  new concepts::Joiner<Element<F>*, 1>(constrElm, 0);
	// add newElm to end of list
	(*endElm)[0] = newElm;
	// update end of list pointer
	endElm = newElm;
      }
      ++nelm_;

      // clean up
      while (t != 0) {
	concepts::TColumn<F>* foo = t; t = t->link(); delete foo;
      }
    }

    elmPairList_.reset(new concepts::ElementPairList<F>(0, false));
    std::auto_ptr<typename Space<F>::Scan> sc2(scan());
    while (*sc2) {
      concepts::ElementPair<F>* elm = &((*sc2)++);
      elmPairList_->add(elm);
    }
  }

  template<typename F>
  std::ostream& Space<F>::info(std::ostream& os) const {
    return os << "constraints::Space(dim = " << dim() << ", nelm = "
	      << nelm() << ')';
  }

  using concepts::Real;
  template class Space<Real>;

} // namespace constraints
