// space of constraints

#ifndef constrSpace_hh
#define constrSpace_hh

#include <memory>
#include "space/space.hh"
#include "toolbox/scannerConnectors.hh"
#include "element.hh"

namespace concepts {

  // ************************************************************** Scanners **

  template<typename F>
  class Scan<constraints::Element<F> > : public Scan<Element<Real> > {
  public:
    virtual constraints::Element<F>& operator++(int) = 0;
  };

} // namespace concepts

namespace constraints {

  template<typename F>
  class ConstraintsList;

  // ***************************************************************** Space **

  /** Space of constraints
      @author Philipp Frauenfelder, 2002
  */
  template<class F>
  class Space : public concepts::Space<F> {
  public:
    /** Constructor.
	@param spc Computational space
	@param constr List of constraints
    */
    Space(concepts::Space<F>& spc, ConstraintsList<F>& constr);
    virtual ~Space();

    typedef concepts::Scan<Element<F> > Scan;

    virtual uint dim() const { return dim_; }
    virtual uint nelm() const { return nelm_; }
    virtual Scan* scan() const;

    /// Rebuilds the space
    void rebuild();

    /// Returns a list of element pairs needed for matrix assembly
    const concepts::ElementPairList<F>& elmPairList() { return *elmPairList_; }
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    /// Dimension of the space
    uint dim_;
    /// Number of elements in the space
    uint nelm_;
    /// Linear list of elements
    concepts::Joiner<Element<F>*, 1>* elm_;

    /// Computational space
    concepts::Space<F>& compSpc_;

    /// List of constraints
    ConstraintsList<F>& constr_;

    /// List of element pairs to be used in matrix assembly
    std::auto_ptr<concepts::ElementPairList<F> > elmPairList_;
  };

} // namespace constraints

#endif // constrSpace_hh
