/* testing analytical constraints
 */

#include <memory>
#include "basics/exceptions.hh"
#include "basics/typedefs.hh"
#include "constraints/analytical.hh"

template<typename F>
class DirichletBoundary : public constraints::AnalyticalConstraint<F> {
public:
  virtual void lengthT(const concepts::Element<F>& e, uint& length);
  virtual void assembly(const concepts::Element<F>& e,
			concepts::TColumn<F>*& t, uint& n, uint& j,
			uint length);
  virtual constraints::AnalyticalConstraint<F>* clone() const;
  virtual void localMatrix(concepts::ElementMatrix<F>& em, uint& j,
			   const constraints::Element<F>& elm) const;
  virtual void localVector(concepts::ElementMatrix<F>& em, uint& j,
			   const constraints::Element<F>& elm) const;
protected:
  virtual std::ostream& info(std::ostream& os) const;
};

template<typename F>
void DirichletBoundary<F>::lengthT(const concepts::Element<F>&,
				   uint&) {
}

template<typename F>
void DirichletBoundary<F>::assembly(const concepts::Element<F>&,
				    concepts::TColumn<F>*&, uint&, uint&,
				    uint) {}

template<typename F>
constraints::AnalyticalConstraint<F>* DirichletBoundary<F>::clone() const {
  return new DirichletBoundary<F>(*this);
}

template<typename F>
void DirichletBoundary<F>::localMatrix
(concepts::ElementMatrix<F>&, uint&,
 const constraints::Element<F>&) const {}

template<typename F>
void DirichletBoundary<F>::localVector
(concepts::ElementMatrix<F>&, uint&,
 const constraints::Element<F>&) const {}

template<typename F>
std::ostream& DirichletBoundary<F>::info(std::ostream& os) const {
  return os << "DirichletBoundary()";
}

int main(int, char**) {

  using concepts::Real;

  try {
    DirichletBoundary<Real> a;
    std::cout << "a = " << a << std::endl;

    constraints::ConstraintsList<Real> b;
    b.add(new DirichletBoundary<Real>());
    b.add(a);
    std::cout << "b = " << b << std::endl;
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
