// eigensolvers

#ifndef eigens_hh
#define eigens_hh

#include "basics/outputOperator.hh"
#include "basics/typedefs.hh"
#include "function/vector.hh"
#include "toolbox/array.hh"

namespace eigensolver {

  using concepts::Real;

  // *********************************************************** EigenSolver **

  /** Interafce for eigenvalue solvers. Every new eigenvalue solver
      should implement this interface.
      @author Norbert Fernandes, 2002
  */
  class EigenSolver : public concepts::OutputOperator {
  public:
    /// Returns an array with the eigen values
    virtual const concepts::Array<Real>& getEV() = 0;
    /// Returns an array with the eigen functions
    virtual const concepts::Array<concepts::Vector<Real>*>& getEF() = 0;
    /// Returns the number of converged eigen pairs
    virtual uint converged() const = 0;
    /// Returns the number of iterations
    virtual uint iterations() const = 0;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  };

}

#endif // eigens_hh
