// fabric for eigensolver InexactInv

#ifndef inexactFabric_hh
#define inexactFabric_hh

#include <memory>
#include "inexactInv.hh"
#include "solverFabric.hh"

namespace eigensolver {

  // ****************************************************** InexactInvFabric **

  /** Fabric class for eigensolver inexact inverse iteration.
      @test test::GolubExample
      @test test::GolubExampleSum
      @test test::MaxwellTransmissionEVP
      @author Philipp Frauenfelder, 2003.
  */
  class InexactInvFabric : public SolverFabric {
  public:
    /** Constructor.
     */
    InexactInvFabric(const concepts::Vector<Real>& EFguess,
		     const Real innerTol = 0.5, const uint innerMax = 100,
		     const Real outerTol = 1e-6, const uint outerMax = 100) :
      solver_(0), EFguess_(EFguess), innerTol_(innerTol), outerTol_(outerTol),
      innerMax_(innerMax), outerMax_(outerMax) {}
    virtual ~InexactInvFabric();
    virtual InexactInv<Real>& operator()(concepts::Operator<Real>& stiff,
					 concepts::Operator<Real>& mass);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    std::auto_ptr<InexactInv<Real> > solver_;

    concepts::Vector<Real> EFguess_;
    Real innerTol_, outerTol_;
    uint innerMax_, outerMax_;

    InexactInvFabric(const InexactInvFabric&);
  };

} // namespace eigensolver

#endif // inexactFabric_hh
