/* Connectors.
 * connectivity information
 */

#include <iostream>
#include "connector.hh"

namespace concepts {

  // ************************************************************* Attribute **

  std::ostream& operator<<(std::ostream& os, const Attribute& a) {
    return os << "Attribute(" << a.attrib_ << ')';
  }

  // ******************************************************************* Key **

  std::ostream& operator<<(std::ostream& os, const Key& k) {
    return os << "Key(" << k.key_ << ')';
  }

  // *************************************************************** related **

  /* checks the relation ship of small and large recursively.
     small is potentially a 'child' (over many generations) of large.
     If small == large is detected, the recursion exited by throwing
     an exception of type uint.
  */
  uint related(const Connector& small,
	       const Connector& large, uint res) throw(uint) {
    // We found a small == large, therefore exit the recursion.
    if (small.key() == large.key()) {
#ifdef __SUNPRO_CC
      std::cout << "throwing exceptions might not work -> abort?" << std::endl;
#endif
      throw res;
    }

    // small can no longer be a child a large if small's key < large's key
    if (small.key().key() < large.key().key()) return 0;

    // check the children of large
    const Connector* cntr = 0;
    for (uint i = 0; (cntr = large.child(i)) != 0; ++i) {
      related(small, *cntr, res+1);
    }
    return 0;
  }

  // ************************************************************* Connector **

  std::ostream& Connector::info(std::ostream& os) const {
    return os << "Connector(key = " << key_ << ", attrib = " << attrib_ << ')';
  }

  int Connector::related(const Connector& conn) const {
    try {
      if (conn.key().key() > key().key())
	return concepts::related(conn, *this, 0);
      else return concepts::related(*this, conn, 0);
    }
    catch (uint& e) {
      if (conn.key().key() > key().key()) return -e;
      else return e;
    }
  }


  // ************************************************************ Connector* **

  uint Connector0::cnt_ = 0;
  uint Connector1::cnt_ = 0;
  uint Connector2::cnt_ = 0;
  uint Connector3::cnt_ = 0;

} // namespace concepts
