/* Topology.
 * how to subdivide a hex
 */

#include <algorithm>
#include "hexsubdiv.hh"
#include "topology3D.hh"

// debugging
#include "basics/debug.hh"

#define SubdivFace5_D 0

// HexSubdivision
#define HexSubdivisionFaceSubdiv4_D 0
#define HexSubdivisionFindFace_D 0

// debugging HexSubdiv8
#define HexSubdiv8Inst_D 0
#define HexSubdiv8Destr_D 0
#define HexSubdiv8createChildren_D 0

// debugging HexSubdiv2x
#define HexSubdiv2xInst_D 0
#define HexSubdiv2xDestr_D 0
#define HexSubdiv2xcreateChildren_D 0

// debugging HexSubdiv2y
#define HexSubdiv2yInst_D 0
#define HexSubdiv2yDestr_D 0
#define HexSubdiv2ycreateChildren_D 0

// debugging HexSubdiv2z
#define HexSubdiv2zInst_D 0
#define HexSubdiv2zDestr_D 0
#define HexSubdiv2zcreateChildren_D 0

// debugging HexSubdiv4x
#define HexSubdiv4xInst_D 0
#define HexSubdiv4xDestr_D 0
#define HexSubdiv4xcreateChildren_D 0
#define HexSubdiv4xdestroyChildren_D 0

// debugging HexSubdiv4y
#define HexSubdiv4yInst_D 0
#define HexSubdiv4yDestr_D 0
#define HexSubdiv4ycreateChildren_D 0
#define HexSubdiv4ydestroyChildren_D 0

// debugging HexSubdiv4z
#define HexSubdiv4zInst_D 0
#define HexSubdiv4zDestr_D 0
#define HexSubdiv4zcreateChildren_D 0
#define HexSubdiv4zdestroyChildren_D 0

namespace concepts {

  // ******************************************************** HexSubdivision **

  void HexSubdivision::subFace0011(Quad& q, Z2 rho, Z4 tau,
				   Quad*& A, Quad*& B) const {
    A = q.child(0);
    B = q.child(1);
    if ( (tau/2) ^ rho ) std::swap(A, B);
  }

  void HexSubdivision::subFace1100(Quad& q, Z2 rho, Z4 tau,
				   Quad*& A, Quad*& B) const {
    subFace0011(q, rho.succ(), tau, A, B);
  }

  void HexSubdivision::subFace0110(Quad& q, Z2 rho, Z4 tau,
				   Quad*& A, Quad*& B) const {
    A = q.child(0);
    B = q.child(1);
    if ( (tau/2) ^ (tau%2) ) std::swap(A, B);
  }

  void HexSubdivision::subFace1001(Quad& q, Z2 rho, Z4 tau,
				   Quad*& A, Quad*& B) const {
    subFace0110(q, rho, tau, B, A);
  }

  void HexSubdivision::subFace0123(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
				   Quad*& C, Quad*& D) const {
    Z4 i = rho + tau;
    A = q.child(i);
    rho?--i:++i;
    B = q.child(i);
    rho?--i:++i;
    C = q.child(i);
    rho?--i:++i;
    D = q.child(i);
  }

  void HexSubdivision::subFace1230(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
				   Quad*& C, Quad*& D) const {
    Z4 i = rho.succ() + tau;
    A = q.child(i);
    rho?--i:++i;    // is from 4x and 4y. 4z has -- and ++ swapped
    B = q.child(i); // is corrected in 4z by calling it with B and C swapped
    rho?--i:++i;
    C = q.child(i);
    rho?--i:++i;
    D = q.child(i);
  }

  void HexSubdivision::subFace2301(Quad& q, Z2 rho, Z4 tau, Quad*& A, Quad*& B,
				   Quad*& C, Quad*& D) const {
    Z4 i = 2 + rho + tau;
    A = q.child(i);
    rho?++i:--i;
    B = q.child(i);
    rho?++i:--i;
    C = q.child(i);
    rho?++i:--i;
    D = q.child(i);
  }

  bool HexSubdivision::faceSubdiv4(Quad& q) const throw(StrategyChange) {
    try {
      q.setStrategy(QuadSubdiv4::instance());
      return true; // it worked in one shot
    }
    catch (StrategyChange& e) {
      DEBUGL(HexSubdivisionFaceSubdiv4_D, "caught " << e);
      const QuadSubdivision* qSubdiv = q.getStrategy();
      conceptsAssert(qSubdiv != 0, Assertion());
      DEBUGL(HexSubdivisionFaceSubdiv4_D, "strategy is " << *qSubdiv);
      Quad* qTmp;

      // loop over all children and set orthogonal strategy to qSubdiv
      for (uint i = 0; (qTmp = q.child(i)) != 0; ++i) {
	DEBUGL(HexSubdivisionFaceSubdiv4_D, "children: " << *qTmp);
	if (qSubdiv == QuadSubdiv2H::instance()) {
	  qTmp->setStrategy(QuadSubdiv2V::instance());
	} else {
	  conceptsAssert(qSubdiv == QuadSubdiv2V::instance(), Assertion());
	  qTmp->setStrategy(QuadSubdiv2H::instance());
	}
      }
    }
    return false; // subdivision hat to be done in two shots
  }

  Quad* HexSubdivision::findFace(const Vertex* const v, Quad* const q[4],
				 const uint tau, const uint rho,
				 const uint vtxIdx) const {
    DEBUGL(HexSubdivisionFindFace_D, "looking for " << *v << " in " << *q[0]
	   << ", " << *q[1] << ", " << *q[2] << ", " << *q[3]
	   << " with tau = " << tau << ", rho = " << rho << ", vtxIdx = "
	   << vtxIdx);
    uint vtxIdxNeg;
    switch (vtxIdx) {
    case 0: vtxIdxNeg = 1; break;
    case 1: vtxIdxNeg = 0; break;
    case 2: vtxIdxNeg = 3; break;
    case 3: vtxIdxNeg = 2; break;
    default: throw conceptsException(ExceptionBase());
    }
    uint idx = (tau + (rho?vtxIdxNeg:vtxIdx))%4;
    DEBUGL(HexSubdivisionFindFace_D, "looking for vertex with index " << idx);
    for (uint i = 0; i < 4; ++i) {
      Vertex* qv = q[i]->vertex(idx);
      if (v == qv) {
	DEBUGL(HexSubdivisionFindFace_D, "q[" << i << "] = " << *(q[i])
	       << " contains " << *v);
	return q[i];
      }
    }

    // Strangely enough, the quad was not found. Try something else here:
    // Just look at every vertex in every quad.
    DEBUGL(1, "did not find " << *v << " in " << *q[0]
	   << ", " << *q[1] << ", " << *q[2] << ", " << *q[3]
	   << " with tau = " << tau << ", rho = " << rho << ", vtxIdx = "
	   << vtxIdx << ", looking at all vertices now");
    for (uint i = 0; i < 4; ++i) {
      Vertex* qv = 0;
      for (uint j = 0; (qv = q[i]->vertex(j)) != 0; ++j) {
	if (v == qv) {
	  DEBUGL(1, "q[" << i << "] = " << *(q[i]) << " contains " << *v);
	  return q[i];
	}
      } // for vertices
    } // for quads    
    conceptsAssert(false, Assertion());
    return 0;
  }

  void HexSubdivision::subdivFace0(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    if (normal) {
      subFace0123(*h.face(0), h.rho(0), h.tau(0), A, B, C, D);
    } else {
      Quad* quads[4];
      quads[0] = h.face(0)->child(0)->child(0);
      quads[1] = h.face(0)->child(0)->child(1);
      quads[2] = h.face(0)->child(1)->child(0);
      quads[3] = h.face(0)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(0), rho = h.rho(0);
      A = findFace(h.vertex(0), quads, tau, rho, 0);
      B = findFace(h.vertex(1), quads, tau, rho, 1);
      C = findFace(h.vertex(2), quads, tau, rho, 2);
      D = findFace(h.vertex(3), quads, tau, rho, 3);
    }
  }

  void HexSubdivision::subdivFace1(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    if (normal) {
      subFace1230(*h.face(1), h.rho(1), h.tau(1), A, B, C, D);
    } else {
      Quad* quads[4];
      quads[0] = h.face(1)->child(0)->child(0);
      quads[1] = h.face(1)->child(0)->child(1);
      quads[2] = h.face(1)->child(1)->child(0);
      quads[3] = h.face(1)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(1), rho = h.rho(1);
      A = findFace(h.vertex(0), quads, tau, rho, 1);
      B = findFace(h.vertex(1), quads, tau, rho, 2);
      C = findFace(h.vertex(5), quads, tau, rho, 3);
      D = findFace(h.vertex(4), quads, tau, rho, 0);
    }
  }

  void HexSubdivision::subdivFace2(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    if (normal) {
      subFace1230(*h.face(2), h.rho(2), h.tau(2), A, B, C, D);
    } else {
      Quad* quads[4];
      quads[0] = h.face(2)->child(0)->child(0);
      quads[1] = h.face(2)->child(0)->child(1);
      quads[2] = h.face(2)->child(1)->child(0);
      quads[3] = h.face(2)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(2), rho = h.rho(2);
      A = findFace(h.vertex(1), quads, tau, rho, 1);
      B = findFace(h.vertex(2), quads, tau, rho, 2);
      C = findFace(h.vertex(6), quads, tau, rho, 3);
      D = findFace(h.vertex(5), quads, tau, rho, 0);
    }
  }

  void HexSubdivision::subdivFace3(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    if (normal) {
      subFace2301(*h.face(3), h.rho(3), h.tau(3), B, A, D, C);
    } else {
      Quad* quads[4];
      quads[0] = h.face(3)->child(0)->child(0);
      quads[1] = h.face(3)->child(0)->child(1);
      quads[2] = h.face(3)->child(1)->child(0);
      quads[3] = h.face(3)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(3), rho = h.rho(3);
      B = findFace(h.vertex(3), quads, tau, rho, 2);
      A = findFace(h.vertex(2), quads, tau, rho, 1);
      D = findFace(h.vertex(6), quads, tau, rho, 0);
      C = findFace(h.vertex(7), quads, tau, rho, 3);
    }
  }

  void HexSubdivision::subdivFace4(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    if (normal) {
      subFace2301(*h.face(4), h.rho(4), h.tau(4), B, A, D, C);
    } else {
      Quad* quads[4];
      quads[0] = h.face(4)->child(0)->child(0);
      quads[1] = h.face(4)->child(0)->child(1);
      quads[2] = h.face(4)->child(1)->child(0);
      quads[3] = h.face(4)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(4), rho = h.rho(4);
      B = findFace(h.vertex(0), quads, tau, rho, 2);
      A = findFace(h.vertex(3), quads, tau, rho, 1);
      D = findFace(h.vertex(7), quads, tau, rho, 0);
      C = findFace(h.vertex(4), quads, tau, rho, 3);
    }
  }

  void HexSubdivision::subdivFace5(Hexahedron& h,
				   Quad*& A, Quad*& B, Quad*& C, Quad*& D,
				   bool normal) const {
    DEBUGL(SubdivFace5_D, h << ", normal = " << normal);
    if (normal) {
      subFace1230(*h.face(5), h.rho(5), h.tau(5), D, A, B, C);
    } else {
      DEBUGL(SubdivFace5_D, "subdividing " << *h.face(5) << " in two shots");
      Quad* quads[4];
      quads[0] = h.face(5)->child(0)->child(0);
      quads[1] = h.face(5)->child(0)->child(1);
      quads[2] = h.face(5)->child(1)->child(0);
      quads[3] = h.face(5)->child(1)->child(1);
      for (uint i = 0; i < 4; ++i)
	conceptsAssert(quads[i], Assertion());

      const uint tau = h.tau(5), rho = h.rho(5);
      A = findFace(h.vertex(7), quads, tau, rho, 2);
      B = findFace(h.vertex(6), quads, tau, rho, 3);
      C = findFace(h.vertex(5), quads, tau, rho, 0);
      D = findFace(h.vertex(4), quads, tau, rho, 1);
    }
  }

  // ************************************************************ HexSubdiv8 **

  std::auto_ptr<HexSubdiv8> HexSubdiv8::instance_(0);

  void HexSubdiv8::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv8createChildren_D, "subdiving " << q);

      bool quadNormal[6];
      for (uint i = 0; i < 6; ++i)
	quadNormal[i] = faceSubdiv4(*(q.quad_[i]));

      Hexahedron** hex;

      // quadxy is the child y of quad x of the hexahedron
      Quad* quad00, * quad10, * quad20, * quad30, * quad40, * quad50;
      Quad* quad01, * quad11, * quad21, * quad31, * quad41, * quad51;
      Quad* quad02, * quad12, * quad22, * quad32, * quad42, * quad52;
      Quad* quad03, * quad13, * quad23, * quad33, * quad43, * quad53;

      // quadNax is one of the new ('N') quads with normal direction a,
      // "child" x
      Quad* quadNx0, * quadNy0, * quadNz0;
      Quad* quadNx1, * quadNy1, * quadNz1;
      Quad* quadNx2, * quadNy2, * quadNz2;
      Quad* quadNx3, * quadNy3, * quadNz3;

      // vtxx is the vertex in the middle of quad_[x]
      Vertex* vtx0, * vtx1, * vtx2, * vtx3, * vtx4, * vtx5;

      // the new edges into the middle of the hex (edgx is from vtxx to vtxN
      Edge* edg0, * edg1, * edg2, * edg3, * edg4, * edg5;
      Edge* tmp1, * tmp2;

      // get the needed quads
      subdivFace0(q, quad00, quad01, quad02, quad03, quadNormal[0]);
      if (q.rho_[0]) {
	vtx0 = quad00->vertex(q.tau_[0].pred());
	tmp1 = quad00->edge(q.tau_[0].pred());
      } else {
	vtx0 = quad00->vertex( ((Z4)q.tau_[0].succ()).succ() );
	tmp1 = quad00->edge(q.tau_[0].succ());
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx0: " << *vtx0);

      subdivFace1(q, quad10, quad11, quad12, quad13, quadNormal[1]);
      if (q.rho_[1]) {
	vtx1 = quad10->vertex( ((Z4)q.tau_[1].succ()).succ() );
      } else {
	vtx1 = quad10->vertex(q.tau_[1].pred());
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx1: " << *vtx1);

      subdivFace2(q, quad20, quad21, quad22, quad23, quadNormal[2]);
      if (q.rho_[2]) {
	vtx2 = quad21->vertex(q.tau_[2].succ());
      } else {
	vtx2 = quad21->vertex(q.tau_[2]);
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx2: " << *vtx2);

      subdivFace3(q, quad30, quad31, quad32, quad33, quadNormal[3]);
      if (q.rho_[3]) {
	vtx3 = quad33->vertex(q.tau_[3].pred());
      } else {
	vtx3 = quad30->vertex(q.tau_[3].pred());
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx3: " << *vtx3);

      subdivFace4(q, quad40, quad41, quad42, quad43, quadNormal[4]);
      if (q.rho_[4]) {
	vtx4 = quad41->vertex(q.tau_[4].succ());
      } else {
	vtx4 = quad41->vertex(q.tau_[4]);
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx4: " << *vtx4);

      subdivFace5(q, quad50, quad51, quad52, quad53, quadNormal[5]);
      if (q.rho_[5]) {
	vtx5 = quad50->vertex(q.tau_[5].succ());
      } else {
	vtx5 = quad50->vertex(q.tau_[5]);
      }
      DEBUGL(HexSubdiv8createChildren_D, "vtx5: " << *vtx5);

      // the new vertex in the middle of the hex
      Vertex* vtxN = new Vertex(q.attrib_);

      // new hex 0
      DEBUGL(HexSubdiv8createChildren_D, "child 0");
      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quad00);
      edg0 = new Edge(*vtx0, *vtxN, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "edg0: " << *edg0);

      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quad10);
      edg1 = new Edge(*vtx1, *vtxN, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "edg1: " << *edg1);
      tmp2 = quad10->edge( ((Z4)q.tau_[1].succ()).succ() );
      quadNx0 = new Quad(*tmp1, *edg0, *edg1, *tmp2, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quadNx0);

      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quad41);

      edg4 = new Edge(*vtx4, *vtxN, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "edg4: " << *edg4);
      tmp1 = quad03->edge(q.tau_[0]);
      tmp2 = quad41->edge(q.tau_[4]);
      quadNy1 = new Quad(*tmp1, *tmp2, *edg4, *edg0, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quadNy1);

      if (q.rho_[4])
	tmp1 = quad41->edge(q.tau_[4].succ());
      else
	tmp1 = quad41->edge(q.tau_[4].pred());
      if (q.rho_[1])
	tmp2 = quad10->edge(q.tau_[1].succ());
      else
	tmp2 = quad10->edge(q.tau_[1].pred());
      quadNz0 = new Quad(*tmp1, *tmp2, *edg1, *edg4, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quadNz0);

      hex = &(q.chld_);
      *hex = new Hexahedron(*quad00, 
			    *quad10, *quadNx0, *quadNy1, *quad41, 
			    *quadNz0, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex0: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 1
      DEBUGL(HexSubdiv8createChildren_D, "child " << "1");

      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quad01);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quad11);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quad20);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quadNx0);

      edg2 = new Edge(*vtx2, *vtxN, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "edg2: " << *edg2);
      tmp1 = quad21->edge(q.tau_[2]);
      tmp2 = quad02->edge(q.tau_[0]);
      quadNy0 = new Quad(*tmp2, *tmp1, *edg2, *edg0, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quadNy0);

      if (q.rho_[1])
	tmp1 = quad11->edge(q.tau_[1].succ());
      else
	tmp1 = quad11->edge(q.tau_[1].pred());
      if (q.rho_[2])
	tmp2 = quad20->edge(q.tau_[2].succ());
      else
	tmp2 = quad20->edge(q.tau_[2].pred());
      quadNz1 = new Quad(*tmp1, *tmp2, *edg2, *edg1, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quadNz1);

      *hex = new Hexahedron(*quad01,
			    *quad11, *quad20, *quadNy0, *quadNx0,
			    *quadNz1, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex1: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 2
      DEBUGL(HexSubdiv8createChildren_D, "child " << "2");

      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quad02);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quadNy0);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quad21);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quad30);

      edg3 = new Edge(*vtx3, *vtxN, q.attrib_);
      if(q.rho_[0])
	tmp1 = quad02->edge(q.tau_[0].succ());
      else
	tmp1 = quad02->edge(q.tau_[0].pred());
      tmp2 = quad31->edge(q.tau_[3]);
      quadNx1 = new Quad(*tmp1, *tmp2, *edg3, *edg0, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quadNx1);

      if (q.rho_[2])
	tmp1 = quad21->edge(q.tau_[2].succ());
      else
	tmp1 = quad21->edge(q.tau_[2].pred());
      if (q.rho_[3])
	tmp2 = quad30->edge(q.tau_[3].succ());
      else
	tmp2 = quad30->edge(q.tau_[3].pred());
      quadNz2 = new Quad(*tmp1, *tmp2, *edg3, *edg2, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quadNz2);

      *hex = new Hexahedron(*quad02,
			    *quadNy0, *quad21, *quad30, *quadNx1,
			    *quadNz2, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex2: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 3
      DEBUGL(HexSubdiv8createChildren_D, "child " << "3");

      if (q.rho_[3])
	tmp1 = quad31->edge(q.tau_[3].succ());
      else
	tmp1 = quad31->edge(q.tau_[3].pred());
      if (q.rho_[4])
	tmp2 = quad40->edge(q.tau_[4].succ());
      else
	tmp2 = quad40->edge(q.tau_[4].pred());
      quadNz3 = new Quad(*tmp1, *tmp2, *edg4, *edg3, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quad03);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quadNy1);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quadNx1);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quad31);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quad40);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quadNz3);

      *hex = new Hexahedron(*quad03,
			    *quadNy1, *quadNx1, *quad31, *quad40,
			    *quadNz3, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex3: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 4
      DEBUGL(HexSubdiv8createChildren_D, "child " << "4");

      edg5 = new Edge(*vtx5, *vtxN, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "edg5: " << *edg5);
      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quadNz0);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quad13);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quad42);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quad53);

      tmp1 = quad12->edge(q.tau_[1]);
      if (q.rho_[5])
	tmp2 = quad53->edge(q.tau_[5].succ());
      else
	tmp2 = quad53->edge(q.tau_[5].pred());
      quadNx3 = new Quad(*tmp1, *tmp2, *edg5, *edg1, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quadNx3);

      tmp1 = quad50->edge(q.tau_[5]);
      tmp2 = quad42->edge(q.tau_[4]);
      quadNy2 = new Quad(*tmp1, *tmp2, *edg4, *edg5, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quadNy2);

      *hex = new Hexahedron(*quadNz0,
			    *quad13, *quadNx3, *quadNy2, *quad42,
			    *quad53, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex4: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 5
      DEBUGL(HexSubdiv8createChildren_D, "child " << "5");

      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quadNz1);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quad12);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quad23);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quadNx3);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quad52);

      tmp1 = quad51->edge(q.tau_[5]);
      tmp2 = quad22->edge(q.tau_[2]);
      quadNy3 = new Quad(*tmp1, *tmp2, *edg2, *edg5, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quadNy3);

      *hex = new Hexahedron(*quadNz1,
			    *quad12, *quad23, *quadNy3, *quadNx3,
			    *quad52, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex5: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 6
      DEBUGL(HexSubdiv8createChildren_D, "child " << "6");

      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quadNz2);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quadNy3);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quad22);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quad33);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quad51);

      if (q.rho_[5])
	tmp1 = quad51->edge(q.tau_[5].pred());
      else
	tmp1 = quad51->edge(q.tau_[5].succ());
      tmp2 = quad32->edge(q.tau_[3]);
      quadNx2 = new Quad(*tmp1, *tmp2, *edg3, *edg5, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quadNx2);

      *hex = new Hexahedron(*quadNz2,
			    *quadNy3, *quad22, *quad33, *quadNx2,
			    *quad51, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex6: " << **hex);
      hex = &(*hex)->lnk_;

      // new hex 7
      DEBUGL(HexSubdiv8createChildren_D, "child " << "7");

      DEBUGL(HexSubdiv8createChildren_D, "quad0: " << *quadNz3);
      DEBUGL(HexSubdiv8createChildren_D, "quad1: " << *quadNy2);
      DEBUGL(HexSubdiv8createChildren_D, "quad2: " << *quadNx2);
      DEBUGL(HexSubdiv8createChildren_D, "quad3: " << *quad32);
      DEBUGL(HexSubdiv8createChildren_D, "quad4: " << *quad43);
      DEBUGL(HexSubdiv8createChildren_D, "quad5: " << *quad50);

      *hex = new Hexahedron(*quadNz3,
			    *quadNy2, *quadNx2, *quad32, *quad43,
			    *quad50, q.attrib_);
      DEBUGL(HexSubdiv8createChildren_D, "hex7: " << **hex);
    }
  }

  void HexSubdiv8::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      // delete new vertex
      delete q.child(0)->vertex(6);

      // delete new edges
      delete q.child(0)->edge(6);
      delete q.child(0)->edge(9);
      delete q.child(0)->edge(10);
      delete q.child(6)->edge(0);
      delete q.child(6)->edge(3);
      delete q.child(6)->edge(4);

      // delete new quads
      delete q.child(0)->quad_[2];
      delete q.child(0)->quad_[3];
      delete q.child(0)->quad_[5];
      delete q.child(1)->quad_[3];
      delete q.child(1)->quad_[5];
      delete q.child(2)->quad_[4];
      delete q.child(2)->quad_[5];
      delete q.child(3)->quad_[5];
      delete q.child(4)->quad_[2];
      delete q.child(4)->quad_[3];
      delete q.child(5)->quad_[3];
      delete q.child(6)->quad_[4];

      // delete new hexahedrons
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv8* HexSubdiv8::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv8Inst_D, "creating new instance");
      instance_.reset(new HexSubdiv8());
    }
    return instance_.get();
  }

  HexSubdiv8::~HexSubdiv8() {
    DEBUGL(HexSubdiv8Destr_D, "done.");
  }

  std::ostream& HexSubdiv8::info(std::ostream& os) const {
    return os << "HexSubdiv8()";
  }

  // *********************************************************** HexSubdiv2x **

  std::auto_ptr<HexSubdiv2x> HexSubdiv2x::instance_(0);

  void HexSubdiv2x::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv2xcreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 0, 1, 3 and 5
      Edge *e0 = 0, *e1 = 0, *e2 = 0, *e3 = 0;
      if (q.tau_[0]%2)
	q.quad_[0]->setStrategy(QuadSubdiv2H::instance());
      else
	q.quad_[0]->setStrategy(QuadSubdiv2V::instance());

      if (q.tau_[1]%2)
	q.quad_[1]->setStrategy(QuadSubdiv2V::instance());
      else
	q.quad_[1]->setStrategy(QuadSubdiv2H::instance());

      if (q.tau_[3]%2)
	q.quad_[3]->setStrategy(QuadSubdiv2V::instance());
      else
	q.quad_[3]->setStrategy(QuadSubdiv2H::instance());

      if (q.tau_[5]%2)
	q.quad_[5]->setStrategy(QuadSubdiv2H::instance());
      else
	q.quad_[5]->setStrategy(QuadSubdiv2V::instance());

      if (q.tau_[0]%2)
	e0 = q.quad_[0]->child(0)->edge(2);
      else
	e0 = q.quad_[0]->child(0)->edge(1);

      if (q.tau_[1]%2)
	e3 = q.quad_[1]->child(0)->edge(1);
      else
	e3 = q.quad_[1]->child(0)->edge(2);

      if (q.tau_[3]%2)
	e1 = q.quad_[3]->child(0)->edge(1);
      else
	e1 = q.quad_[3]->child(0)->edge(2);

      if (q.tau_[5]%2)
	e2 = q.quad_[5]->child(0)->edge(2);
      else
	e2 = q.quad_[5]->child(0)->edge(1);

      // create new face in the middle
      Quad* newQ = new Quad(*e0, *e1, *e2, *e3, q.attrib_);
      DEBUGL(HexSubdiv2xcreateChildren_D, "new face " << *newQ);

      // identify faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0, *qG = 0,
	*qH = 0;

      subFace0011(*q.quad_[0], q.rho_[0], q.tau_[0], qA, qB);
      subFace0110(*q.quad_[1], q.rho_[1], q.tau_[1], qC, qD);
      subFace1001(*q.quad_[3], q.rho_[3], q.tau_[3], qE, qF);
      subFace1100(*q.quad_[5], q.rho_[5], q.tau_[5], qG, qH);

      // create two new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*qA,
			    *qC, *newQ, *qE, *(q.quad_[4]),
			    *qG, q.attrib_);
      DEBUGL(HexSubdiv2xcreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qB,
			    *qD, *(q.quad_[2]), *qF, *newQ,
			    *qH, q.attrib_);
      DEBUGL(HexSubdiv2xcreateChildren_D, "child 1 = " << **hex);
    }
  }

  void HexSubdiv2x::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      delete q.child(0)->quad_[2];
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv2x* HexSubdiv2x::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv2xInst_D, "creating new instance");
      instance_.reset(new HexSubdiv2x());
    }
    return instance_.get();
  }

  HexSubdiv2x::~HexSubdiv2x() {
    DEBUGL(HexSubdiv2xDestr_D, "done.");
  }

  std::ostream& HexSubdiv2x::info(std::ostream& os) const {
    return os << "HexSubdiv2x()";
  }

  // *********************************************************** HexSubdiv2y **

  std::auto_ptr<HexSubdiv2y> HexSubdiv2y::instance_(0);

  void HexSubdiv2y::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv2ycreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 0, 2, 4 and 5
      uint I[4] = { 0, 2, 4, 5 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
      }

      Edge *e[4] = { 0, 0, 0, 0 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  e[i] = q.quad_[I[i]]->child(0)->edge(1);
	else
	  e[i] = q.quad_[I[i]]->child(0)->edge(2);
      }

      // create new face in the middle
      Quad* newQ = new Quad(*e[0], *e[1], *e[3], *e[2], q.attrib_);
      DEBUGL(HexSubdiv2ycreateChildren_D, "new face " << *newQ);

      // identify faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0,
	*qG = 0, *qH = 0;

      subFace0110(*q.quad_[0], q.rho_[0], q.tau_[0], qA, qB);
      subFace0110(*q.quad_[2], q.rho_[2], q.tau_[2], qC, qD);
      subFace0110(*q.quad_[5], q.rho_[5], q.tau_[5], qE, qF);
      subFace1001(*q.quad_[4], q.rho_[4], q.tau_[4], qG, qH);

      // create two new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*qA,
			    *(q.quad_[1]), *qC, *newQ, *qG,
			    *qE, q.attrib_);
      DEBUGL(HexSubdiv2ycreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qB,
			    *newQ, *qD, *(q.quad_[3]), *qH,
			    *qF, q.attrib_);
      DEBUGL(HexSubdiv2ycreateChildren_D, "child 1 = " << **hex);
    }
  }

  void HexSubdiv2y::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      delete q.child(0)->quad_[3];
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv2y* HexSubdiv2y::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv2yInst_D, "creating new instance");
      instance_.reset(new HexSubdiv2y());
    }
    return instance_.get();
  }

  HexSubdiv2y::~HexSubdiv2y() {
    DEBUGL(HexSubdiv2yDestr_D, "done.");
  }

  std::ostream& HexSubdiv2y::info(std::ostream& os) const {
    return os << "HexSubdiv2y()";
  }

  // *********************************************************** HexSubdiv2z **

  std::auto_ptr<HexSubdiv2z> HexSubdiv2z::instance_(0);

  void HexSubdiv2z::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv2zcreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 1, 2, 3 and 4
      uint I[4] = { 1, 2, 3, 4 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
      }

      Edge *e[4] = { 0, 0, 0, 0 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  e[i] = q.quad_[I[i]]->child(0)->edge(2);
	else
	  e[i] = q.quad_[I[i]]->child(0)->edge(1);
      }

      // create new face in the middle
      Quad* newQ = new Quad(*e[0], *e[1], *e[2], *e[3], q.attrib_);
      DEBUGL(HexSubdiv2zcreateChildren_D, "new face " << *newQ);

      // identify faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0,
	*qG = 0, *qH = 0;

      subFace1100(*q.quad_[1], q.rho_[1], q.tau_[1], qA, qB);
      subFace1100(*q.quad_[2], q.rho_[2], q.tau_[2], qC, qD);
      subFace1100(*q.quad_[3], q.rho_[3], q.tau_[3], qE, qF);
      subFace1100(*q.quad_[4], q.rho_[4], q.tau_[4], qG, qH);

      // create two new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*(q.quad_[0]),
			    *qA, *qC, *qE, *qG,
			    *newQ, q.attrib_);
      DEBUGL(HexSubdiv2zcreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*newQ,
			    *qB, *qD, *qF, *qH,
			    *(q.quad_[5]), q.attrib_);
      DEBUGL(HexSubdiv2zcreateChildren_D, "child 1 = " << **hex);
    }
  }

  void HexSubdiv2z::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      delete q.child(0)->quad_[5];
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv2z* HexSubdiv2z::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv2zInst_D, "creating new instance");
      instance_.reset(new HexSubdiv2z());
    }
    return instance_.get();
  }

  HexSubdiv2z::~HexSubdiv2z() {
    DEBUGL(HexSubdiv2zDestr_D, "done.");
  }

  std::ostream& HexSubdiv2z::info(std::ostream& os) const {
    return os << "HexSubdiv2z()";
  }

  // *********************************************************** HexSubdiv4x **

  std::auto_ptr<HexSubdiv4x> HexSubdiv4x::instance_(0);

  void HexSubdiv4x::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv4xcreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 0, 2, 4 and 5
      uint I[2] = { 0, 5 };
      for (uint i = 0; i < 2; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
      }
      I[0] = 1; I[1] = 3;
      for (uint i = 0; i < 2; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
      }

      // try to subdivide the other two faces into 4
      bool quad2Normal = faceSubdiv4(*(q.quad_[2]));
      bool quad4Normal = faceSubdiv4(*(q.quad_[4]));

      // create new edge
      Vertex* vtx1 = 0, *vtx2 = 0;
      if (quad2Normal)
	vtx1 = q.quad_[2]->child(0)->vertex(2);
      else {
	vtx1 = q.quad_[2]->child(0)->child(0)->vertex(2);
      }
      if (quad4Normal)
	vtx2 = q.quad_[4]->child(0)->vertex(2);
      else {
	vtx2 = q.quad_[4]->child(0)->child(0)->vertex(2);
      }
      Edge* newE = new Edge(*vtx1, *vtx2, q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "new edge " << *newE);

      // find subdivided faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0, *qG = 0,
	*qH = 0, *qI = 0, *qJ = 0, *qK = 0, *qL = 0, *qM = 0, *qN = 0, *qO,
	*qP = 0;

      subFace0110(*q.quad_[0], q.rho_[0], q.tau_[0], qA, qB);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qA = " << *qA);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qB = " << *qB);

      subFace1100(*q.quad_[1], q.rho_[1], q.tau_[1], qC, qD);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qC = " << *qC);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qD = " << *qD);

      subdivFace2(q, qE, qF, qG, qH, quad2Normal);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qE = " << *qE);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qF = " << *qF);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qG = " << *qG);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qH = " << *qH);

      subFace1100(*q.quad_[3], q.rho_[3], q.tau_[3], qI, qJ);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qI = " << *qI);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qJ = " << *qJ);

      subdivFace4(q, qL, qK, qN, qM, quad4Normal);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qK = " << *qK);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qL = " << *qL);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qM = " << *qM);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qN = " << *qN);

      subFace0110(*q.quad_[5], q.rho_[5], q.tau_[5], qO, qP);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qO = " << *qO);
      DEBUGL(HexSubdiv4xcreateChildren_D, "qP = " << *qP);

      // create new quads
      Quad* newQ0 = new Quad(*(q.quad_[0]->child(0)->edge(2-(q.tau_[0]%2))),
			     *(qF->edge(q.tau_[2])),
			     *newE,
			     *(qK->edge(q.tau_[4])), q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "new quad 0 " << *newQ0);

      Quad* newQ1 = new Quad(*(q.quad_[1]->child(0)->edge(1+(q.tau_[1]%2))),
			     *(qE->edge( (q.tau_[2]+(q.rho(2)?1:3))%4 )),
			     *newE,
			     *(qK->edge( (q.tau_[4]+(q.rho(4)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "new quad 1 " << *newQ1);

      Quad* newQ2 = new Quad(*newE,
			     *(qF->edge( (q.tau_[2]+(q.rho(2)?1:3))%4 )),
			     *(q.quad_[3]->child(0)->edge(1+(q.tau_[3]%2))),
			     *(qL->edge( (q.tau_[4]+(q.rho(4)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "new quad 2 " << *newQ2);

      Quad* newQ3 = new Quad(*(qG->edge(q.tau_[2])),
			     *newE,
			     *(qN->edge(q.tau_[4])),
			     *(q.quad_[5]->child(0)->edge(2-(q.tau_[5]%2))),
			     q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "new quad 3 " << *newQ3);

      // create four new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*qA,
			    *qC, *qE, *newQ0, *qK,
			    *newQ1, q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qB,
			    *newQ0, *qF, *qI, *qL,
			    *newQ2, q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "child 1 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*newQ2,
			    *newQ3, *qG, *qJ, *qM,
			    *qP, q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "child 2 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*newQ1,
			    *qD, *qH, *newQ3, *qN,
			    *qO, q.attrib_);
      DEBUGL(HexSubdiv4xcreateChildren_D, "child 3 = " << **hex);
    }
  }

  void HexSubdiv4x::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      Edge* newE = q.child(2)->edge(0);
      delete q.child(0)->quad_[3];
      delete q.child(0)->quad_[5];
      delete q.child(2)->quad_[0];
      delete q.child(2)->quad_[1];
      delete newE;
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv4x* HexSubdiv4x::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv4xInst_D, "creating new instance");
      instance_.reset(new HexSubdiv4x());
    }
    return instance_.get();
  }

  HexSubdiv4x::~HexSubdiv4x() {
    DEBUGL(HexSubdiv4xDestr_D, "done.");
  }

  std::ostream& HexSubdiv4x::info(std::ostream& os) const {
    return os << "HexSubdiv4x()";
  }

  // *********************************************************** HexSubdiv4y **

  std::auto_ptr<HexSubdiv4y> HexSubdiv4y::instance_(0);

  void HexSubdiv4y::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv4ycreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 0, 2, 4 and 5
      uint I[4] = { 0, 2, 4, 5 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
      }

      // try to subdivide the other two faces into 4
      bool quad1Normal = faceSubdiv4(*(q.quad_[1]));
      bool quad3Normal = faceSubdiv4(*(q.quad_[3]));

      // create new edge
      Vertex* vtx1 = 0, *vtx2 = 0;
      if (quad1Normal)
	vtx1 = q.quad_[1]->child(0)->vertex(2);
      else {
	vtx1 = q.quad_[1]->child(0)->child(0)->vertex(2);
      }
      if (quad3Normal)
	vtx2 = q.quad_[3]->child(0)->vertex(2);
      else {
	vtx2 = q.quad_[3]->child(0)->child(0)->vertex(2);
      }
      Edge* newE = new Edge(*vtx1, *vtx2, q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "new edge " << *newE);

      // find subdivided faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0, *qG = 0,
	*qH = 0, *qI = 0, *qJ = 0, *qK = 0, *qL = 0, *qM = 0, *qN = 0, *qO,
	*qP = 0;

      subFace0011(*q.quad_[0], q.rho_[0], q.tau_[0], qA, qB);
      subdivFace1(q, qC, qD, qE, qF, quad1Normal);
      subFace1100(*q.quad_[2], q.rho_[2], q.tau_[2], qG, qH);
      subdivFace3(q, qJ, qI, qL, qK, quad3Normal);
      subFace1100(*q.quad_[4], q.rho_[4], q.tau_[4], qM, qN);
      subFace1100(*q.quad_[5], q.rho_[5], q.tau_[5], qO, qP);

      // create new quads
      Quad* newQ0 = new Quad(*(q.quad_[0]->child(0)->edge(1+(q.tau_[0]%2))),
			     *(qI->edge(q.tau_[3])),
			     *newE,
			     *(qC->edge((q.tau_[1]+2)%4)), q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "new quad 0 " << *newQ0);

      Quad* newQ1 = new Quad(*(q.quad_[4]->child(0)->edge(1+(q.tau_[4]%2))),
			     *(qC->edge( (q.tau_[1]+(q.rho(1)?1:3))%4 )),
			     *newE,
			     *(qI->edge( (q.tau_[3]+(q.rho(3)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "new quad 1 " << *newQ1);

      Quad* newQ2 = new Quad(*newE,
			     *(qD->edge( (q.tau_[1]+(q.rho(1)?1:3))%4 )),
			     *(q.quad_[2]->child(0)->edge(1+(q.tau_[2]%2))),
			     *(qJ->edge( (q.tau_[3]+(q.rho(3)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "new quad 2 " << *newQ2);

      Quad* newQ3 = new Quad(*(qE->edge(q.tau_[1])),
			     *newE,
			     *(qL->edge(q.tau_[3])),
			     *(q.quad_[5]->child(0)->edge(1+(q.tau_[5]%2))),
			     q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "new quad 3 " << *newQ3);

      // create four new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*qA,
			    *qC, *newQ0, *qI, *qM,
			    *newQ1, q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qB,
			    *qD, *qG, *qJ, *newQ0,
			    *newQ2, q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "child 1 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*newQ2,
			    *qE, *qH, *qK, *newQ3,
			    *qP, q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "child 2 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*newQ1,
			    *qF, *newQ3, *qL, *qN,
			    *qO, q.attrib_);
      DEBUGL(HexSubdiv4ycreateChildren_D, "child 3 = " << **hex);
    }
  }

  void HexSubdiv4y::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      Edge* newE = q.child(3)->edge(1);
      delete q.child(0)->quad_[2];
      delete q.child(0)->quad_[5];
      delete q.child(2)->quad_[0];
      delete q.child(2)->quad_[4];
      delete newE;
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv4y* HexSubdiv4y::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv4yInst_D, "creating new instance");
      instance_.reset(new HexSubdiv4y());
    }
    return instance_.get();
  }

  HexSubdiv4y::~HexSubdiv4y() {
    DEBUGL(HexSubdiv4yDestr_D, "done.");
  }

  std::ostream& HexSubdiv4y::info(std::ostream& os) const {
    return os << "HexSubdiv4y()";
  }

  // *********************************************************** HexSubdiv4z **

  std::auto_ptr<HexSubdiv4z> HexSubdiv4z::instance_(0);

  void HexSubdiv4z::createChildren(Hexahedron& q) const {
    if (q.chld_ == 0) {
      DEBUGL(HexSubdiv4zcreateChildren_D, "subdividing " << q);
      // set subdivision strategy on the quads 0, 2, 4 and 5
      uint I[4] = { 1, 2, 3, 4 };
      for (uint i = 0; i < 4; ++i) {
	if (q.tau_[I[i]]%2)
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2V::instance());
	else
	  q.quad_[I[i]]->setStrategy(QuadSubdiv2H::instance());
      }

      // try to subdivide the other two faces into 4
      bool quad0Normal = faceSubdiv4(*(q.quad_[0]));
      bool quad5Normal = faceSubdiv4(*(q.quad_[5]));

      // create new edge
      Vertex* vtx1 = 0, *vtx2 = 0;
      if (quad0Normal)
	vtx1 = q.quad_[0]->child(0)->vertex(2);
      else {
	vtx1 = q.quad_[0]->child(0)->child(0)->vertex(2);
      }
      if (quad5Normal)
	vtx2 = q.quad_[5]->child(0)->vertex(2);
      else {
	vtx2 = q.quad_[5]->child(0)->child(0)->vertex(2);
      }
      Edge* newE = new Edge(*vtx1, *vtx2, q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "new edge " << *newE);

      // find subdivided faces
      Quad *qA = 0, *qB = 0, *qC = 0, *qD = 0, *qE = 0, *qF = 0, *qG = 0,
	*qH = 0, *qI = 0, *qJ = 0, *qK = 0, *qL = 0, *qM = 0, *qN = 0, *qO,
	*qP = 0;

      subdivFace0(q, qA, qB, qC, qD, quad0Normal);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qA = " << *qA);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qB = " << *qB);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qC = " << *qC);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qD = " << *qD);

      subFace0110(*q.quad_[1], q.rho_[1], q.tau_[1], qE, qF);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qE = " << *qE);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qF = " << *qF);

      subFace0110(*q.quad_[2], q.rho_[2], q.tau_[2], qG, qH);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qG = " << *qG);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qH = " << *qH);

      subFace1001(*q.quad_[3], q.rho_[3], q.tau_[3], qI, qJ);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qI = " << *qI);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qJ = " << *qJ);

      subFace1001(*q.quad_[4], q.rho_[4], q.tau_[4], qK, qL);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qK = " << *qK);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qL = " << *qL);

      if (quad5Normal) {
	// here, P and N are swapped because of a different sequence
	// subdivFace5 cannot be used!
	subFace1230(*q.quad_[5], q.rho_[5], q.tau_[5], qM, qP, qO, qN);
      } else {
	Quad* quads[4];
	quads[0] = q.quad_[5]->child(0)->child(0);
	quads[1] = q.quad_[5]->child(0)->child(1);
	quads[2] = q.quad_[5]->child(1)->child(0);
	quads[3] = q.quad_[5]->child(1)->child(1);
	for (uint i = 0; i < 4; ++i)
	  conceptsAssert(quads[i], Assertion());

	const uint tau = q.tau(5), rho = q.rho(5);
	qM = findFace(q.vertex(4), quads, tau, rho, 0);
	qN = findFace(q.vertex(5), quads, tau, rho, 3);
	qO = findFace(q.vertex(6), quads, tau, rho, 2);
	qP = findFace(q.vertex(7), quads, tau, rho, 1);
      }
      DEBUGL(HexSubdiv4zcreateChildren_D, "qM = " << *qM);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qN = " << *qN);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qO = " << *qO);
      DEBUGL(HexSubdiv4zcreateChildren_D, "qP = " << *qP);

      // create new quads
      Quad* newQ0 = new Quad(*(q.quad_[1]->child(0)->edge(2-(q.tau_[1]%2))),
			     *(qB->edge( (q.tau_[0]+(q.rho_[0]?1:3))%4 )),
			     *newE,
			     *(qM->edge( (q.tau_[5]+(q.rho(5)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "new quad 0 " << *newQ0);

      Quad* newQ1 = new Quad(*(qC->edge(q.tau_[0])),
			     *(q.quad_[2]->child(0)->edge(2-(q.tau_[2]%2))),
			     *(qO->edge(q.tau_[5])),
			     *newE,
			     q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "new quad 1 " << *newQ1);

      Quad* newQ2 = new Quad(*newE,
			     *(qC->edge( (q.tau_[0]+(q.rho_[0]?1:3))%4 )),
			     *(q.quad_[3]->child(0)->edge(2-(q.tau_[3]%2))),
			     *(qP->edge( (q.tau_[5]+(q.rho(5)?1:3))%4 )),
			     q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "new quad 2 " << *newQ2);

      Quad* newQ3 = new Quad(*(qP->edge(q.tau_[5])),
			     *newE,
			     *(qD->edge(q.tau_[0])),
			     *(q.quad_[4]->child(0)->edge(2-(q.tau_[4]%2))),
			     q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "new quad 3 " << *newQ3);

      // create four new hexahedra
      Hexahedron** hex = &(q.chld_);

      *hex = new Hexahedron(*qA,
			    *qE, *newQ0, *newQ3, *qK,
			    *qM, q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "child 0 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qB,
			    *qF, *qG, *newQ1, *newQ0,
			    *qN, q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "child 1 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qC,
			    *newQ1, *qH, *qJ, *newQ2,
			    *qO, q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "child 2 = " << **hex);
      hex = &(*hex)->lnk_;

      *hex = new Hexahedron(*qD,
			    *newQ3, *newQ2, *qI, *qL,
			    *qP, q.attrib_);
      DEBUGL(HexSubdiv4zcreateChildren_D, "child 3 = " << **hex);
    }
  }

  void HexSubdiv4z::removeChildren(Hexahedron& q) const {
    if (q.chld_ != 0) {
      Edge* newE = q.child(0)->edge(6);
      delete q.child(0)->quad_[2];
      delete q.child(0)->quad_[3];
      delete q.child(2)->quad_[1];
      delete q.child(2)->quad_[4];
      delete newE;
      for(Hexahedron* hex = q.chld_;
	  hex != 0;
	  q.chld_ = hex, hex = q.chld_->lnk_, delete q.chld_);
    }
  }

  const HexSubdiv4z* HexSubdiv4z::instance() {
    if (instance_.get() == 0) {
      DEBUGL(HexSubdiv4zInst_D, "creating new instance");
      instance_.reset(new HexSubdiv4z());
    }
    return instance_.get();
  }

  HexSubdiv4z::~HexSubdiv4z() {
    DEBUGL(HexSubdiv4zDestr_D, "done.");
  }

  std::ostream& HexSubdiv4z::info(std::ostream& os) const {
    return os << "HexSubdiv4z()";
  }

} // namespace concepts
