/* Topology.
 * how to subdivide a quad
 */

#include "topology.hh"
#include "quadsubdiv.hh"
#include "basics/debug.hh"

// debugging QuadSubdiv4
#define QuadSubdiv4Constr_D 0
#define QuadSubdiv4Destr_D 0
#define QuadSubdiv4createChildren_D 0 
#define QuadSubdiv4removeChildren_D 0

// debugging QuadSubdiv2H
#define QuadSubdiv2HDestr_D 0
#define QuadSubdiv2HcreateChildren_D 0
#define QuadSubdiv2HremoveChildren_D 0 

// debugging QuadSubdiv2V
#define QuadSubdiv2VDestr_D 0
#define QuadSubdiv2VcreateChildren_D 0
#define QuadSubdiv2VremoveChildren_D 0

namespace concepts {

  // *********************************************************** QuadSubdiv4 **

  std::auto_ptr<QuadSubdiv4> QuadSubdiv4::instance_(0);

  void QuadSubdiv4::createChildren(Quad& q) const {
    if (q.chld_ == 0) {
      DEBUGL(QuadSubdiv4createChildren_D, "subdiving " << q);
      conceptsAssert(q.getStrategy() == this, Assertion());
      Edge* edgA = q.edg_[0]->child(q.rho_[0]);
      Edge* edgB = q.edg_[0]->child(q.rho_[0].succ());
      Edge* edgC = q.edg_[1]->child(q.rho_[1]);
      Edge* edgD = q.edg_[1]->child(q.rho_[1].succ());
      Edge* edgE = q.edg_[2]->child(q.rho_[2]);
      Edge* edgF = q.edg_[2]->child(q.rho_[2].succ());
      Edge* edgG = q.edg_[3]->child(q.rho_[3]);
      Edge* edgH = q.edg_[3]->child(q.rho_[3].succ());

      Vertex* vtx4 = edgB->vertex(q.rho_[0]);
      Vertex* vtx5 = edgD->vertex(q.rho_[1]);
      Vertex* vtx6 = edgF->vertex(q.rho_[2]);
      Vertex* vtx7 = edgH->vertex(q.rho_[3]);

      Vertex* vtx8 = new Vertex(q.attrib_);

      Edge* edgI = new Edge(*vtx4, *vtx8, q.attrib_);
      Edge* edgJ = new Edge(*vtx5, *vtx8, q.attrib_);
      Edge* edgK = new Edge(*vtx6, *vtx8, q.attrib_);
      Edge* edgL = new Edge(*vtx7, *vtx8, q.attrib_);

      Quad* qd = 0;

      qd = new Quad(*edgA, *edgI, *edgL, *edgH, q.attrib_);
      DEBUGL(QuadSubdiv4createChildren_D, "child 0 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);

      qd = new Quad(*edgB, *edgC, *edgJ, *edgI, q.attrib_);
      DEBUGL(QuadSubdiv4createChildren_D, "child 1 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);

      qd = new Quad(*edgJ, *edgD, *edgE, *edgK, q.attrib_);
      DEBUGL(QuadSubdiv4createChildren_D, "child 2 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);

      qd = new Quad(*edgL, *edgK, *edgF, *edgG, q.attrib_);
      DEBUGL(QuadSubdiv4createChildren_D, "child 3 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);
    }
  }

  void QuadSubdiv4::removeChildren(Quad& q) const {
    if (q.chld_ != 0) {
      try {
	q.child(0);
      }
      catch (concepts::StrategyChange& e) {
	DEBUGL(QuadSubdiv4removeChildren_D, "removes 4 children of " << q);
	conceptsAssert(q.getStrategy() == this, Assertion());


	delete newVertex(q, 0);
	for (uint i = 0; i < 4; ++i)
	  delete newEdge(q, i);

	concepts::Joiner<Quad*, 1>::destructor(q.chld_);
      }
    }
  }

  Edge* QuadSubdiv4::newEdge(const Quad& q, uint i) const {
    conceptsAssert(q.getStrategy() == this, Assertion());
    if (q.chld_)
      switch(i) {
      case 0: return q.child(0)->edge(1);
      case 1: return q.child(2)->edge(0);
      case 2: return q.child(2)->edge(3);
      case 3: return q.child(0)->edge(2);
      default: return 0;
      }
    else return 0;
  }

  Vertex* QuadSubdiv4::newVertex(const Quad& q, uint i) const {
    conceptsAssert(q.getStrategy() == this, Assertion());
    if (q.chld_ && (i == 0)) return q.child(0)->vertex(2);
    else return 0;
  }

  const QuadSubdiv4* QuadSubdiv4::instance() {
    if (instance_.get() == 0)
      instance_.reset(new QuadSubdiv4());
    return instance_.get();
  }

  QuadSubdiv4::~QuadSubdiv4() {
    DEBUGL(QuadSubdiv4Destr_D, "done.");
  }

  std::ostream& QuadSubdiv4::info(std::ostream& os) const {
    return os << "QuadSubdiv4()";
  }

  // ********************************************************** QuadSubdiv2H **

  std::auto_ptr<QuadSubdiv2H> QuadSubdiv2H::instance_(0);

  void QuadSubdiv2H::createChildren(Quad& q) const {
    if (q.chld_ == 0) {
      conceptsAssert(q.getStrategy() == this, Assertion());
      DEBUGL(QuadSubdiv2HcreateChildren_D, "subdiving " << q);
      // create things
      Edge* edgA = q.edg_[0];
      Edge* edgB = q.edg_[1]->child(q.rho_[1]);
      Edge* edgD = q.edg_[3]->child(q.rho_[3].succ());
      Edge* edgC = new Edge(*(edgB->vertex(q.rho_[1].succ())),
			    *(edgD->vertex(q.rho_[3])), q.attrib_);
      Edge* edgE = q.edg_[1]->child(q.rho_[1].succ());
      Edge* edgF = q.edg_[2];
      Edge* edgG = q.edg_[3]->child(q.rho_[3]);

      Quad* qd = 0;

      qd = new Quad(*edgA, *edgB, *edgC, *edgD, q.attrib_);
      DEBUGL(QuadSubdiv2HcreateChildren_D, "child 0 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);

      qd = new Quad(*edgC, *edgE, *edgF, *edgG, q.attrib_);
      DEBUGL(QuadSubdiv2HcreateChildren_D, "child 1 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);
    }
  }

  void QuadSubdiv2H::removeChildren(Quad& q) const {
    if (q.chld_ != 0) {
      try {
	q.child(0);
      }
      catch (concepts::StrategyChange& e) {
	DEBUGL(QuadSubdiv2HremoveChildren_D, "removes 2H children of " << q);
	conceptsAssert(q.getStrategy() == this, Assertion());
	// delete edge in the middle
	delete q.child(0)->edg_[2];

	concepts::Joiner<Quad*, 1>::destructor(q.chld_);
      }
    }
  }

  Edge* QuadSubdiv2H::newEdge(const Quad& q, uint i) const {
    conceptsAssert(q.getStrategy() == this, Assertion());
    if (i == 0) return q.child(0)->edge(2);
    else return 0;
  }

  Vertex* QuadSubdiv2H::newVertex(const Quad&, uint) const {
    return 0;
  }

  const QuadSubdiv2H* QuadSubdiv2H::instance() {
    if (instance_.get() == 0)
      instance_.reset(new QuadSubdiv2H());
    return instance_.get();
  }

  QuadSubdiv2H::~QuadSubdiv2H() {
    DEBUGL(QuadSubdiv2HDestr_D, "done.");
  }

  std::ostream& QuadSubdiv2H::info(std::ostream& os) const {
    return os << "QuadSubdiv2H()";
  }

  // ********************************************************** QuadSubdiv2V **

  std::auto_ptr<QuadSubdiv2V> QuadSubdiv2V::instance_(0);

  void QuadSubdiv2V::createChildren(Quad& q) const {
    if (q.chld_ == 0) {
      DEBUGL(QuadSubdiv2VcreateChildren_D, "subdiving " << q);
      conceptsAssert(q.getStrategy() == this, Assertion());
      // create things
      Edge* edgA = q.edg_[0]->child(q.rho_[0]);
      Edge* edgC = q.edg_[2]->child(q.rho_[2].succ());
      Edge* edgD = q.edg_[3];
      Edge* edgB = new Edge(*(edgA->vertex(q.rho_[0].succ())),
			    *(edgC->vertex(q.rho_[2])), q.attrib_);
      Edge* edgE = q.edg_[0]->child(q.rho_[0].succ());
      Edge* edgF = q.edg_[1];
      Edge* edgG = q.edg_[2]->child(q.rho_[2]);

      Quad* qd = 0;

      qd = new Quad(*edgA, *edgB, *edgC, *edgD, q.attrib_);
      DEBUGL(QuadSubdiv2VcreateChildren_D, "child 0 = " << *qd);
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);

      qd = new Quad(*edgE, *edgF, *edgG, *edgB, q.attrib_);
      DEBUGL(QuadSubdiv2VcreateChildren_D, "child 1 = " << *qd)
      q.chld_ = new concepts::Joiner<Quad*, 1>(qd, q.chld_);
    }
  }

  void QuadSubdiv2V::removeChildren(Quad& q) const {
    if (q.chld_ != 0) {
      try {
	q.child(0);
      }
      catch (concepts::StrategyChange& e) {
	DEBUGL(QuadSubdiv2VremoveChildren_D, "removes 2V children of " << q);
	conceptsAssert(q.getStrategy() == this, Assertion());
	// delete edge in the middle
	delete q.child(0)->edg_[1];
	
	concepts::Joiner<Quad*, 1>::destructor(q.chld_);
      }
    }
  }

  Edge* QuadSubdiv2V::newEdge(const Quad& q, uint i) const {
    conceptsAssert(q.getStrategy() == this, Assertion());
    if (i == 0) return q.child(0)->edge(1);
    else return 0;
  }

  Vertex* QuadSubdiv2V::newVertex(const Quad&, uint) const {
    return 0;
  }

  const QuadSubdiv2V* QuadSubdiv2V::instance() {
    if (instance_.get() == 0)
      instance_.reset(new QuadSubdiv2V());
    return instance_.get();
  }

  QuadSubdiv2V::~QuadSubdiv2V() {
    DEBUGL(QuadSubdiv2VDestr_D, "done.");
  }

  std::ostream& QuadSubdiv2V::info(std::ostream& os) const {
    return os << "QuadSubdiv2V()";
  }

} // namespace concepts
