/* testsuite for 2D mesh classes in geometry
 */

#include <iostream>
#include "basics.hh"
#include "geometry.hh"

using namespace concepts;

class testMesh : public Mesh2 {
private:
  class S : public Scan<Cell2> {
    unsigned int idx_;
    Cell2 *(&cell_)[1];
  public:
    inline S(Cell2 *(&cell)[1]) : idx_(0), cell_(cell) {}
    inline S(const S& scan) : idx_(scan.idx_),  cell_(scan.cell_) {}
    inline bool eos() const { return idx_ == 1; }
    inline Cell2& operator++(int) { return *cell_[idx_++]; }
    inline Scan2* clone() const { return new S(*this); }
  };

  Vertex *vtx_[3];
  Edge *edg_[3];
  Triangle *tri_[1];
  Cell2 *cell_[1];

public:
  testMesh(const unsigned int r = 2);
  virtual ~testMesh();
  inline unsigned int ncell() const { return 1; }
  inline Scan2* scan() { return new S(cell_); }
  virtual std::ostream& info(std::ostream& os) const;
};

testMesh::testMesh(const unsigned int r) {

  for (int i = 0; i < 3; ++i)
    vtx_[i] = new Vertex(1);

  edg_[0] = new Edge(*vtx_[0], *vtx_[1], 1);
  edg_[1] = new Edge(*vtx_[1], *vtx_[2], 2);
  edg_[2] = new Edge(*vtx_[2], *vtx_[0], 3);

  tri_[0] = new Triangle(*edg_[0], *edg_[1], *edg_[2]);
  VertexTriangle2d testMap(Real2d(0.0, 0.0), Real2d(1.0, 0.0), 
			   Real2d(0.0, 1.0));
  cell_[0] = new Triangle2d(*tri_[0], testMap);
}

testMesh::~testMesh() {
  delete cell_[0];
  delete tri_[0];
  for (int i = 0; i < 3; ++i)
    delete edg_[i];
  for (int i = 0; i < 3; ++i)
    delete vtx_[i];
}

std::ostream& testMesh::info(std::ostream& os) const {
  os << "testMesh(" << *cell_[0] << ')';
  return os;
}

int main(int, char**) {

  Vertex c(1);
  Vertex d(2);
  Edge e(c, d);
  Vertex f(3);
  Edge g(d, f);
  Edge h(f, c);
  Triangle i(e, g, h);

  std::cout << "Triangle: " << i << std::endl;

  VertexTriangle2d j(Real2d(1.0, 1.0), Real2d(2.0, 2.0), Real2d(0.0, 1.5));
  Triangle2d k(i, j);

  std::cout << "Triangle2d: " << k << std::endl;

  for (int n = 0; n < 4; ++n)
    std::cout << "Child: " << *(k.child(n)) << std::endl;

  testMesh t;
  std::cout << "Mesh: " << t << std::endl;

  return 0;
}
