/* testsuite for 3D cells
 */

#include <memory>
#include "basics/exceptions.hh"
#include "geometry.hh"

int main(int, char**) {

  using namespace concepts;

  const uint ncell_ = 1;
  std::auto_ptr<Vertex> vtx_[8];
  std::auto_ptr<Edge> edg_[12];
  std::auto_ptr<Quad> quad_[6];
  std::auto_ptr<Hexahedron> hex_[ncell_];
  std::auto_ptr<Hexahedron3d> cell_[ncell_];

  try {

    for (uint i = 0; i < 8; ++i)
      vtx_[i].reset(new Vertex);

    edg_[0].reset(new Edge(*vtx_[0], *vtx_[1]));
    edg_[1].reset(new Edge(*vtx_[1], *vtx_[2]));
    edg_[2].reset(new Edge(*vtx_[2], *vtx_[3]));
    edg_[3].reset(new Edge(*vtx_[0], *vtx_[3]));
    edg_[4].reset(new Edge(*vtx_[0], *vtx_[4]));
    edg_[5].reset(new Edge(*vtx_[1], *vtx_[5]));
    edg_[6].reset(new Edge(*vtx_[2], *vtx_[6]));
    edg_[7].reset(new Edge(*vtx_[3], *vtx_[7]));
    edg_[8].reset(new Edge(*vtx_[4], *vtx_[5]));
    edg_[9].reset(new Edge(*vtx_[5], *vtx_[6]));
    edg_[10].reset(new Edge(*vtx_[6], *vtx_[7]));
    edg_[11].reset(new Edge(*vtx_[7], *vtx_[4]));

    quad_[0].reset(new Quad(*edg_[0], *edg_[1], *edg_[2], *edg_[3], 1));
    quad_[1].reset(new Quad(*edg_[0], *edg_[5], *edg_[8], *edg_[4], 1));
    quad_[2].reset(new Quad(*edg_[1], *edg_[6], *edg_[9], *edg_[5], 2));
    quad_[3].reset(new Quad(*edg_[2], *edg_[6], *edg_[10], *edg_[7], 3));
    quad_[4].reset(new Quad(*edg_[3], *edg_[7], *edg_[11], *edg_[4], 1));
    quad_[5].reset(new Quad(*edg_[8], *edg_[9], *edg_[10], *edg_[11], 4));

    hex_[0].reset(new Hexahedron(*quad_[0], 
				 *quad_[1], *quad_[2], *quad_[3], *quad_[4], 
				 *quad_[5]));
    cell_[0].reset(new Hexahedron3d (*hex_[0], 
				     MapHexahedron3d(Real3d(0.0, 0.0, 0.0),
						     Real3d(1.0, 0.0, 0.0),
						     Real3d(1.0, 1.0, 0.0),
						     Real3d(0.0, 1.0, 0.0),
						     Real3d(0.0, 0.0, 1.0),
						     Real3d(1.0, 0.0, 1.0),
						     Real3d(1.0, 1.0, 1.0),
						     Real3d(0.0, 1.0, 1.0))));

    // start tests here
    std::cout << "hex = " << *(hex_[0]) << std::endl;
    hex_[0]->setStrategy(HexSubdiv4z::instance());
    Hexahedron* hex = 0;
    for (uint i = 0; (hex = hex_[0]->child(i)) != 0; ++i)
      std::cout << "child " << i << " = " << *hex << std::endl;

    concepts::Stacktrace::doit() = false;

    std::cout << "Triggering 'illegal change of strategy':" << std::endl;
    hex_[0]->setStrategy(HexSubdiv8::instance());
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }
  
  return 0;
}
