/* graphics of the mesh in DAT
 */

#ifndef graphMeshDAT_hh
#define graphMeshDAT_hh

#include <string>
#include <fstream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "basics/exceptions.hh"
#include "toolbox/dynArray.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declartions
  class Mesh2;
  class Mesh3;
} // namespace concepts

namespace graphics {

  using concepts::Real;

  // *************************************************************** MeshDAT **

  /** Creates files which are suitable for mesh input using \c
      Import2dMesh or \c Import3dMesh.
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  class MeshDAT {
  public:
    /// Constructor
    MeshDAT(concepts::Mesh& msh, const std::string& filename);
    /// Constructor
    MeshDAT(concepts::Space<F>& spc, const std::string& filename);
  private:
    std::string coordFilename_, boundFilename_, elmFilename_;
    void createFilenames_(const std::string& filename);
  };

  /** Trampoline function to create an instance of MeshDAT.
      @author Philipp Frauenfelder, 2003
  */
  void drawMeshDAT(concepts::Mesh& msh, const std::string& filename);

  /** Trampoline function to create an instance of MeshDAT.
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  void drawMeshDAT(concepts::Space<F>& spc, const std::string& filename);

  // *********************************************************** MeshDATCell **

  /** Cell by cell output for \c MeshDAT
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  class MeshDATCell : public concepts::CellPostprocess<F> {
  public:
    /// Constructor.
    MeshDATCell(const std::string& coordFilename,
		const std::string& boundFilename,
		const std::string& elmFilename);
    virtual ~MeshDATCell();
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature);
  private:
    std::ofstream coordFile_, boundFile_, elmFile_;

    /// Vertex lists
    std::map<uint, uint> vtxList_;
    std::map<uint, concepts::Real3d> vtxCoord_;
    concepts::DynArray<bool> vtxBndNotDone_, edgBndNotDone_;

    /// Counters
    uint vtxCnt_, bndCnt_, elmCnt_;

    uint getVertexNumber_(const uint key, const concepts::Real3d* coord = 0);
  };

} // namespace graphics

#endif // graphMeshDAT_hh
