/* graphics of the mesh in Tecplot
 */

#ifndef graphMeshTecplot_hh
#define graphMeshTecplot_hh

#include <string>
#include <fstream>
#include <memory>
#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "basics/exceptions.hh"
#include "toolbox/dynArray.hh"
#include "space/postProcess.hh"

namespace concepts {
  // forward declartions
  class Mesh2;
  class Mesh3;
} // namespace concepts

namespace graphics {

  // forward declartions
  class VertexList;

  using concepts::Real;

  // *********************************************************** MeshTecplot **

  /** Creates files to show the mesh using Tecplot. The default
      extension for Tecplot ASCII files is '.dat'.
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  class MeshTecplot {
  public:
    /// Constructor
    MeshTecplot(concepts::Mesh& msh, const std::string& filename);
    /// Constructor
    MeshTecplot(concepts::Space<F>& spc, const std::string& filename,
		const uint dim);
  private:
    const std::string filename_;
  };

  /** Trampoline function to create an instance of MeshTecplot.
      @author Philipp Frauenfelder, 2003
  */
  void drawMeshTecplot(concepts::Mesh& msh, const std::string& filename);

  /** Trampoline function to create an instance of MeshTecplot.
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  void drawMeshTecplot(concepts::Space<F>& spc, const std::string& filename,
		       const uint dim);

  // ******************************************************* MeshTecplotCell **

  /** Cell by cell output for \c MeshTecplot
      @author Philipp Frauenfelder, 2003
  */
  template<typename F>
  class MeshTecplotCell : public concepts::CellPostprocess<F> {
  public:
    /// Constructor.
    MeshTecplotCell(const std::string& filename, VertexList& vtxList,
		    const uint nelm, const uint dim);
    virtual ~MeshTecplotCell();
    virtual void operator() (const concepts::Element<F>& elm)
      throw(concepts::MissingFeature);
    virtual void operator() (const concepts::Cell& cell)
      throw(concepts::MissingFeature);
  private:
    /// Stream to file
    std::ofstream file_;

    /// Vertex list
    VertexList& vtxList_;
  };

} // namespace graphics

#endif // graphMeshTecplot_hh
