/* Bilinear forms for hp 1D FEM
 */

#ifndef bilinearform1D_hh
#define bilinearform1D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatricesForward.hh"
#include "operator/bilinearForm.hh"

namespace concepts {
  // forward declarations
  template<class F>
  class Element;

  template<class F>
  class ElementMatrix;

  template<class F>
  class Array;
}

namespace hp1D {
  using concepts::Real;

  // *************************************************************** Laplace **

  /** A function class to calculate element matrices for the Laplacian.
      @author Philipp Frauenfelder, 2002
  */
  class Laplace : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    Laplace();
    virtual ~Laplace();

    void operator()(const concepts::Element<Real>& elmX,
		    const concepts::Element<Real>& elmY, 
		    concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
  };

  // ************************************************************** Identity **

  /** A function class to calculate element matrices for the mass matrix.
      @author Philipp Frauenfelder, 2002
  */
  class Identity : public concepts::BilinearForm<Real> {
  public:
    /// Constructor
    Identity();
    virtual ~Identity();

    void operator()(const concepts::Element<Real>& elmX,
		    const concepts::Element<Real>& elmY, 
		    concepts::ElementMatrix<Real>& em);
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
  };

} // namespace hp1D

#endif // bilinearform1D_hh
