/* abstract base class for hp elements in 1D
 */

#include "element.hh"
#include "basics/debug.hh"
#include "basics/exceptions.hh"

#define CONST_POINTS 0

#if CONST_POINTS
#define NUM_POINTS 10
#else
#define NUM_POINTS_ADD 2
#endif

// debugging Element
#define ElementConstr_D 0
#define ElementShpfctIndices_D 0

namespace hp1D {

  // *************************************************************** Element **

  template<class F>
  std::auto_ptr<LineGraphics> Element<F>::graphics_(0);

  template<class F>
  Element<F>::Element(concepts::Edge1d& cell, uint p,
		      concepts::TColumn<Real>* T0,
		      concepts::TColumn<Real>* T1) :
    T_(T1), cell_(cell), shpfct_(0), shpfctD_(0), int_(0), p_(p) {
    if (T0)
      T_.append(T0);
    DEBUGL(ElementConstr_D, cell << ", p = " << p << ", T = " << T_);

    int_.reset(new concepts::Quadrature<INTEGRATION_RULE>(
#if CONST_POINTS
							  NUM_POINTS
#else
							  p+NUM_POINTS_ADD
#endif
							  ));
    // shape functions
    shpfct_.reset
      (new concepts::Karniadakis<1,0>(p, int_->abscissas(), int_->n()));
    // derivatives of shape functions
    shpfctD_.reset
      (new concepts::Karniadakis<1,1>(p, int_->abscissas(), int_->n()));
  }

  template<class F>
  concepts::Real3d Element<F>::vertex(uint i) const {
    return cell_.vertex(i);
  }

  template<class F>
  const concepts::ElementGraphics<Real>* Element<F>::graphics() const {
    if (! graphics_.get())
      graphics_.reset(new LineGraphics());
    return graphics_.get();
  }

  template<class F>
  std::ostream& Element<F>::info(std::ostream& os) const {
    return os << "Element(" << cell_ << ", " << T_ << ")";
  }

  template class Element<Real>;

} // namespace hp1D
