/* makes a function (formula) from a weight (can then be plotted)
 */

#ifndef formulaFromWeight2D_hh
#define formulaFromWeight2D_hh

#include <typeinfo>
#include "space/element.hh"
#include "shortestDist.hh"
#include "formulaFromWeight.hh"
#include "quad.hh"
#include "basics/exceptions.hh"

// debugging
#include "basics/debug.hh"

#define FormulaFromWeightConstr_D 0
#define FormulaFromWeightAppl_D 0

namespace hp2D {

  // ***************************************************** FormulaFromWeight **

  /** Makes it possible to plot a given \c Weight.
      @author Philipp Frauenfelder, 2003
  */
  template<typename Weight = TrivialWeight>
  class FormulaFromWeight : public concepts::ElementFormula<Real> {
  public:
    /** Constructor.
	@param singularities Set of singularly marked edges and corners
    */
    FormulaFromWeight(const SingularSet& singularities);

    virtual Real operator()(const concepts::Element<Real>& elm, const Real p,
			    const Real t = 0.0) const;
    virtual Real operator()(const concepts::Element<Real>& elm,
			    const concepts::Real2d& p,
			    const Real t = 0.0) const;
    virtual Real operator()(const concepts::Element<Real>& elm,
			    const concepts::Real3d& p,
			    const Real t = 0.0) const;
  private:
    Weight weight_;
    const SingularSet& singularities_;
  };

  template<typename Weight>
  FormulaFromWeight<Weight>::FormulaFromWeight
  (const SingularSet& singularities) :
    weight_(), singularities_(singularities) {
    DEBUGL(FormulaFromWeightConstr_D, "weight is " << typeid(weight_).name());
    DEBUGL(FormulaFromWeightConstr_D, "singularities: " << singularities_);
  }

  template<typename Weight>
  Real FormulaFromWeight<Weight>::operator()
    (const concepts::Element<Real>& elm, const Real p, const Real t) const {
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    conceptsAssert(quad, concepts::Assertion());
    return weight_(singularities_, concepts::Real2d(p, 0), quad);
  }

  template<typename Weight>
  Real FormulaFromWeight<Weight>::operator()
    (const concepts::Element<Real>& elm, const concepts::Real2d& p,
     const Real t) const {
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    conceptsAssert(quad, concepts::Assertion());
    DEBUGL(FormulaFromWeightAppl_D, "p = " << p << ", singularities = "
	   << singularities_);
    return weight_(singularities_, p, quad);
  }

  template<typename Weight>
  Real FormulaFromWeight<Weight>::operator()
    (const concepts::Element<Real>& elm, const concepts::Real3d& p,
     const Real t) const {
    const Quad* quad = dynamic_cast<const Quad*>(&elm);
    conceptsAssert(quad, concepts::Assertion());
    return weight_(singularities_, concepts::Real2d(p[0], p[1]), quad);
  }

} // namespace hp2D

#endif // formulaFromWeight2D_hh
