// additional functionality for quadrilaterals

#ifndef quadFunctions_hh
#define quadFunctions_hh

#include "basics/typedefs.hh"

namespace concepts {
  // forward declaration
  class Quad;
}

namespace hp2D {

  // ********************************************************* QuadFunctions **

  /** Auxiliary functions for quadrilaterals.
      @author Philipp Frauenfelder, 2004
  */
  class QuadFunctions {
  public:
    /** Helps distributing the polynomial degree to the edges
	@param i Index of the edge [0..3]
	@param pIndex Index of the polynomial degree
	              of this edge in the array of the element
	@param qIndex Other index
	@return pIndex, qIndex
    */
    static void edgeOrientation(const uint i, uint& pIndex, uint& qIndex);

    /** Computes the tensor product index \c ll for a vertex with index \c i.
	@return ll
    */
    static void vertexIndex(const uint i, uint ll[2]);

    /** Computes the tensor product index \c ll for an edge with index \c i
	and precomputed indices \c qIndex and \c rIndex (use edgeOrientation to
	compute them).
	@return ll
    */
    static void edgeIndex(const uint i, const uint qIndex, uint ll[2]);

    /** Computes the linear index from the tensor product index \c ll using
	the polynomial degrees \c p.
	@return Linear index
    */
    static uint index(const uint ll[2], const ushort p[2]) {
      // copy'n paste from TColumnTensor::operator[]
      return ll[0]+ll[1]*(1+p[0]);
    }
  };

} // namespace hp2D

#endif // quadFunctions_hh
