/* space class for 2D hp FEM
 * enrichElm_
 */

#include "hp2D/space.hh"
#include "basics/debug.hh"

#define hpSpace2Dminimum_D 0
#define hpSpace2Denrich_D 0

namespace hp2D {

  // ***************************************************************** Space **

  void Space::minimumRule_(const concepts::Connector2& cntr,
			   const concepts::Connector1& edge) {
    DEBUGL(hpSpace2Dminimum_D, cntr.key() << " - start " << edge);

    concepts::AdaptiveControlP<2>& ctrl2(ctrl2_[cntr.key()]);

    if (ctrl2.tag_.isMember()) {
      concepts::AdaptiveControlP<1>& ctrl1(ctrl1_[edge.key()]);
      concepts::Connector1* cntr1 = 0;
      for (uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	if ((cntr1 == &edge) || (edge.related(*cntr1) != 0)) {
	  uint p = 0;
	  switch(i) {
	  case 0: case 2: p = ctrl2.p_[0]; break;
	  case 1: case 3: p = ctrl2.p_[1]; break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  if (p < ctrl1.p_[0]) {
	    DEBUGL(hpSpace2Dminimum_D, cntr.key()
		   << " - changing p of " << edge << " from " << ctrl1.p_[0]
		   << " to " << p);
	    ctrl1.p_[0] = p;
	  }
	} // if (edges are related / equal)
      } // for edges
    } // is member
    else { // is not member
      const concepts::Connector2* child = 0;
      for (uint i = 0; (child = cntr.child(i)) != 0; ++i) {
	minimumRule_(*child, edge);
      } // for children
    } // is not member
  }

  void Space::enrichElm_(const concepts::Quad2d& cell, int* p) {
    DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - start.");

    concepts::Quad& cntr = cell.connector();
    concepts::AdaptiveControlP<2>& ctrl2(ctrl2_[cntr.key()]);

    DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - " << cntr << ", "
	   << ctrl2);
    DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - P = (i "
	   << p[0] << ", " << p[1]
	   << ", e0 " << p[2] << ", e1 " << p[3]
	   << ", e2 " << p[4] << ", e3 " << p[5] << ')');

    if (ctrl2.tag_.isMember()) {
      // store given information in the right place
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1(ctrl1_[cntr1->key()]);
	if (p[2+i] > ctrl1.p_[0]) {
	  ctrl1.p_[0] = p[2+i];
	  DEBUGL(hpSpace2Denrich_D, cell.connector().key()
		 << " - updated " << *cntr1 << " to " << ctrl1);
	}
      } // for edges
      DEBUGL(hpSpace2Denrich_D, cell.connector().key()
	     << " - stored information.");
    } else {
      DEBUGL(hpSpace2Denrich_D, cell.connector().key()
	     << " - gathering information.");
      // check if there are degrees of freedom here and store the relevant
      // polynomial degrees
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1(ctrl1_[cntr1->key()]);
	std::map<uint, concepts::EdgeData>::const_iterator j =
	  edgeList_.find(cntr1->key().key());
	if (ctrl1.tag_.isMember() && ctrl1.tag_.isActive() &&
	    (j != edgeList_.end()) && (j->second.hasCell(cell.connector()))) {
	  if (ctrl1.p_[0] > 0) {
	    if (p[2+i] < ctrl1.p_[0]) p[2+i] = ctrl1.p_[0];
	  }
	}
      } // for edges
      DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - P = (i "
	     << p[0] << ", " << p[1]
	     << ", e0 " << p[2] << ", e1 " << p[3]
	     << ", e2 " << p[4] << ", e3 " << p[5] << ')');

      // get subdivision strategy
      enum subdivTypes { NONE = 0, X = 1, Y = 2 };
      uint subdiv = 0;
      const concepts::Quad2dSubdivision* strategy = cell.getStrategy();
      conceptsAssert(strategy != 0, concepts::Assertion());
      if (strategy == concepts::Quad2dSubdiv4::instance()) subdiv = X | Y;
      if (strategy == concepts::Quad2dSubdiv2H::instance()) subdiv = Y;
      if (strategy == concepts::Quad2dSubdiv2V::instance()) subdiv = X;
      conceptsAssert(subdiv != 0, concepts::Assertion());
      DEBUGL(hpSpace2Denrich_D, cell.connector().key()
	     << " - subdiv = " << subdiv);

      // loop over the children of the cell
      const concepts::Quad2d* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
	DEBUGL(hpSpace2Denrich_D, cell.connector().key()
	       << " - child " << i << " = " << *child);
	// initialize data for child
	int P[6];
	for (uint j = 0; j < 6; ++j) P[j] = -1;

	// update data for child from edges
	const concepts::Connector1* cntr1 = 0;
	for(uint e = 0; (cntr1 = cntr.edge(e)) != 0; ++e) {
// 	  DEBUGL(hpSpace2Denrich_D, cell.connector().key()
// 		 << " - edge " << e << " = " << *cntr1);
	  uint idxFrom = 2+e;
	  int idxTo = -1;
	  switch (e) {
	  case 0: if (subdiv & X) idxTo = 2+2; break;
	  case 1: if (subdiv & Y) idxTo = 2+3; break;
	  case 2: if (subdiv & X) idxTo = 2+0; break;
	  case 3: if (subdiv & Y) idxTo = 2+1; break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  } // switch edge
	  if (idxTo != -1) {
	    bool doUpdate = false;
	    const concepts::Connector1* edge_child = 0;
	    // update only oppositely edge if the child cell is adjoint to edge e
	    for(uint k = 0; (edge_child = cntr1->child(k)) != 0; ++k) {
	      if (edge_child->key() == child->connector().edge(e)->key())
		doUpdate = true;
	    }
	    if (doUpdate) {
	      // the oppositely edge in the child should has at least the
	      // polynomial degree of parent cell
	      if (P[idxTo] < p[idxFrom]) P[idxTo] = p[idxFrom];
	    }
	  }
	  // the same edge in the child should has at least the polynomial
	  // degree of parent cell
	  if (P[idxFrom] < p[idxFrom]) P[idxFrom] = p[idxFrom];
	} // for edges
	DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - P = (i "
	       << p[0] << ", " << p[1]
	       << ", e0 " << p[2] << ", e1 " << p[3]
	       << ", e2 " << p[4] << ", e3 " << p[5] << ')');

	// go to child
	enrichElm_(*child, P);
      } // for children
    }

    DEBUGL(hpSpace2Denrich_D, cell.connector().key() << " - done.");
  }

} // namespace hp2D
