/* Functor class for sum factorization in 2D
 */

#ifndef sumfact_hh
#define sumfact_hh

#include "toolbox/array.hh"

namespace concepts {

  // forward declarations
  template<typename T>
  class ElementMatrix;

} // namespace concepts

namespace hp2D {

  using concepts::Real;

  // forward declarations
  template<typename F>
  class ShapeFunction2D;

  // ****************************************************** SumFactorization **

  /** Sum factorization for an element matrix.
      @todo Implement changing order of sum factorization, possibility is there
      @bug If the loops are reordered using \c a and \c b on line 58,
           \c coeff (by use of \c coeffTmp) is not accessed in the correct
	   order in the first loop to fill \c sumfact1_.
      @author Kersten Schmidt, 2003
  */
  class SumFactorization {
  public:
    /** Application operator.
	Adds the contribution of \c coeff to \c em.
	@param em Element matrix
	@param jacobian Jacobian
	@param coeff Coefficients for the integration
	@param u Left shape functions
	@param v Right shape functions
    */
    template<typename T>
    void operator()(concepts::ElementMatrix<Real>& em,
		    const concepts::Array<Real>& jacobian, T* coeff,
		    const ShapeFunction2D<Real>& u,
		    const ShapeFunction2D<Real>& v);
  private:
    /// Intermediate matrix for sum factorization
    concepts::Array<Real> sumfact_;
  };

} // namespace hp2D


#endif // sumfact_hh
