/* space class for 3D hp FEM
 * doCorrectRefinement
 */

#include "hp3D/space.hh"

// debugging
#include "basics/debug.hh"

#define DoCorrectRefinement_D 0

namespace hp3D {

  // ***************************************************************** Space **

  void Space::doCorrectRefinement_(concepts::Hexahedron3d& cell,
				   concepts::Level<3>& L) const {
    const concepts::Level<3>& l(cell.level());
    DEBUGL(DoCorrectRefinement_D, l << " - " << cell.connector());
    concepts::Level<3> levelDiff = L - l;
    if (cell.getStrategy() == 0) { // setting refinement strategy
      conceptsAssert(! (L == concepts::Level<3>(-1)), concepts::Assertion());
      conceptsAssert(! (levelDiff < concepts::Level<3>(0)),
		     concepts::Assertion());
      bool refinementDone = false;
      while (! refinementDone) {
	if (levelDiff > concepts::Level<3>(0)) {
	  enum subdivTypes { NONE = 0, X = 1, Y = 2, Z = 4 };
	  uint subdiv = 0;
	  if (levelDiff.l_[0] > 0) subdiv |= X;
	  if (levelDiff.l_[1] > 0) subdiv |= Y;
	  if (levelDiff.l_[2] > 0) subdiv |= Z;
	  switch (subdiv) {
	  case Z | Y | X:
	    cell.setStrategy(concepts::Hex3dSubdiv8::instance()); break;
	  case X | Y:
	    cell.setStrategy(concepts::Hex3dSubdiv4z::instance()); break;
	  case Y | Z:
	    cell.setStrategy(concepts::Hex3dSubdiv4x::instance()); break;
	  case X | Z:
	    cell.setStrategy(concepts::Hex3dSubdiv4y::instance()); break;
	  case X: cell.setStrategy(concepts::Hex3dSubdiv2x::instance()); break;
	  case Y: cell.setStrategy(concepts::Hex3dSubdiv2y::instance()); break;
	  case Z: cell.setStrategy(concepts::Hex3dSubdiv2z::instance()); break;
	  default: break;
	  }
	}

	// testing, if refinement is allowed
	DEBUGL(DoCorrectRefinement_D, l << " - testing, if refinement for "
	       << levelDiff << " is allowed");
	try {
	  cell.child(0);
	  refinementDone = true;
	  DEBUGL(DoCorrectRefinement_D, l << " - success!");
	}
	catch (concepts::StrategyChange& e) {
	  DEBUGL(DoCorrectRefinement_D, "caught " << e << " in "
		 << cell.connector());
	  conceptsAssert(! (L == concepts::Level<3>(-1)),
			 concepts::Assertion());

	  if (&(e.newStrategy()) != concepts::QuadSubdiv4::instance() &&
	      !(levelDiff == concepts::Level<3>(1))) {
	    // find out on which face it went wrong
	    const concepts::Connector3& cntr = cell.connector();
	    const concepts::Connector2* face = 0;
	    uint faceNr = (uint)-1;
	    for (uint i = 0; (face = cntr.face(i)) != 0; ++i) {
	      if (face == &(e.connector()))
		faceNr = i;
	    }
	    if (faceNr == (uint)-1) {
	      // If we did not yet find the face, it must be a child of
	      // one of the faces
	      for (uint i = 0; (face = cntr.face(i)) != 0; ++i) {
		DEBUGL(DoCorrectRefinement_D, "testing " << *face);
		if (face->related(e.connector()))
		  faceNr = i;
	      }
	    }
	    conceptsAssert(faceNr != (uint)-1, concepts::Assertion());
	    // upgrade subdivision
	    switch (faceNr) {
	    case 0: case 5: {
	      if (levelDiff.l_[0] == 0) ++levelDiff.l_[0];
	      else ++levelDiff.l_[1];
	      break; }
	    case 1: case 3: {
	      if (levelDiff.l_[0] == 0) ++levelDiff.l_[0];
	      else ++levelDiff.l_[2];
	      break; }
	    case 2: case 4: {
	      if (levelDiff.l_[1] == 0) ++levelDiff.l_[1];
	      else ++levelDiff.l_[2];
	      break; }
	    } // case faceNr
	  } else {
	    L = l;
	    refinementDone = true;
	    DEBUGL(1, "refinement did not work!");
	  }
	} // catch
      } // while
    } // if
  }

} // namespace hp3D
