/* space class for 3D hp FEM
 * enrichElm_
 */

#include "hp3D/space.hh"
#include "basics/debug.hh"

#define hpSpace3Denrich_D 0

namespace hp3D {

  // ***************************************************************** Space **

  void Space::enrichElm_(const concepts::Hexahedron3d& cell, int* p) {
    DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - start.");

    concepts::Hexahedron& cntr = cell.connector();
    concepts::AdaptiveControlP<3>& ctrl3(ctrl3_[cntr.key()]);

    DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - " << cntr << ", "
	   << ctrl3);
    DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - P = (i "
	   << p[0] << ", " << p[1] << ", " << p[2] << ", f0 "
	   << p[3] << ", " << p[4] << ", f1 " << p[5] << ", "
	   << p[6] << ", f2 " << p[7] << ", " << p[8] << ", f3 "
	   << p[9] << ", " << p[10] << ", f4 " << p[11] << ", "
	   << p[12] << ", f5 " << p[13] << ", " << p[14] << ", e0 "
	   << p[15] << ", e1 " << p[16] << ", e2 " << p[17] << ", e3 "
	   << p[18] << ", e4 " << p[19] << ", e5 " << p[20] << ", e6 "
	   << p[21] << ", e7 " << p[22] << ", e8 " << p[23] << ", e9 "
	   << p[24] << ", e10 " << p[25] << ", e11 " << p[26]
	   << ')');

    if (ctrl3.tag_.isMember()) {
      // store given information in the right place
      concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
	concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[cntr2->key()]);
	for (uint j = 0; j < 2; ++j)
	  if (p[3+2*i+j] > ctrl2.p_[j]) {
	    ctrl2.p_[j] = p[3+2*i+j];
	    DEBUGL(hpSpace3Denrich_D, cell.connector().key()
		   << " - updated " << *cntr2 << " to " << ctrl2);
	  }
      } // for faces
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	if (p[15+i] > ctrl1.p_[0]) {
	  ctrl1.p_[0] = p[15+i];
	  DEBUGL(hpSpace3Denrich_D, cell.connector().key()
		 << " - updated " << *cntr1 << " to " << ctrl1);
	}
      } // for edges
      DEBUGL(hpSpace3Denrich_D, cell.connector().key()
	     << " - stored information.");
    } else {
      DEBUGL(hpSpace3Denrich_D, cell.connector().key()
	     << " - gathering information.");
      // check if there are degrees of freedom here and store the relevant
      // polynomial degrees
      concepts::Connector2* cntr2 = 0;
      for(uint i = 0; (cntr2 = cntr.face(i)) != 0; ++i) {
	concepts::AdaptiveControlP<2>& ctrl2((*ctrl2_)[cntr2->key()]);
	std::map<uint, concepts::FaceData>::const_iterator j =
	  faceList_.find(cntr2->key().key());
	if (ctrl2.tag_.isMember() && ctrl2.tag_.isActive() &&
	    (j != faceList_.end()) && (j->second.hasCell(cell.connector()))) {
	  if ((ctrl2.p_[0] > 0) && (ctrl2.p_[1] > 0)) {
	    for (uint j = 0; j < 2; ++j)
	      if (p[3+2*i+j] > ctrl2.p_[j]) p[3+2*i+j] = ctrl2.p_[j];
	  }
	}
      } // for faces
      concepts::Connector1* cntr1 = 0;
      for(uint i = 0; (cntr1 = cntr.edge(i)) != 0; ++i) {
	concepts::AdaptiveControlP<1>& ctrl1((*ctrl1_)[cntr1->key()]);
	std::map<uint, concepts::EdgeData>::const_iterator j =
	  edgeList_.find(cntr1->key().key());
	if (ctrl1.tag_.isMember() && ctrl1.tag_.isActive() &&
	    (j != edgeList_.end()) && (j->second.hasCell(cell.connector()))) {
	  if (ctrl1.p_[0] > 0) {
	    if (p[15+i] > ctrl1.p_[0]) p[15+i] = ctrl1.p_[0];
	  }
	}
      } // for edges
      DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - P = (i "
	     << p[0] << ", " << p[1] << ", " << p[2] << ", f0 "
	     << p[3] << ", " << p[4] << ", f1 " << p[5] << ", "
	     << p[6] << ", f2 " << p[7] << ", " << p[8] << ", f3 "
	     << p[9] << ", " << p[10] << ", f4 " << p[11] << ", "
	     << p[12] << ", f5 " << p[13] << ", " << p[14] << ", e0 "
	     << p[15] << ", e1 " << p[16] << ", e2 " << p[17] << ", e3 "
	     << p[18] << ", e4 " << p[19] << ", e5 " << p[20] << ", e6 "
	     << p[21] << ", e7 " << p[22] << ", e8 " << p[23] << ", e9 "
	     << p[24] << ", e10 " << p[25] << ", e11 " << p[26]
	     << ')');

      // get subdivision strategy
      enum subdivTypes { NONE = 0, X = 1, Y = 2, Z = 4 };
      uint subdiv = 0;
      const concepts::Hex3dSubdivision* strategy = cell.getStrategy();
      conceptsAssert(strategy != 0, concepts::Assertion());
      if (strategy == concepts::Hex3dSubdiv8::instance()) subdiv = X | Y | Z;
      if (strategy == concepts::Hex3dSubdiv4z::instance()) subdiv = X | Y;
      if (strategy == concepts::Hex3dSubdiv4y::instance()) subdiv = X | Z;
      if (strategy == concepts::Hex3dSubdiv4x::instance()) subdiv = Y | Z;
      if (strategy == concepts::Hex3dSubdiv2z::instance()) subdiv = Z;
      if (strategy == concepts::Hex3dSubdiv2y::instance()) subdiv = Y;
      if (strategy == concepts::Hex3dSubdiv2x::instance()) subdiv = X;
      conceptsAssert(subdiv != 0, concepts::Assertion());
      DEBUGL(hpSpace3Denrich_D, cell.connector().key()
	     << " - subdiv = " << subdiv);

      // loop over the children of the cell
      const concepts::Hexahedron3d* child = 0;
      for (uint i = 0; (child = cell.child(i)) != 0; ++i) {
	DEBUGL(hpSpace3Denrich_D, cell.connector().key()
	       << " - child " << i << " = " << *child);
	// initialize data for child
	int P[27];
	for (uint j = 0; j < 27; ++j) P[j] = -1;

	// update data for child from faces
	const concepts::Connector2* cntr2 = 0;
	for(uint f = 0; (cntr2 = cntr.face(f)) != 0; ++f) {
// 	  DEBUGL(hpSpace3Denrich_D, cell.connector().key()
// 		 << " - face " << f << " = " << *cntr2);
	  uint idxFrom = 3+2*f;
	  int idxTo = -1;
	  switch (f) {
	  case 0: if (subdiv & Z) idxTo = 3+2*5; break;
	  case 1: if (subdiv & Y) idxTo = 3+2*3; break;
	  case 2: if (subdiv & X) idxTo = 3+2*4; break;
	  case 3: if (subdiv & Y) idxTo = 3+2*1; break;
	  case 4: if (subdiv & X) idxTo = 3+2*2; break;
	  case 5: if (subdiv & Z) idxTo = 3+2*0; break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  } // switch face
	  if (idxTo != -1) {
	    bool doUpdate = false;
	    const concepts::Connector2* face_child = 0;
	    for(uint k = 0; (face_child = cntr2->child(k)) != 0; ++k) {
	      if (face_child->key() == child->connector().face(f)->key())
		doUpdate = true;
	    }
	    if (doUpdate) {
	      for (uint j = 0; j < 2; ++j)
		if (P[idxTo+j] < p[idxFrom+j]) P[idxTo+j] = p[idxFrom+j];
	    }
	  }
	  for (uint j = 0; j < 2; ++j)
	    if (P[idxFrom+j] < p[idxFrom+j]) P[idxFrom+j] = p[idxFrom+j];
	} // for faces
	DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - P = (i "
	       << P[0] << ", " << P[1] << ", " << P[2] << ", f0 "
	       << P[3] << ", " << P[4] << ", f1 " << P[5] << ", "
	       << P[6] << ", f2 " << P[7] << ", " << P[8] << ", f3 "
	       << P[9] << ", " << P[10] << ", f4 " << P[11] << ", "
	       << P[12] << ", f5 " << P[13] << ", " << P[14] << ", e0 "
	       << P[15] << ", e1 " << P[16] << ", e2 " << P[17] << ", e3 "
	       << P[18] << ", e4 " << P[19] << ", e5 " << P[20] << ", e6 "
	       << P[21] << ", e7 " << P[22] << ", e8 " << P[23] << ", e9 "
	       << P[24] << ", e10 " << P[25] << ", e11 " << P[26]
	       << ')');

	// update data for child from edges
	const concepts::Connector1* cntr1 = 0;
	for (uint e = 0; (cntr1 = cntr.edge(e)) != 0; ++e) {
// 	  DEBUGL(hpSpace3Denrich_D, cell.connector().key()
// 		 << " - edge " << e << " = " << *cntr1);
	  uint idxFrom = 15+e;
	  int idxTo[2] = { -1, -1};
	  if ((e < 4) && (subdiv & Z)) idxTo[0] = 15+e+8;
	  if ((e > 7) && (subdiv & Z)) idxTo[0] = 15+e-8;
	  switch (e) {
	  case 0: if (subdiv & Y) idxTo[1] = 15+2; break;
	  case 1: if (subdiv & X) idxTo[1] = 15+3; break;
	  case 2: if (subdiv & Y) idxTo[1] = 15+0; break;
	  case 3: if (subdiv & X) idxTo[1] = 15+1; break;
	  case 4: if (subdiv & X) idxTo[1] = 15+5;
	    if (subdiv & Y) idxTo[0] = 15+7;
	    break;
	  case 5: if (subdiv & X) idxTo[1] = 15+4;
	    if (subdiv & Y) idxTo[0] = 15+6;
	    break;
	  case 6: if (subdiv & Y) idxTo[1] = 15+5;
	    if (subdiv & X) idxTo[0] = 15+7;
	    break;
	  case 7: if (subdiv & Y) idxTo[1] = 15+4;
	    if (subdiv & X) idxTo[0] = 15+6;
	    break;
	  case 8: if (subdiv & Y) idxTo[1] = 15+10; break;
	  case 9: if (subdiv & X) idxTo[1] = 15+11; break;
	  case 10: if (subdiv & Y) idxTo[1] = 15+8; break;
	  case 11: if (subdiv & X) idxTo[1] = 15+9; break;
	  default: throw conceptsException(concepts::ExceptionBase());
	  }
	  for (uint l = 0; l < 2; ++l) {
	    if (idxTo[l] != -1) {
	      bool doUpdate = false;
	      const concepts::Connector1* edge_child = 0;
	      for(uint k = 0; (edge_child = cntr1->child(k)) != 0; ++k) {
		if (edge_child->key() == child->connector().edge(e)->key())
		  doUpdate = true;
	      }
	      if (doUpdate)
		if (P[idxTo[l]] < p[idxFrom]) P[idxTo[l]] = p[idxFrom];
	    }
	  }
	  if (P[idxFrom] < p[idxFrom]) P[idxFrom] = p[idxFrom];
	} // for edges
	DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - P = (i "
	       << P[0] << ", " << P[1] << ", " << P[2] << ", f0 "
	       << P[3] << ", " << P[4] << ", f1 " << P[5] << ", "
	       << P[6] << ", f2 " << P[7] << ", " << P[8] << ", f3 "
	       << P[9] << ", " << P[10] << ", f4 " << P[11] << ", "
	       << P[12] << ", f5 " << P[13] << ", " << P[14] << ", e0 "
	       << P[15] << ", e1 " << P[16] << ", e2 " << P[17] << ", e3 "
	       << P[18] << ", e4 " << P[19] << ", e5 " << P[20] << ", e6 "
	       << P[21] << ", e7 " << P[22] << ", e8 " << P[23] << ", e9 "
	       << P[24] << ", e10 " << P[25] << ", e11 " << P[26]
	       << ')');

	// go to child
	enrichElm_(*child, P);
      } // for children
    }

    DEBUGL(hpSpace3Denrich_D, cell.connector().key() << " - done.");
  }

}
