/* testsuite for 3D cells with geometric deadlock
 * create space
 */

#include <memory>
#include "basics/exceptions.hh"
#include "geometry.hh"
#include "hp3D/space.hh"
#include "deadlock2.hh"

namespace test {

  // ************************************************************* DeadLock2 **

  void DeadLock2::run() {
    testSpace();
  }

  void DeadLock2::testSpace() {
    using namespace concepts;

    try {
      Import3dMesh mesh
	("../../../../geometry/testsuite/deadlock1.Coord.dat",
	 "../../../../geometry/testsuite/deadlock1.Elements.dat",
	 "../../../../geometry/testsuite/deadlock1.Boundary.dat");
      _test(mesh.ncell() == 2);

      const Hex3dSubdivision* subdivs[] = {
	Hex3dSubdiv2z::instance(), Hex3dSubdiv4y::instance(),
	Hex3dSubdiv2x::instance()
      };

      Level<3> level[] = {
	Level<3>(0), Level<3>(1)
      };
      level[0].l_[2] = 1; level[1].l_[0] = 0;

      concepts::Stacktrace::doit() = false;

      const uint childIdx[] = { 3, 5 };

      std::auto_ptr<Scan3> sc(mesh.scan());
      uint i = 0;
      while (*sc) {
	Cell3& cell = (*sc)++;
	Hexahedron3d* ch = dynamic_cast<Hexahedron3d*>(&cell);
	if (ch) {
	  ch->setStrategy(subdivs[i]);
	  _test(cell.child(1)->connector().key().key() == childIdx[i]);
	  ++i;
	}
      }

      hp3D::Space space(mesh, 0, 1);
      space.rebuild();
      _test(space.dim() == 12);
      _test(space.nelm() == 2);
      std::auto_ptr<hp3D::Space::Scan> scS(space.scan());
      i = 0;
      while (*scS) {
	hp3D::Element<Real>& elm = (*scS)++;
	space.adjust(elm, AdaptiveAdjustP<3>(level[i], 1));
      }
      space.rebuild();
      _test(space.dim() == 83);
      _test(space.nelm() == 6);
    }
    catch (ExceptionBase& e) {
      std::cout << e << std::endl;
    }
  }

} // namespace test

int main(int, char**) {
  test::DeadLock2 t;
  t.run();
  return t.report();
}
