/* testsuite for 3D cells with geometric deadlock
 * create space
 */

#include <memory>
#include "basics/exceptions.hh"
#include "geometry.hh"
#include "hp3D/space.hh"
#include "graphics.hh"
#include "deadlock3.hh"

namespace test {

  // ************************************************************* DeadLock3 **

  void DeadLock3::run() {
    testSpace();
  }

  void DeadLock3::testSpace() {
    using namespace concepts;

    try {
      Import3dMesh mesh
	("../../../../hp3D/testsuite/deadlock3.Coord.dat",
	 "../../../../hp3D/testsuite/deadlock3.Elements.dat",
	 "../../../../geometry/testsuite/deadlock1.Boundary.dat");
      _test(mesh.ncell() == 3);

      uint i = 0;
      std::auto_ptr<Scan3> sc(mesh.scan());
      while (*sc) {
	Cell3& cell = (*sc)++;
	_test(cell.connector().key().key() == i++);
      }

      Level<3> level[] = {
	Level<3>(0), Level<3>(0), Level<3>(0)
      };
      level[0].l_[2] = 1; level[1].l_[1] = 1; level[2].l_[0] = 1;

      concepts::Stacktrace::doit() = false;

      hp3D::Space space(mesh, 0, 1);
      space.rebuild();
      _test(space.dim() == 16);
      _test(space.nelm() == 3);
      graphics::drawMeshTecplot(space, "deadlock3-1.tecplot.dat", 3);

      std::auto_ptr<hp3D::Space::Scan> scS(space.scan());
      i = 0;
      while (*scS) {
	hp3D::Element<Real>& elm = (*scS)++;
	space.adjust(elm, AdaptiveAdjustP<3>(level[i++], 0));
      }
      space.rebuild();
      _test(space.dim() == 30);
      _test(space.nelm() == 12);
      graphics::drawMeshTecplot(space, "deadlock3-2.tecplot.dat", 3);

      scS.reset(space.scan());
      while (*scS) {
	hp3D::Element<Real>& elm = (*scS)++;
	space.adjust(elm, AdaptiveAdjustP<3>(0, 1));
      }
      space.rebuild();
      _test(space.dim() == 143);
      _test(space.nelm() == 12);
    }
    catch (ExceptionBase& e) {
      std::cout << e << std::endl;
    }
  }

} // namespace test

int main(int, char**) {
  test::DeadLock3 t;
  t.run();
  return t.report();
}
