// Boundary integrals for linear DG FEM in 1D

#ifndef linDG1DbilinearForm_hh
#define linDG1DbilinearForm_hh

#include "basics/typedefs.hh"
#include "operator/bilinearForm.hh"
#include "geometry/boundaryConditions.hh"

namespace linDG1D {

  using concepts::Real;

  // *********************************************************** BoundaryInt **

  /** Boundary integral for the DG FEM in 1D. This bilinear form integrates
      \f[ \int_{K\cap K'} \langle \phi_i \rangle \cdot [\phi_j] \, dx. \f]
      To generate the complete boundary integral
      \f[ \int_v \langle \phi_i \rangle \cdot [\phi_j] -
      [\phi_i] \cdot \langle \phi_j' \rangle \f]
      combine this matrix with the negative transpose of itself.
      @author Thomas Wihler, 2003
  */
  class BoundaryInt : public concepts::BilinearForm<Real> {
  public:
    BoundaryInt(concepts::BoundaryConditions* bc = 0) : bc_(bc) {}
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em);
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const concepts::ElementPair<Real>& ep);
  private:
    concepts::BoundaryConditions* bc_;
  };

  // ******************************************************* BoundaryIntStab **

  /** Stabilizing boundary integral for the DG FEM in 1D. This
      bilinear form integrates
      \f[ \int_{K\cap K'} [\phi_i] \cdot [\phi_j] \, dx. \f]
      @author Thomas Wihler, 2003
  */
  class BoundaryIntStab : public concepts::BilinearForm<Real> {
  public:
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em) {
      conceptsAssert(false, concepts::Assertion());
    }
    virtual void operator()(const concepts::Element<Real>& elmX,
			    const concepts::Element<Real>& elmY,
			    concepts::ElementMatrix<Real>& em,
			    const concepts::ElementPair<Real>& ep);
  };

} // namespace linDG1D

#endif // linDG1DbilinearForm_hh
