#include "edgeInfo.hh"
#include "triangle.hh"

namespace linDG2D {

  using concepts::Real;
  using concepts::Real2d;

  EdgeInfo::EdgeInfo() {
    cells_[0] = 0;
    cells_[1] = 0;
  }

  void EdgeInfo::addCell(const concepts::Triangle2d* cell, const uint idx) {
    conceptsAssert(cells_[1] == 0, concepts::Assertion());
    if (cells_[0] == 0) {
      cells_[0] = cell;
      idxs_[0] = idx;
    }
    else {
      cells_[1] = cell;
      idxs_[1] = idx;
    }
  }

  Real EdgeInfo::length() const {
    Real2d edgeVector = 
      cells_[0]->vertex((idxs_[0]+1)%3) -
      cells_[0]->vertex(idxs_[0]);
    return edgeVector.l2();
  }

  Real EdgeInfo::meanHeight() const {
    Real res = 1./cells_[0]->area();
    if(cells_[1]) {
      res += 1./cells_[1]->area();
      res *= 0.5;
    }
    return 1./(res*length());
  }


  const Real2d EdgeInfo::normal() const {
    Real2d edgeVector = 
      cells_[0]->vertex((idxs_[0]+1)%3) -
      cells_[0]->vertex(idxs_[0]);
    Real2d normal(edgeVector[1], -edgeVector[0]);
    return normal * (1.0/normal.l2());
  }

  Real2d EdgeInfo::mapRefTri(Real xi, uint i) const {
    Real xim = i == 0 ? xi : 1.0 - xi;
    switch(idxs_[i]) {
    case 0: return Real2d(xim, 0.0);
    case 1: return Real2d(1.0, xim);
    case 2: return Real2d(1.0-xim, 1.0-xim);
    default: conceptsAssert(false, concepts::Assertion());
      return Real2d(0.0, 0.0);
    }
  }

  Real2d EdgeInfo::map(Real xi) const {
    Real2d v0 = cells_[0]->vertex(idxs_[0]);
    Real2d v1 = cells_[0]->vertex((idxs_[0]+1)%3);
    return v1*xi + v0*(1.-xi);
  }

} // namespace linDG2D
