#include "elementPairList.hh"
#include "elementPair.hh"
#include "meshInfo.hh"
#include "space.hh"

namespace linDG2D {

  ElementPairList::ElementPairList(const Space& spc0,
                                   const Space& spc1,
                                   const MeshInfo& meshInfo) {
    std::auto_ptr<MeshInfo::Scanner> sc(meshInfo.scan());
    while (*sc) {
      EdgeInfo& ei = (*sc)++;
      const concepts::Triangle2d* cell0 = ei.cell(0);
      const concepts::Triangle2d* cell1 = ei.cell(1);
      add(new ElementPair(spc0.elm(cell0), spc1.elm(cell0), ei));
      if (cell1 != 0) {
        // inner edge
        add(new ElementPair(spc0.elm(cell0), spc1.elm(cell1), ei));
        add(new ElementPair(spc0.elm(cell1), spc1.elm(cell0), ei));
        add(new ElementPair(spc0.elm(cell1), spc1.elm(cell1), ei));
      }
    }
  }

} // namespace linDG2D
