#ifndef DGElementPairList_hh
#define DGElementPairList_hh

#include "space/elementPairs.hh"

namespace linDG2D {

  // forward declarations
  class Space;
  class MeshInfo;

  /** Element pair list for two DG spaces. Given two spaces \c spc1 and \c
      spc2 on the same mesh, this element pair list is built up such that
      it contains 4 element pairs for each inner edge and 1 element pair
      for each boundary edge. The first element of
      the pair is taken from \c spc1, the second one from \c spc2. So
      \c spc1 represents the test space, and \c spc2 the ansatz space.
      @author Christoph Winkelmann, 2004
  */
  class ElementPairList : public concepts::ElementPairList<concepts::Real> {
  public:
    /** Constructor. Scans \c meshInfo and adds to the ElementPairList
        represented by this class 4 element pairs for each inner edge
        and 1 element pair for each boundary edge. An edge is an inner
        edge if both cells are != 0 in DGEdgeInfo. The first element of
        the pair is taken from \c spc1, the second one from \c spc2.
    */
    ElementPairList(const Space& spc1, const Space& spc2, 
                    const MeshInfo& meshInfo);
  };

} // namespace linDG2D

#endif // DGElementPairList_hh
