// testing linDG2D::ElementPair

#include "basics/testsuite.hh"
#include "basics/exceptions.hh"

#include "elementPairTest.hh"
#include "triangle.hh"
#include "edgeInfo.hh"
#include "elementPair.hh"

namespace test {
  
  // *************************************************** DGElementPair2dTest **

  void DGElementPair2dTest::run() {
    concepts::Vertex v0, v1, v2, v3;
    concepts::Edge e0(v0, v1);
    concepts::Edge e1(v1, v2);
    concepts::Edge e2(v2, v0);
    concepts::Edge e3(v1, v3);
    concepts::Edge e4(v3, v2);
    concepts::Triangle tco0(e0, e1, e2);
    concepts::Triangle tco1(e1, e3, e4);
    concepts::Real2d p0(-4.5, -3.0);
    concepts::Real2d p1(3.0, -1.5);
    concepts::Real2d p2(-1.5, 3.0);
    concepts::Real2d p3(3.0, 3.0);
    concepts::VertexTriangle2d tm0(p0, p1, p2);
    concepts::VertexTriangle2d tm1(p2, p1, p3);
    concepts::Triangle2d tce0(tco0, tm0);
    concepts::Triangle2d tce1(tco1, tm1);

    linDG2D::EdgeInfo ei[5];
    ei[0].addCell(&tce0, 0);
    ei[1].addCell(&tce0, 1);
    ei[1].addCell(&tce1, 0);
    ei[2].addCell(&tce0, 2);
    ei[3].addCell(&tce1, 1);
    ei[4].addCell(&tce1, 2);

    uint idx[3] = {0, 1, 2};
    linDG2D::TriangleP1 tel0(tce0, idx);
    idx[0] = 3;
    idx[1] = 4;
    idx[2] = 5;
    linDG2D::TriangleP1 tel1(tce1, idx);

    linDG2D::ElementPair* ep[5];
    ep[0] = new linDG2D::ElementPair(tel0, tel0, ei[0]);
    ep[1] = new linDG2D::ElementPair(tel0, tel1, ei[1]);
    ep[2] = new linDG2D::ElementPair(tel0, tel0, ei[2]);
    ep[3] = new linDG2D::ElementPair(tel1, tel1, ei[3]);
    ep[4] = new linDG2D::ElementPair(tel1, tel1, ei[4]);

    testShapeFct(ep);
  }

  void DGElementPair2dTest::testShapeFct(linDG2D::ElementPair** ep) {
    _numtest(ep[0]->shapeFct(0, 0, 0), 1);
    _numtest(ep[0]->shapeFct(0, 0, 0.28), 1-2*0.28);
    _numtest(ep[0]->shapeFct(0, 0, 1), -1);
    _numtest(ep[0]->shapeFct(0, 1, 0), -1);
    _numtest(ep[0]->shapeFct(0, 1, 0.28), 2*0.28-1);
    _numtest(ep[0]->shapeFct(0, 1, 1), 1);
    _numtest(ep[0]->shapeFct(0, 2, 0), -1);
    _numtest(ep[0]->shapeFct(0, 2, 0.28), -1);
    _numtest(ep[0]->shapeFct(0, 2, 1), -1);

    _numtest(ep[1]->shapeFct(0, 0, 0), -1);
    _numtest(ep[1]->shapeFct(0, 0, 0.28), -1);
    _numtest(ep[1]->shapeFct(0, 0, 1), -1);
    _numtest(ep[1]->shapeFct(0, 1, 0), 1);
    _numtest(ep[1]->shapeFct(0, 1, 0.28), 1-2*0.28);
    _numtest(ep[1]->shapeFct(0, 1, 1), -1);
    _numtest(ep[1]->shapeFct(0, 2, 0), -1);
    _numtest(ep[1]->shapeFct(0, 2, 0.28), 2*0.28-1);
    _numtest(ep[1]->shapeFct(0, 2, 1), 1);

    _numtest(ep[1]->shapeFct(1, 0, 0), -1);
    _numtest(ep[1]->shapeFct(1, 0, 0.28), 2*0.28-1);
    _numtest(ep[1]->shapeFct(1, 0, 1), 1);
    _numtest(ep[1]->shapeFct(1, 1, 0), 1);
    _numtest(ep[1]->shapeFct(1, 1, 0.28), 1-2*0.28);
    _numtest(ep[1]->shapeFct(1, 1, 1), -1);
    _numtest(ep[1]->shapeFct(1, 2, 0), -1);
    _numtest(ep[1]->shapeFct(1, 2, 0.28), -1);
    _numtest(ep[1]->shapeFct(1, 2, 1), -1);

    _numtest(ep[2]->shapeFct(0, 0, 0), -1);
    _numtest(ep[2]->shapeFct(0, 0, 0.28), 2*0.28-1);
    _numtest(ep[2]->shapeFct(0, 0, 1), 1);
    _numtest(ep[2]->shapeFct(0, 1, 0), -1);
    _numtest(ep[2]->shapeFct(0, 1, 0.28), -1);
    _numtest(ep[2]->shapeFct(0, 1, 1), -1);
    _numtest(ep[2]->shapeFct(0, 2, 0), 1);
    _numtest(ep[2]->shapeFct(0, 2, 0.28), 1-2*0.28);
    _numtest(ep[2]->shapeFct(0, 2, 1), -1);
  }

}
