// linear form for 1D linear FEM

#ifndef linearForm1D_hh
#define linearForm1D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "function/linearForm.hh"
#include "toolbox/array.hh"
#include "toolbox/formula.hh"
#include "integration/quadrature.hh"

namespace linearFEM {

  using concepts::Real;

  // *************************************************************** Riesz1d **

  /** Linear form form 1D linear FEM. This linear form takes a formula \c frm
      and computes the load vector \f[ \int_K f v \, dx, \f]
      where f ist \c frm and v are the test functions.
      @author Philipp Frauenfelder, 2003
  */
  class Riesz1d : public concepts::LinearForm<Real> {
  public:
    /** Constructor
	@param frm Formula
	@param gauss_p Numer of points for the Gaussian quadrature
    */
    Riesz1d(const concepts::Formula& frm, const uint gauss_p = 1) :
      frm_(frm.clone()), quad_(gauss_p) {}
    virtual void operator()(const concepts::Element<Real>& elm,
			    concepts::ElementMatrix<Real>& em);
  private:
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    /// Quadrature formula
    const concepts::Quadrature<4> quad_;
  };

} // namespace linearFEM

#endif // linearForm2D_hh

