// linear form for 2D linear FEM (quads and triangles)

#ifndef linearForm2D_hh
#define linearForm2D_hh

#include "basics/typedefs.hh"
#include "basics/vectorsMatrices.hh"
#include "function/linearForm.hh"
#include "toolbox/array.hh"
#include "toolbox/formula.hh"

namespace linearFEM {

  using concepts::Real;

  class Quad;
  class Triangle;

  // *************************************************************** Riesz2d **

  /** Linear form form 2D linear FEM. This linear form takes a formula \c frm
      and computes the load vector \f[ \int_K f v \, dx, \f]
      where f ist \c frm and v are the test functions.
      @author Philipp Frauenfelder, 2002
  */
  class Riesz2d : public concepts::LinearForm<Real> {
  public:
    /** Constructor
	@param frm Formula
    */
    Riesz2d(const concepts::Formula& frm) : frm_(frm.clone()), shpfct_(0) {}
    virtual void operator()(const concepts::Element<Real>& elm,
			    concepts::ElementMatrix<Real>& em);
    /// Computes the local load vector for a quad
    void operator()(const linearFEM::Quad& elm,
		    concepts::ElementMatrix<Real>& em);
    /// Computes the local load vector for a triangle
    void operator()(const linearFEM::Triangle& elm,
		    concepts::ElementMatrix<Real>& em);
  private:
    /// Formula
    std::auto_ptr<const concepts::Formula> frm_;
    concepts::Array<concepts::Real> shpfct_;
    /// Integration weights (3rd order)
    static const Real weights_[4];
    /// Integration points (3rd order)
    static const concepts::Real2d coord_[4];
  };

} // namespace linearFEM

#endif // linearForm2D_hh

