# version number of the library
MAJOR=2
MINOR=0
PATCHLEVEL=0
VERSION=$(MAJOR).$(MINOR).$(PATCHLEVEL)

# LAPACK
LAPACK_LIB = -llapack -lblas -lg2c

# PETSc - iterative solvers and much more
HAS_PETSC = true
ifeq ($(HAS_PETSC),true)
CPPFLAGS += -DHAS_PETSC
export PETSC_DIR = /usr/local/app/petsc/SRC/petsc-2.1.3
export PETSC_ARCH = $(shell $(PETSC_DIR)/bin/petscarch)
BOPT = O
PETSC_LIB_PATH = $(PETSC_DIR)/lib/lib$(BOPT)/$(PETSC_ARCH)
LDFLAGS += $(LDRPATH)$(PETSC_LIB_PATH) -L$(PETSC_LIB_PATH)
endif

# MPI - message passing
HAS_MPI = false
#MPI_VERSION=MPICH
MPI_VERSION=LAM
ifeq ($(HAS_MPI),true)
CPPFLAGS += -D$(MPI_VERSION)
endif

# SuperLU - sparse direct solver
HAS_SuperLU = false
ifeq ($(HAS_SuperLU),true)
SuperLU_PATH = $(CONCEPTSDIR)/../SuperLU
SuperLU_LIB_PATH = $(SuperLU_PATH)
SuperLU_INC_PATH = $(SuperLU_PATH)/SRC
CPPFLAGS += -DHAS_SuperLU
SuperLU_LIB = -lsuperlu $(LAPACK_LIB)
endif

# JDBSYM - eigenvalue solver
HAS_JDBSym = false
ifeq ($(HAS_JDBSym),true)
JDBSym_PATH = $(CONCEPTSDIR)/../jdbsym
JDBSym_INC_PATH = $(JDBSym_PATH)
JDBSym_LIB_PATH = $(JDBSym_PATH)
JDBSym_LIB = -ljdbsym -ljdqrgen $(LAPACK_LIB)
endif

# ARPACK - eigenvalue solver
HAS_ARPACK = false
ifeq ($(HAS_ARPACK),true)
ARPACK_PATH = $(CONCEPTSDIR)/../ARPACK
ARPACK_LIB_PATH = $(ARPACK_PATH)
ARPACK_LIB = -larpack -lg2c
endif

# NRLsolver - sparse direct solver
HAS_NRLsolver = false
ifeq ($(HAS_NRLsolver),true)
NRLsolver_PATH = $(CONCEPTSDIR)/../NRLsolver
NRLsolver_LIB_PATH = $(NRLsolver_PATH)/lib
NRLsolver_LIB = -lcholesky -lreorder -lmetis
NRLsolver_INC_PATH = $(NRLsolver_PATH)/include
endif

# UMFPACK - sparse direct solver
HAS_UMFPACK = false
ifeq ($(HAS_UMFPACK),true)
UMFPACK_PATH = $(CONCEPTSDIR)/../umfpack-4.1
UMFPACK_LIB_PATH = $(UMFPACK_PATH)/lib
UMFPACK_LIB = -lumfpack -lamd
UMFPACK_INC_PATH = $(UMFPACK_PATH)/UMFPACK/Include
CPPFLAGS += -DUMFPACK_41
endif

# PARDISO - sparse direct solver
HAS_PARDISO = false
ifeq ($(HAS_PARDISO),true)
PARDISO_PATH = $(CONCEPTSDIR)/../pardiso_example
PARDISO_LIB_PATH = $(PARDISO_PATH)
PARDISO_LIB = -lpardiso32
PARDISO_INC_PATH = $(PARDISO_PATH)
endif

# parts of the project, ie. most of the subdirectories of the root directory
PROJECTS = basics function geometry integration operator space toolbox \
	   graphics constraints vectorial \
	   eigensolver sparseqr timestepping

ifeq ($(HAS_MPI),true)
PROJECTS += toolbox_p geometry_p distributor_p
endif

# extending parts of the project, ie. parts which don't belong to the
# core library
extPROJECTS = linearFEM hp1D hp3D hp2D \
	      linDG3D linDG1D linDG2D

# libraries are created from extPROJECTS, lib from PROJECTS
# more than one extPROJECT can go into the same library
LIBRARIES = lib liblinearFEM libhp1D libhp3D libhp2D libDG

# Applications use the library and are link against it. If you want to
# keep your things separatly, create a new applications directory and
# register it here.
APPLICATIONS = applications app-walser \
	       app-fvdgfem app-cwinkelm app-emanuel

RELEASE_MODULES = $(PROJECTS) \
		  linearFEM hp1D hp3D hp2D linDG3D linDG1D linDG2D
RELEASE_DIRS = builds doc makesupport applications app-cwinkelm app-walser \
		app-emanuel app-fvdgfem
