###############################################################################
##  check for correct $(config)
###############################################################################
ifeq ($(filter $(config),$(subst /, ,$(shell pwd))),$(config))
else
$(error wrong configuration $$(config) = $(config))
endif

###############################################################################
##  some paths
###############################################################################

srcdir = $(CONCEPTSDIR)
VPATH += $(srcdir) $(srcdir)/builds/$(config) $(srcdir)/$(project)

###############################################################################
#  standard targets
###############################################################################

.PHONY: all clean mostlyclean 

all: export $(executable) $(library)

test: $(executable)
	rm -f *.out
ifneq ($(executable),)
	set -e; \
	for i in $(executable); do \
	  echo testing $$i ;\
	  ./$$i > $$i.out ;\
	done
endif

###############################################################################
#  generate and include source file dependencies
###############################################################################

%.d: %.cc 
	$(SHELL) -ec '$(D) $< | $(PERL) $(CONCEPTSDIR)/makesupport/changedepends.pl $(project) $(*F) $(@F) > $@'

%.d: %.c 
	$(SHELL) -ec '$(D) $< | $(PERL) $(CONCEPTSDIR)/makesupport/changedepends.pl $(project) $(*F) $(@F) > $@'

ifneq ($(strip $(object)), )
-include $(object:.o=.d)
endif

###############################################################################
#  generate the table of contents for a given project 
###############################################################################

export: $(object) $(library)
	$(RM) -f export
	$(if $(strip $(object) $(library)),\
	  echo $(strip $(object) $(library)) > export \
	)

###############################################################################
#  expand projects into libraries and relocatable objects
###############################################################################

extractLibs = $(filter -l% -L%,\
	$(patsubst $(CONCEPTSDIR)/builds/$(config)/lib/lib%.so,-l%,\
           $(strip $(foreach p, $(1),\
               $(if $(filter %.o %.a %.so -l% -L%, $(p)),\
                 $(p),\
                 $(foreach q,\
                   $(shell cat $(srcdir)/builds/$(config)/$(p)/export),\
                   $(filter -l% -L% /%, $(q)) \
                   $(addprefix $(p)/,$(filter-out -l% -L% /%, $(q)))\
                 )\
               )\
           ))\
         ))

expand = $(filter-out -l%,\
	$(patsubst $(CONCEPTSDIR)/builds/$(config)/lib/lib%.so,-l%,\
           $(strip $(foreach p, $(1),\
               $(if $(filter %.o %.a %.so -l%, $(p)),\
                 $(p),\
                 $(foreach q,\
                   $(shell cat $(srcdir)/builds/$(config)/$(p)/export),\
                   $(filter -l% /%, $(q)) \
                   $(addprefix $(p)/,$(filter-out -l% -L% /%, $(q)))\
                 )\
               )\
           ))\
         ))

getlib = $(shell cat \
             $(CONCEPTSDIR)/builds/$(config)/$(strip $(1))/libraries)

###############################################################################
#  include & update imported projects
###############################################################################

CPPFLAGS += -I$(srcdir) -I.
# CPPFLAGS += $(addprefix -I$(srcdir)/, $(sort $(dir $(import)))) -I.
# CPPFLAGS += $(addprefix -I$(srcdir)/, $(sort $(dir $(import)))) -I. -I$(PETSC_DIR)/include -I$(PETSC_DIR)/bmake/$(PETSC_ARCH) -I$(PETSC_DIR)/src/sys/src/mpiuni

importing:
	@$(PERL) $(CONCEPTSDIR)/makesupport/makedepend.pl $(CONCEPTSDIR) $(config) $(import)

.PHONY: importing

-include importing

###############################################################################
#  link executables
###############################################################################

%: %.o
	$(CXX) -o $@ $(LDFLAGS) \
	  $(patsubst -L%,$(LDRPATH)%,$(filter -L$(CONCEPTSDIR)/builds/$(config)/%,$(LIBS))) \
	  $(addprefix $(LDRPATH),$(filter-out /usr/lib/ /usr/local/lib/, \
	    $(sort $(dir $^)))) \
	  $^ $(LIBS)

###############################################################################
#  generate parser files
###############################################################################

%.yy.c: %.l
	$(FLEX) -o$(notdir $(<:.l=.yy.c)) $^

%.tab.cc: %.y
	$(YACC) $(YFLAGS) $< -o $(notdir $(<:.y=.tab.cc))

.SECONDARY: input.yy.c input.tab.cc

.SUFFIXES: .cc .o .d .c .hh .h .l .y .mk
