/* Basic operator declarations such as compositions of operators
 */

#include "compositions.cc"

#if defined(__hpux) || defined(_IBMR2)
  #define FORTRAN(X) X
#else
  #define FORTRAN(X) X ## _
#endif

#ifdef __cplusplus
extern "C" {
#endif
  extern void FORTRAN(zgemm)(const char& TRANSA, const char& TRANSB, 
			     const int& M, const int& N, const int& K, 
			     const std::complex<double>& ALPHA, 
			     const std::complex<double>* const A,
			     const int& LDA,
			     const std::complex<double>* const B,
			     const int& LDB, 
			     const std::complex<double>& BETA,
			     const std::complex<double>* C,
			     const int& LDC);
#ifdef __cplusplus
}
#endif

namespace concepts {

  // ************************************************************** Operator **

  template class Operator<Cmplx>;

  // *************************************************************** Compose **

  template<>
  void Compose<Cmplx>::collapse_(DenseMatrix<Cmplx>& A, DenseMatrix<Cmplx>& B,
				 Matrix<Cmplx>& dest) const {
    // action with zgemm_
    const int m = B.spaceY().dim();
    const int n = A.spaceX().dim();
    Array<Cmplx> res(m*n);
    const int k = B.spaceX().dim();
    const Cmplx alpha = 1.0;
    const int lda = B.spaceY().dim();
    const Cmplx beta = 0.0;
    const int ldb = A.spaceY().dim();
    const int ldc = B.spaceY().dim();
    DEBUGL(ComposeCollapse_D, "calling zgemm with parameters " 
	   << 'N' << ", " << 'N' << ", " << m << ", " << n << ", " << k 
	   << ", " << alpha << ", " << (Cmplx*)B << ", " << lda
	   << ", " << (Cmplx*)A << ", " << ldb << ", " << beta
	   << ", " << dest << ", " << ldc);
    FORTRAN(zgemm)('N', 'N', m, n, k, alpha, (std::complex<double>*)B,
		   lda, (std::complex<double>*)A, ldb, beta,
		   (std::complex<double>*)res, ldc);
    DEBUGL(ComposeCollapse_D, "res = " << res);
    const Cmplx* a = (Cmplx*)res;
    for (uint i = 0; i < (uint)n; ++i)
      for (uint j = 0; j < (uint)m; ++j, ++a)
	if (*a != 0.0) dest(i,j) = *a;
  }

  template class Compose<Cmplx>;

  // ***************************************************************** LiCoI **

  template class LiCoI<Cmplx>;

  // ****************************************************************** LiCo **

  template class LiCo<Cmplx>;

} // namespace concepts
