// solves a constrained source problem

#include "constrained.hh"
#include "solverFabric.hh"
#include "cg.hh"

namespace concepts {

  // *********************************************************** Constrained **

  Constrained::Constrained(SparseMatrix<Real>& S, SparseMatrix<Real>& C,
			   const Vector<Real>& g,
			   SolverFabric<Real>& fabric) :
    C_(C), S_(S), Ct_(C, true), g_(g), SRg_(S.spaceX()), CCtinv_(0),
    solver_(0), R_(0), Qmin_(0), CtC_(0), QS_(0), QSQ_(0), Q_(0), Sp_(0),
    fabric_(fabric) {
    conceptsAssert(S_.spaceX().dim() == S_.spaceY().dim(), Assertion());
    conceptsAssert(S_.spaceX().dim() == C_.spaceY().dim(), Assertion());

  }

  void Constrained::operator()(const Function<Real>& fncY,
			       Function<Real>& fncX) {
    if (solver_.get() == 0) {
      // Compute CCtinv
      Compose<Real> CCt(C_, Ct_);
      SparseMatrix<Real> CCtmatrix(CCt);
      CCtinv_.reset(fabric_(CCtmatrix));

      // Compute R and Q
      R_.reset(new concepts::Compose<Real>(Ct_, *CCtinv_));
      Qmin_.reset(new Compose<Real>(*R_, C_));
      Q_.reset(new LiCoI<Real>(*Qmin_, -1.0));

      // Compute S'
      QS_.reset(new Compose<Real>(*Q_, S_));
      QSQ_.reset(new Compose<Real>(*QS_, *Q_));
      CtC_.reset(new Compose<Real>(Ct_, C_));
      Sp_.reset(new LiCo<Real>(*CtC_, *QSQ_));

      // Prepare computation of f'
      Vector<Real> Rg(S_.spaceX());
      (*R_)(g_, Rg);
      S_(Rg, SRg_);

      solver_.reset(new CG<Real>(*Sp_, 1e-15, 2000));
    }

    // Compute f'
    fncX = fncY;
    fncX -= SRg_;
    Vector<Real> fp(fncX.space());
    (*Q_)(fncX, fp);
    Ct_(g_, SRg_);
    fp += SRg_;

    (*solver_)(fp, fncX);
  }

  const Space<Real>& Constrained::spaceX() const {
    return S_.spaceY();
  }

  const Space<Real>& Constrained::spaceY() const {
    return S_.spaceX();
  }

  std::ostream& Constrained::info(std::ostream& os) const {
    return os << "Constrained()";
  }

} // namespace concepts
