// fabric for an optimized solver class for diagonal matrices

#include "diagonalFabric.hh"

namespace concepts {

  // ************************************************** DiagonalSolverFabric **

  template <typename F>
  Operator<F>* DiagonalSolverFabric<F>::operator()(Operator<F>& matrix) {
    Operator<F>* solver = 0;
    try {
      const DiagonalMatrix<F>& diagMatrix =
	dynamic_cast<const DiagonalMatrix<F>& >(matrix);
      solver = new DiagonalSolver<F>(diagMatrix);
    } catch (std::bad_cast& e) {
      throw conceptsException
	(MissingFeature("takes only DiagonalMatrix"));
    }
    return solver;
  }
  
  template <typename F>
  std::ostream& DiagonalSolverFabric<F>::info(std::ostream& os) const {
    return os << "DiagonalSolverFabric()";
  }
  
} // namespace concepts
