// fabric class for Pardiso

#ifndef pardisoFabric_hh
#define pardisoFabric_hh

#include "pardiso.hh"
#include "solverFabric.hh"

namespace concepts {

  // ********************************************************* PardisoFabric **

  /** Fabric class for \c Pardiso.
      @author Philipp Frauenfelder, 2002
  */
  class PardisoFabric : public SolverFabric<Real> {
  public:
    PardisoFabric(bool symmetric = false) : symmetric_(symmetric) {}
    virtual Operator<Real>* operator()(Operator<Real>& matrix) {
      return new Pardiso(matrix, symmetric_);
    }
    virtual Operator<Real>* operator()(SparseMatrix<Real>& matrix) {
      return new Pardiso(matrix, symmetric_);
    }
  protected:
    virtual std::ostream& info(std::ostream& os) const {
      return os << "PardisoFabric()";
    }
  private:
    bool symmetric_;
  };

} // concepts

#endif // pardisoFabric_hh
