// abstract class for solver of linear systems

#ifndef systemSolverFabric_hh
#define systemSolverFabric_hh

#include "sparseMatrix.hh"

namespace concepts {

  // ********************************************************** SolverFabric **

  /** Abstract fabric class for linear solvers. The fabric delivers a
      created object to a new owner which is then responsible for the
      deletion of the object.
      @author Philipp Frauenfelder, 2002
      @ingroup solver
  */
  template<typename F>
  class SolverFabric : public concepts::OutputOperator {
  public:
    /** Creates a linear solver from \c matrix. The caller is responsible
	for the deletion of the delivered object.
	@return Reference to the linear solver
    */
    virtual Operator<F>* operator()(Operator<F>& matrix) = 0;
    virtual Operator<F>* operator()(SparseMatrix<F>& matrix) {
      Operator<F>* m = dynamic_cast<Operator<F>*>(&matrix);
      return operator()(*m);
    }
  };

} // concepts

#endif // systemSolverFabric_hh
