/* A sparse matrix
 */

#include "basics/typedefs.hh"
#include "sparseMatrix.cc"

namespace concepts {

  template<>
  void SparseMatrix<Real>::histogram(std::map<int, uint>& hist) const {
    int hashBits = m_->HashBits();
    int p = 1 << hashBits;
    uint n = spaceX().dim();
    HashedSparseMatrix<Real>::Value** matrix = m_->Matrix();

    for (uint r = 0; r < n; ++r) {
      for (int c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	  if (v->val != 0.0) {
	    ++(hist[(int)(std::ceil(std::log10(std::fabs(v->val))))]);
	  }
	}
      }
    }
  }

  template class SparseMatrix<Real>;
}
