/* testsuite for gmres
 */

#include "basics/exceptions.hh"
#include "operator/matrix.hh"
#include "operator/denseMatrix.hh"
#include "operator/sparseMatrix.hh"
#include "operator/PETSc.hh"
#include "operator/gmres.hh"

int main(int, char**) {

  using namespace concepts;

  try {
    DummySpace<Real> spc(5);
    DummySpace<Cmplx> spcC(5);
    std::cout << spc << std::endl;
    std::cout << spcC << std::endl;

    std::cout << "=== DenseMatrix, Real ===" << std::endl;
    DenseMatrix<Real> m(spc, spc);
    m(0,0) = 1.0; m(0,1) = 0.1; m(0,3) = 3.6;
    m(1,1) = 2.0;
    m(2,2) = 3.0; m(2,4) = 9.0;
    m(3,3) = 4.0; m(3,4) = 18.0;
    m(4,4) = 5.0;
    std::cout << m << std::endl;

    Vector<Real> rhs(spc), sol(spc);
    rhs[0] = 1.0; rhs[1] = 2.0; rhs[2] = 3.0; rhs[3] = 4.0; rhs[4] = 5.0;
    std::cout << "rhs = " << rhs << std::endl;

    GMRes<Real> gmresm(m, 1e-12, 10);
    gmresm(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [13.5, 1, -2, -3.5, 1]" << std::endl;
    std::cout << gmresm << std::endl;

    std::cout << "=== DenseMatrix, Cmplx ===" << std::endl;
    DenseMatrix<Cmplx> mC(spcC, spcC);
    // matlab
    // mC=[1+2i 0 0 0 0; 2+3i 4+3i 0 0 0; 0 0 4+1i 2+3i 1+3i;
    //     1+2i 0 0 0 1+2i; 0 0 5+1i 2+3i 1+3i];
    // rhsC=[1+2i; 2+3i; 3+4i; 4+5i; 5+6i];
    mC(0,0) = Cmplx(1,2);
    mC(1,0) = Cmplx(2,3); mC(1,1) = Cmplx(4,3);
    mC(2,2) = Cmplx(4,1); mC(2,3) = Cmplx(2,3); mC(2,4) = Cmplx(1,3);
    mC(3,0) = Cmplx(1,2); mC(3,4) = Cmplx(1,2);
    mC(4,2) = Cmplx(5,1); mC(4,3) = Cmplx(2,3); mC(4,4) = Cmplx(1,3);
    std::cout << mC << std::endl;

    Vector<Cmplx> rhsC(spcC), solC(spcC);
    rhsC[0] = Cmplx(1,2); rhsC[1] = Cmplx(2,3); rhsC[2] = Cmplx(3,4);
    rhsC[3] = Cmplx(4,5); rhsC[4] = Cmplx(5,6);
    std::cout << "rhsC = " << rhsC << std::endl;

    GMRes<Cmplx> gmresmC(mC, 1e-12, 7);
    gmresmC(rhsC, solC);
    std::cout << "solC = " << solC
	      << " should be [(1,0), (0,0), (2,2), (-3.5077,-0.1385), "
	      << "(1.8,-0.6)]" << std::endl;
    std::cout << gmresmC << std::endl;

    std::cout << "=== SparseMatrix, Real ===" << std::endl;
    SparseMatrix<Real> mS(spc, spc);
    mS(0,0) = 2.0; mS(0,1) = 0.3; mS(0,3) = 3.5;
    mS(1,1) = 3.0;
    mS(2,2) = 2.0; mS(2,4) = 7.0;
    mS(3,3) = 6.0; mS(3,4) = 8.0;
    mS(4,4) = 7.0;
    std::cout << mS << std::endl;
    rhs[0] = 5.14; rhs[1] = 9.9; rhs[2] = 10.3; rhs[3] = 3.8; rhs[4] = 0.7;
    std::cout << "rhs = " << rhs << std::endl;

    GMRes<Real> gmresmS(mS, 1e-12, 6);
    sol = 0.0;
    gmresmS(rhs, sol);
    std::cout << "sol = " << sol
	      << " should be [1.2, 3.3, 4.8, 0.5, 0.1]" << std::endl;
    std::cout << gmresmS << std::endl;

    std::cout << "=== SparseMatrix, Cmplx ===" << std::endl;
    SparseMatrix<Cmplx> msC(spcC, spcC);
    // matlab
    // mC=[1+2i 0 0 0 0; 2+3i 4+3i 0 0 0; 0 0 4+1i 2+3i 1+3i;
    //     1+2i 0 0 0 1+2i; 0 0 5+1i 2+3i 1+3i];
    // rhsC=[1+2i; 2+3i; 3+4i; 4+5i; 5+6i];
    msC(0,0) = Cmplx(1,2);
    msC(1,0) = Cmplx(2,3); msC(1,1) = Cmplx(4,3);
    msC(2,2) = Cmplx(4,1); msC(2,3) = Cmplx(2,3); msC(2,4) = Cmplx(1,3);
    msC(3,0) = Cmplx(1,2); msC(3,4) = Cmplx(1,2);
    msC(4,2) = Cmplx(5,1); msC(4,3) = Cmplx(2,3); msC(4,4) = Cmplx(1,3);
    std::cout << msC << std::endl;

    rhsC[0] = Cmplx(1,2); rhsC[1] = Cmplx(2,3); rhsC[2] = Cmplx(3,4);
    rhsC[3] = Cmplx(4,5); rhsC[4] = Cmplx(5,6);
    std::cout << "rhsC = " << rhsC << std::endl;

    GMRes<Cmplx> gmresmsC(msC, 1e-12, 7);
    solC = 0.0;
    gmresmsC(rhsC, solC);
    std::cout << "solC = " << solC
	      << " should be [(1,0), (0,0), (2,2), (-3.5077,-0.1385), "
	      << "(1.8,-0.6)]" << std::endl;
    std::cout << gmresmsC << std::endl;

// #ifdef HAS_PETSC
//     std::cout << "=== PETScMat ===" << std::endl;
//     PETScMat mP(spc3, spc3);
//     std::cout << mP << std::endl;

//     const PETScMat& nP = mP;
//     std::cout << "n(1,1) = " << nP(1,1) << std::endl;
//     std::cout << "n(1,2) = " << nP(1,2) << std::endl;
//     std::cout << mP << std::endl;
//     std::cout << "Triggering 'missing feature: reference access "
// 	      << "to matrix not allowed':" << std::endl;
//     mP(1,1) = 1.0;
// #endif
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
