/* testsuite for HashedMatrix
 */

#include "basics/exceptions.hh"
#include "operator/hashedSMatrix.hh"
#include <limits>

using namespace concepts;

int main(int argc, char** argv) {

  try {
    HashedSparseMatrix<Real> m(4, 4, 1);
    m(0,0) = 1; m(0,1) = 1; m(0,2) = 1; m(0,3) = 1;
    m(1,1) = 1; m(2,2) = 1; 
    m(2,3) = 1; m(3,1) = 1;
    m(3,3) = 1e-20;
    m(0,3) = 1e-15;
    m(1,3) = 1e-16;
    m(2,3) = 1e-17;
    std::cout << m << std::endl;
    std::cout << "memory = " << m.memory() << std::endl;
    std::cout << "Walking through entries:" << std::endl;

    uint hashBits = m.HashBits();
    uint p = 1 << hashBits;
    uint n = 4;
    HashedSparseMatrix<Real>::Value** matrix = m.Matrix();

    for (uint r = 0; r < n; ++r) {
      for (uint c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	  std::cout << "r = " << r << ", c_mod = " << c_mod << ", idx = "
		    << v->idx << ", val = " << v->val << std::endl;
	}
      }
    }

    m.compress(std::numeric_limits<Real>::epsilon());

    std::cout << m << std::endl;
    std::cout << "memory = " << m.memory() << std::endl;
    std::cout << "Walking through entries:" << std::endl;

    matrix = m.Matrix();
    for (uint r = 0; r < n; ++r) {
      for (uint c_mod = 0; c_mod < p; ++c_mod) {
	for (HashedSparseMatrix<Real>::Value* v =
	       matrix[(r<<hashBits) + c_mod]; v != NULL; v = v->lnk) {
	  std::cout << "r = " << r << ", c_mod = " << c_mod << ", idx = "
		    << v->idx << ", val = " << v->val << std::endl;
	}
      }
    }
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
