/* testsuite for permutation and matrices
 */

#include "basics/exceptions.hh"
#include "operator/compositions.hh"
#include "operator/matrix.hh"
#include "operator/denseMatrix.hh"
#include "operator/sparseMatrix.hh"
#include "operator/PETSc.hh"
#include "operator/permutation.hh"
#include "toolbox/array.hh"

int main(int, char**) {

  using namespace concepts;

  try {
    DummySpace<Real> spc(4);
    std::cout << spc << std::endl;

    Array<int> perm(spc.dim());
    perm[0] = 1; perm[1] = 2; perm[2] = 0; perm[3] = 3;
    Permutation<Real> P(spc, perm);
    std::cout << "P = " << P << std::endl;
//     SparseMatrix<Real>(P).storeMatlab("permutation.matlab.out");

    Permutation<Real> Q(spc, perm, true);
    std::cout << "Q = " << Q << std::endl;
//     SparseMatrix<Real>(Q).storeMatlab("permutationT.matlab.out");

    std::cout << "=== DenseMatrix ===" << std::endl;
    DenseMatrix<Real> m(spc, spc);
    m(0,1) = 1.0; m(0,3) = 7.0;
    m(1,0) = 2.0;
    m(2,0) = 5.0; m(2,1) = 6.0;
    m(3,2) = 4.0;
    std::cout << "m = " << m << std::endl;

    Compose<Real> c(m, P);
    std::cout << "c = " << c << std::endl;
//     Operator<Real>* tmp = dynamic_cast<Operator<Real>*>(&c);
//     SparseMatrix<Real>(*tmp).storeMatlab("permutation.c.matlab.out");

//     DenseMatrix<Real> mP(c);
//     std::cout << "mP = " << mP << std::endl;

    std::cout << "=== SparseMatrix ===" << std::endl;
    SparseMatrix<Real> mS(spc, spc);
    mS(0,1) = 1.0; mS(0,3) = 7.0;
    mS(1,0) = 2.0;
    mS(2,0) = 5.0; mS(2,1) = 6.0;
    mS(3,2) = 4.0;
    std::cout << "mS = " << mS << std::endl;

    Compose<Real> cS(mS, P);
    std::cout << "cS = " << cS << " = " << std::endl;
    std::cout << "  = " << SparseMatrix<Real>(cS) << std::endl;
    std::cout << "  = " << DenseMatrix<Real>(cS) << std::endl;
    SparseMatrix<Real> mP(spc, spc); P.convertToMatrix(mP);
    Compose<Real> cS2(mS, mP);
    std::cout << "  = " << DenseMatrix<Real>(cS2) << std::endl;

    Compose<Real> CS(P, mS);
    std::cout << "CS = " << CS << " = " << std::endl;
    std::cout << "  = " << SparseMatrix<Real>(CS) << std::endl;
    std::cout << "  = " << DenseMatrix<Real>(CS) << std::endl;

    Compose<Real> csp(Q, CS);
    std::cout << "csp = " << csp << " = " << std::endl;
    std::cout << "  = " << SparseMatrix<Real>(csp) << std::endl;
    std::cout << "  = " << DenseMatrix<Real>(csp) << std::endl;
    SparseMatrix<Real> mQ(spc, spc); Q.convertToMatrix(mQ);
    Compose<Real> csp2(mQ, CS);
    std::cout << "  = " << DenseMatrix<Real>(csp2) << std::endl;

    std::cout << "=== TrivExtendRestrict ===" << std::endl;
    DummySpace<Real> spcS(2);
    TrivExtendRestrict<Real> restr(spcS, spc, false);
    Compose<Real> restrP(restr, P);
    std::cout << "restrP = " << restrP << std::endl;
    Operator<Real>* tmp2 = dynamic_cast<Operator<Real>*>(&restrP);
    SparseMatrix<Real>(*tmp2).storeMatlab("restrP.m");
    std::cout << "  = " << DenseMatrix<Real>(restrP) << std::endl;

    TrivExtendRestrict<Real> ext(spc, spcS, true);
    Compose<Real> Pext(P, ext);
    std::cout << "Pext = " << Pext << std::endl;
    Operator<Real>* tmp3 = dynamic_cast<Operator<Real>*>(&Pext);
    SparseMatrix<Real>(*tmp3).storeMatlab("Pext.m");
    std::cout << "  = " << DenseMatrix<Real>(Pext) << std::endl;
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
