/* testsuite for petsc matrices
 */

#include "operator/sparseMatrix.hh"
#include "operator/denseMatrix.hh"
#ifdef HAS_PETSC
#include "operator/PETSc.hh"
#endif

int main(int argc, char** argv) {

  using namespace concepts;

  try {
    DummySpace<Real> spc2(2), spc3(3);
    std::cout << spc2 << ", " << spc3 << std::endl;

    std::cout << "=== SparseMatrix ===" << std::endl;
    SparseMatrix<Real> mS(spc2, spc3);
    std::cout << mS << std::endl;
    mS(0,0) = 1.0; mS(0,1) = 2.0; mS(0,2) = 3.0;
    mS(1,0) = 4.0; mS(1,1) = 5.0; mS(1,2) = 6.0;
    std::cout << mS << std::endl;

#ifdef HAS_PETSC
    std::cout << "=== PETScMat ===" << std::endl;
    PETScMat mP(spc3, spc3);
    std::cout << mP << std::endl;

    PETScMat mPS(mS);
    std::cout << mS << " = " << DenseMatrix<Real>(mS) << std::endl;
#endif

    DummySpace<Real> spc(5);
    SparseMatrix<Real> mU(spc, spc);
    mU(0,0) = 2.0; mU(0,1) = mU(1,0) = 0.3; mU(0,3) = mU(3,0) = 0.5;
    mU(1,1) = 3.0;
    mU(2,2) = 7.0; mU(2,4) = mU(4,2) = 2.0;
    mU(3,3) = 8.0; mU(3,4) = mU(4,3) = 6.0;
    mU(4,4) = 7.0;
    std::cout << mU << " = " << DenseMatrix<Real>(mU) << std::endl;
    PETScMat m(mU);
    PETSc solver(m, 1e-10, "cg");

    Vector<Real> b(spc), sol(spc);
    b[0] = 5.14; b[1] = 9.9; b[2] = 10.3; b[3] = 3.8; b[4] = 0.7;
    std::cout << "b = " << b << std::endl;
    solver(b, sol);
    std::cout << "sol = " << sol << " should be "
	      << "[1.56618, 3.14338, 2.13892, 2.12927, -2.33621]" << std::endl
	      << solver << std::endl;
//     std::cout << "Solve again with starting value " << sol << std::endl;
//     solver(b, sol);
//     std::cout << "sol = " << sol << " should be "
// 	      << "[1.56618, 3.14338, 2.13892, 2.12927, -2.33621]" << std::endl
// 	      << solver << std::endl;
  }
  catch (ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
