/* post processing, ie. looping over all cells / elements
 * and doing something
 */

#include <memory>
#include "postProcess.hh"
#include "space.hh"
#include "element.hh"
#include "geometry/cell.hh"
#include "geometry/mesh.hh"
#include "basics/debug.hh"

// debugging GlobalPostprocess
#define GlobalPostprocessAppl_D 0

namespace concepts {

  // ******************************************************* CellPostprocess **

  template<class F>
  std::ostream& CellPostprocess<F>::info(std::ostream& os) const {
    return os << "CellPostprocess()";
  }

  // ***************************************************** GlobalPostprocess **

  template<class F>
  GlobalPostprocess<F>::GlobalPostprocess(Space<F>& spc) :
    spc_(&spc), msh_(0) {
    conceptsAssert(spc_ != 0, Assertion());
  }

  template<class F>
  GlobalPostprocess<F>::GlobalPostprocess(Mesh& msh) :
    spc_(0), msh_(&msh) {
    conceptsAssert(msh_ != 0, Assertion());
  }

  template<class F>
  void GlobalPostprocess<F>::operator()(CellPostprocess<F>& perCell) const {
    if (spc_) {
      DEBUGL(GlobalPostprocessAppl_D, "scanning space");
      std::auto_ptr<typename Space<F>::Scanner> sc(spc_->scan());
      while (!sc->eos()) {
	Element<F>& elm = (*sc)++;
	DEBUGL(GlobalPostprocessAppl_D, "elm = " << elm);
	perCell(elm);
      } // while
      return;
    } // if

    if (msh_) {
      DEBUGL(GlobalPostprocessAppl_D, "scanning mesh");
      std::auto_ptr<Scan<Cell> > sc(msh_->scan());
      while (!sc->eos()) {
	Cell& cell = (*sc)++;
	DEBUGL(GlobalPostprocessAppl_D, "cell = " << cell);
	perCell(cell);
      } // while
      return;
    } // if

    throw conceptsException(ExceptionBase());
  }

  template<class F>
  std::ostream& GlobalPostprocess<F>::info(std::ostream& os) const {
    os << "GlobalPostprocess(";
    if (msh_) return os << *msh_ << ')';
    else return os << *spc_ << ')';
  }

} // namespace concepts
