/* testsuite for S matrix from smatrix.hh
 */

#include <iostream>
#ifdef __SUNPRO_CC
#  define __EXTENSIONS__
#endif
#include <unistd.h>
#include "smatrix.hh"
#include "tmatrix.hh"
#include "basics/exceptions.hh"
#include "integration/karniadakis.hh"
#include "integration/quadrature.hh"
#include "toolbox/array.hh"

using namespace concepts;

int main(int argc, char** argv) {

  int p = 2;
  int opt;

  // scan options
  while ((opt = getopt(argc, argv, "-p:")) != EOF)
    switch(opt) {
    case 'p': p = std::atoi(optarg); break;
    default:
      std::cout << "Usage: " << argv[0] << " -p P" << std::endl
		<< "where P denotes the polynomial degree"
		<< std::endl;
      break;
    }

  try {
    std::cout << "p = " << p << std::endl;

    Quadrature<0> Int(p+5);
    std::cout << "Int = " << Int << std::endl;
    Karniadakis<1,0> small(p+4, Int.abscissas(), Int.n());

    Array<Real> leftP(Int.n());
    for (uint i = 0; i<Int.n(); ++i)
      leftP[i] = (Int.abscissas()[i]-1.0)/2.0;
    std::cout << "leftP = " << leftP << std::endl;
    Karniadakis<1,0> left(p+4, (Real*)leftP, Int.n(), 0, 0, false);

    Array<Real> rightP(Int.n());
    for (uint i = 0; i<Int.n(); ++i)
      rightP[i] = (Int.abscissas()[i]+1.0)/2.0;
    std::cout << "rightP = " << rightP << std::endl;
    Karniadakis<1,0> right(p+4, (Real*)rightP, Int.n(), 0, 0, false);

    SMatrix1D s(small, left, right);
    std::cout << s << std::endl;

    SMatrix1D t(s, true);
    std::cout << t << std::endl;

    TColumn<Real> tc0(p+1, 1, 0);
    TColumn<Real> tc1(p+3, 1, 0);
    for (uint i = 0; i < tc0.n(); ++i)
      tc0[i] = 1.0+i;
    std::cout << "tc0 = " << tc0 << std::endl;
    s(tc0, tc1);
    std::cout << "tc1 = " << tc1 << std::endl;
    t(tc0, tc1);
    std::cout << "tc1 = " << tc1 << std::endl;
    t(tc1, tc0);
    std::cout << "tc0 = " << tc0 << std::endl;

    std::cout << "Test in 2D:" << std::endl;
    uint dim2[2] = { 2, 2 };
    TColumnTensor<Real, 2> tc21(dim2, 1);
    uint idx2[2];
    uint k = 1;
    for (idx2[1] = 0; idx2[1] < dim2[1]; idx2[1]++)
      for (idx2[0] = 0; idx2[0] < dim2[0]; idx2[0]++) {
// 	std::cout << "tc21[" << idx2[0] << ", " << idx2[1] << "] = "
// 		  << k << std::endl;
	tc21[idx2] = k++;
      }
    std::cout << "tc21 = " << tc21 << " = "
	      << (TColumn<Real>)tc21 << std::endl;
    SMatrixTensor<Real, 2> smt2Dleft1(s, 1);
    SMatrixTensor<Real, 2> smt2Dleft2(s, 2);
    SMatrixTensor<Real, 2> smt2Dright1(t, 1);
    SMatrixTensor<Real, 2> smt2Dright2(t, 2);
    SMatrixCompose<Real> smc2D1(smt2Dleft1, smt2Dleft2);
    std::cout << "smc2D1 = " << smc2D1 << std::endl;
    std::cout << "smc2D1 should represent S matrix of first child in quad"
	      << std::endl;
    uint K2 = dim2[0]*dim2[1];
    TColumnTensor<Real, 2> tc22(dim2, 1);
    for (uint k = 0; k < K2; ++k) {
      for (idx2[0] = 0; idx2[0] < tc21.n()[0]; ++idx2[0])
	for (idx2[1] = 0; idx2[1] < tc21.n()[1]; ++idx2[1])
	  tc21[idx2] = 0.0;

      for (uint& j = idx2[1] = 0; j < tc21.n()[1]; ++j)
	for (uint& i = idx2[0] = 0; i < tc21.n()[0]; ++i)
	  if (k == i+j*tc21.n()[0])
	    tc21[idx2] = 1.0;
      std::cout << tc21;
      smc2D1(tc21, tc22);
      std::cout << " gives " << tc22 << std::endl;
    }

    SMatrixCompose<Real> smc2D2(smt2Dright1, smt2Dleft2);
    std::cout << "smc2D2 = " << smc2D2 << std::endl;
    std::cout << "smc2D2 should represent S matrix of second child in quad"
	      << std::endl;
    for (uint k = 0; k < K2; ++k) {
      for (idx2[0] = 0; idx2[0] < tc21.n()[0]; ++idx2[0])
	for (idx2[1] = 0; idx2[1] < tc21.n()[1]; ++idx2[1])
	  tc21[idx2] = 0.0;

      for (uint& i = idx2[0] = 0; i < tc21.n()[0]; ++i)
	for (uint& j = idx2[1] = 0; j < tc21.n()[1]; ++j)
	  if (k == i+j*tc21.n()[0])
	    tc21[idx2] = 1.0;
      std::cout << tc21;
      smc2D2(tc21, tc22);
      std::cout << " gives " << tc22 << std::endl;
    }

    SMatrixCompose<Real> smc2D3(smt2Dright1, smt2Dright2);
    std::cout << "smc2D3 = " << smc2D3 << std::endl;
    std::cout << "smc2D3 should represent S matrix of third child in quad"
	      << std::endl;
    for (uint k = 0; k < K2; ++k) {
      for (idx2[0] = 0; idx2[0] < tc21.n()[0]; ++idx2[0])
	for (idx2[1] = 0; idx2[1] < tc21.n()[1]; ++idx2[1])
	  tc21[idx2] = 0.0;

      for (uint& i = idx2[0] = 0; i < tc21.n()[0]; ++i)
	for (uint& j = idx2[1] = 0; j < tc21.n()[1]; ++j)
	  if (k == i+j*tc21.n()[0])
	    tc21[idx2] = 1.0;
      std::cout << tc21;
      smc2D3(tc21, tc22);
      std::cout << " gives " << tc22 << std::endl;
    }

    SMatrixCompose<Real> smc2D4(smt2Dleft1, smt2Dright2);
    std::cout << "smc2D4 = " << smc2D4 << std::endl;
    std::cout << "smc2D4 should represent S matrix of fourth child in quad"
	      << std::endl;
    for (uint k = 0; k < K2; ++k) {
      for (idx2[0] = 0; idx2[0] < tc21.n()[0]; ++idx2[0])
	for (idx2[1] = 0; idx2[1] < tc21.n()[1]; ++idx2[1])
	  tc21[idx2] = 0.0;

      for (uint& i = idx2[0] = 0; i < tc21.n()[0]; ++i)
	for (uint& j = idx2[1] = 0; j < tc21.n()[1]; ++j)
	  if (k == i+j*tc21.n()[0])
	    tc21[idx2] = 1.0;
      std::cout << tc21;
      smc2D4(tc21, tc22);
      std::cout << " gives " << tc22 << std::endl;
    }

    std::cout << "Test in 3D:" << std::endl;
    uint dim3[3] = { 2, 2, 2 };
    TColumnTensor<Real, 3> tc31(dim3, 1);
    uint idx3[3];
    k = 1;
    for (idx3[2] = 0; idx3[2] < dim3[2]; idx3[2]++)
      for (idx3[1] = 0; idx3[1] < dim3[1]; idx3[1]++)
	for (idx3[0] = 0; idx3[0] < dim3[0]; idx3[0]++) {
// 	  std::cout << "tc31[" << idx3[0] << ", " << idx3[1] << ", "
// 		    << idx3[2] << "] = " << k << std::endl;
	  tc31[idx3] = k++;
	}
    std::cout << "tc31 = " << tc31 << " = "
	      << (TColumn<Real>)tc31 << std::endl;

    // testing the S matrix for the first child
    TColumnTensor<Real, 3> tc32(dim3, 1);
    SMatrixTensor<Real, 3> smt3Dleft1(s, 1);
    SMatrixTensor<Real, 3> smt3Dleft2(s, 2);
    SMatrixTensor<Real, 3> smt3Dleft3(s, 3);
    SMatrixCompose<Real> smc3D1(smt3Dleft1, smt3Dleft2, &smt3Dleft3);
    std::cout << "smc3D1 = " << smc3D1 << std::endl;
    std::cout << "smc3D1 should represent S matrix of first child"
	      << std::endl;
    const uint K3 = dim3[0]*dim3[1]*dim3[2];
    for (uint k = 0; k < K3; ++k) {
      for (idx3[0] = 0; idx3[0] < tc31.n()[0]; ++idx3[0])
	for (idx3[1] = 0; idx3[1] < tc31.n()[1]; ++idx3[1])
	  for (idx3[2] = 0; idx3[2] < tc31.n()[2]; ++idx3[2])
	    tc31[idx3] = 0.0;
      for (uint& i = idx3[0] = 0; i < tc31.n()[0]; ++i)
	for (uint& j = idx3[1] = 0; j < tc31.n()[1]; ++j)
	  for (uint& l = idx3[2] = 0; l < tc31.n()[2]; ++l)
	    if (k == i+j*tc31.n()[0]+l*tc31.n()[1]*tc31.n()[0])
	      tc31[idx3] = 1.0;
      std::cout << tc31;
      smc3D1(tc31, tc32);
      std::cout << " gives " << tc32 << std::endl;
    }
    // testing the S matrix for the 2nd child
    SMatrixTensor<Real, 3> smt3Dright1(t, 1);
    SMatrixCompose<Real> smc3D2(smt3Dright1, smt3Dleft2, &smt3Dleft3);
    std::cout << "smc3D1 = " << smc3D2 << std::endl;
    std::cout << "smc3D1 should represent S matrix of second child"
	      << std::endl;
    for (uint k = 0; k < K3; ++k) {
      for (idx3[0] = 0; idx3[0] < tc31.n()[0]; ++idx3[0])
	for (idx3[1] = 0; idx3[1] < tc31.n()[1]; ++idx3[1])
	  for (idx3[2] = 0; idx3[2] < tc31.n()[2]; ++idx3[2])
	    tc31[idx3] = 0.0;
      for (uint& i = idx3[0] = 0; i < tc31.n()[0]; ++i)
	for (uint& j = idx3[1] = 0; j < tc31.n()[1]; ++j)
	  for (uint& l = idx3[2] = 0; l < tc31.n()[2]; ++l)
	    if (k == i+j*tc31.n()[0]+l*tc31.n()[1]*tc31.n()[0])
	      tc31[idx3] = 1.0;
      std::cout << tc31;
      smc3D2(tc31, tc32);
      std::cout << " gives " << tc32 << std::endl;
    }
  }
  catch(ExceptionBase& e) {
    std::cout << e << std::endl;
  }

  return 0;
}
