/*----------------------------------------------------------------------------
Sparse QR solver for linear systems of equations or least squares problems
Copyright (C) 1996, Thomas H. Robey

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

Inquiries can be directed to trobey@arc.unm.edu or

Thomas H. Robey
925 Madison NE
Albuquerque, NM  87110
USA
----------------------------------------------------------------------------*/

#include <cassert>
#include "sparseqr.hh"

namespace sparseqr {

  int Smatrix::match(Smatrix *rptr2) {
    Smatrix *rptr1=this;
    while (rptr1!=NULL || rptr2!=NULL) {
      if (rptr1==NULL || rptr2==NULL || rptr1->col!=rptr2->col)
	return(0);
      rptr1=rptr1->right;
      rptr2=rptr2->right;
    }
    return(1);
  };

  Smatrix *Smatrix::MoveDown(Smatrix **sptr) {
    *sptr=this;
    return(down);
  };

  Smatrix *Smatrix::MoveRight(Smatrix **sptr) {
    *sptr=this;
    return(right);
  };

  Smatrix *Smatrix::MarkEntry(Smatrix **sptr) {
    *sptr=right;
    row=-1;
    return(right);
  };
 
  void *Smatrix::operator new(size_t,int r,int c,double v,Smatrix *rptr,
			      Smatrix *dptr) {
    Smatrix *sptr=::new Smatrix;
    assert(sptr!=NULL);
    sptr->row=r;
    sptr->col=c;
    sptr->value=v;
    sptr->right=rptr;
    sptr->down=dptr;
    return sptr;
  };

}
