// testing qr factorizations

#include "driverTest.hh"
#include "operator/denseMatrix.hh"

namespace test {

  // ************************************************************ DriverTest **

  DriverTest::DriverTest() :
    spc1(5), spc2(8), a(spc2, spc2), b(spc2, spc1), decomp(b) {
    a(0,0) = 2.0; a(0,1) = a(1,0) = 0.3; a(0,3) = a(3,0) = 0.5;
    a(1,1) = 3.0;
    a(2,2) = 7.0; a(2,4) = a(4,2) = 2.0;
    a(3,3) = 8.0; a(3,4) = a(4,3) = 6.0;
    a(4,4) = 7.0;
    a(7,4) = 1;

    b(0,0) = 1; b(0,3) = 7; b(0,4) = -3;
    b(1,1) = 8; b(1,2) = 5;
    b(2,0) = 3; b(2,3) = 10;
    b(3,2) = 2;
    b(4,1) = 1; b(4,3) = 9; b(4,4) = 1;
    b(5,2) = 1;
    b(6,2) = 7;
    b(7,4) = 1;
  }

  DriverTest::~DriverTest() {}

  void DriverTest::run() {
    testQ(); testQt(); testSolver();
  }

  void DriverTest::testQ() {
    sparseqr::GivensRotations<Real>* q = decomp.Q();

    concepts::Compose<Real> aq(a, *q);
    concepts::SparseMatrix<Real> aqs(aq);
    compare_(aqs, aq);

    concepts::Compose<Real> qa(*q, a);
    concepts::SparseMatrix<Real> qas(qa);
    compare_(qas, qa);
  }

  void DriverTest::testQt() {
    sparseqr::GivensRotations<Real>* qt = decomp.Qt();
    std::cout << *qt << std::endl;

    concepts::Compose<Real> aqt(a, *qt);
    concepts::SparseMatrix<Real> aqts(aqt);
    compare_(aqts, aqt);

    concepts::Compose<Real> qta(*qt, a);
    concepts::SparseMatrix<Real> qtas(qta);
    compare_(qtas, qta);
  }

  void DriverTest::testSolver() {
    // same as example in UMFPACK Guide p. 10
    concepts::SparseMatrix<Real> mU(spc1, spc1);
    mU(0,0) = 2; mU(0,1) = 3;
    mU(1,0) = 3; mU(1,2) = 4; mU(1,4) = 6;
    mU(2,1) = -1; mU(2,2) = -3; mU(2,3) = 2;
    mU(3,2) = 1;
    mU(4,1) = 4; mU(4,2) = 2; mU(4,4) = 1;

    concepts::Vector<Real> b(spc1), sol(spc1);
    b[0] = 8; b[1] = 45; b[2] = -3; b[3] = 3; b[4] = 19;
    sparseqr::Driver solver(mU);
    solver(b, sol);
//     for (uint i = 0; i < spc1.dim(); ++i)
//       _numtest(sol[i], i+1);
  }

  void DriverTest::compare_(const concepts::SparseMatrix<Real>& s,
			    concepts::Operator<Real>& o) {
    DEBUGL(0, concepts::DenseMatrix<Real>(s));
    _test(s.spaceX().dim() == o.spaceX().dim());
    _test(s.spaceY().dim() == o.spaceY().dim());
    concepts::Vector<Real> tmp(s.spaceY()), res(s.spaceX());
    for (uint i = 0; i < s.spaceY().dim(); ++i) {
      tmp[i] = 1.;
      o(tmp, res);
      DEBUGL(0, "i = " << i << ", res = " << res);
      for (uint j = 0; j < s.spaceX().dim(); ++j) {
	_numtest(s(j, i), res[j]);
      }
      tmp[i] = 0.;
    }
  }

} // namespace test

int main(int, char**) {
  concepts::Stacktrace::doit() = false;
  try {
    std::cout << std::setprecision(20);

    test::DriverTest t1;

    t1.run();
    return t1.report();
  }
  catch(concepts::ExceptionBase& e) {
    std::cout << e << std::endl;
  }
}
