// nystroem strategy

#include "nystroem.hh"
#include "timestepping/vectors.hh"
#include "operator/compositions.hh"

// debugging
#include "basics/debug.hh"

#define Nystroem_D 0

namespace timestepping {

  // ************************************************************** Nystroem **

  Nystroem::Nystroem(concepts::SolverFabric<Real>& fabric,
                     concepts::Operator<Real>& D2,
                     concepts::Operator<Real>& D0,
                     timestepping::TimeVector& trhs, 
                     const concepts::Vector<Real>& Y0,
                     const concepts::Vector<Real>& Z0,
                     Real dt) :
    TimeStepStrategy(fabric, &D2, Y0, Z0, dt),
    D0_(D0), trhs_(trhs), Z_(Z0) { }

  void Nystroem::next() {
    if(t_ == 0) {
      Z_ *= 0.5*dt_;
    }

    t_ += dt_;

    sol_ += Z_;
    D0_(sol_, rhs_);
    trhs_.time(t_ - 0.5*dt_);
    rhs_ -= trhs_;

    concepts::Vector<Real> temp(rhs_);
    (*solver_)(rhs_, temp);
    temp *= 0.5*dt_*dt_;
    Z_ -= temp;
    sol_ += Z_;
  }

  std::ostream& Nystroem::info(std::ostream& os) const {
    return os << "Nystroem(dt = " << dt_ << ')';
  }

}
