// runge kutta 4 strategy

#include "rungekutta4.hh"
#include "vectors.hh"
#include "operator/compositions.hh"

// debugging
#include "basics/debug.hh"

#define RungeKutta4_D 0

namespace timestepping {

  // ************************************************* Runge Kutta 4th order **

  RungeKutta4::RungeKutta4(concepts::SolverFabric<Real>& fabric,
			   concepts::Operator<Real>& D1,
			   concepts::Operator<Real>& D0,
			   timestepping::TimeVector& trhs, 
			   const concepts::Vector<Real>& Y0,
			   Real dt) :
    TimeStepStrategy(fabric, &D1, Y0, Y0, dt),
    D1_(D1), D0_(D0), trhs_(trhs), Yn1_(Y0) {}

  RungeKutta4::~RungeKutta4() {}

  void RungeKutta4::next() {
    Yn1_ = sol_;
    TimeVector temp1(sol_);
    TimeVector temp2(sol_);

    temp1.time(t_);
    D0_(temp1, temp2);
    trhs_.time(t_);
    temp2 -= trhs_;
    (*solver_)(temp2, temp1);
    temp1 *= -dt_ / 6.0;
    DEBUGL(RungeKutta4_D, "1. Summand = " << temp1 << std::endl);
    sol_ += temp1;

    temp1 *= 3.0;
    temp1 += Yn1_;
    temp1.time(t_ + 0.5 * dt_);
    D0_(temp1, temp2);
    trhs_.time(t_ + 0.5 * dt_);
    temp2 -= trhs_;
    (*solver_)(temp2, temp1);
    temp1 *= -dt_ / 3.0;
    DEBUGL(RungeKutta4_D, "2. Summand = " << temp1 << std::endl);
    sol_ += temp1;

    temp1 *= 3.0 / 2.0;
    temp1 += Yn1_;
    temp1.time(t_ + 0.5 * dt_);
    D0_(temp1, temp2);
    temp2 -= trhs_;
    (*solver_)(temp2, temp1);
    temp1 *= -dt_ / 3.0;
    DEBUGL(RungeKutta4_D, "3. Summand = " << temp1 << std::endl);
    sol_ += temp1;

    temp1 *= 3.0;
    temp1 += Yn1_;
    temp1.time(t_ + dt_);
    D0_(temp1, temp2);
    trhs_.time(t_ + dt_);
    temp2 -= trhs_;
    (*solver_)(temp2, temp1);
    temp1 *= -dt_ / 6.0;
    DEBUGL(RungeKutta4_D, "4. Summand = " << temp1 << std::endl);
    sol_ += temp1;

    t_ += dt_;
  }

  std::ostream& RungeKutta4::info(std::ostream& os) const {
    return os << "RungeKutta4(dt = " << dt_ << ')';
  }

}
