/* The boundary class is used the describe an element on the boundary
 */

#include "basics/exceptions.hh"
#include "basics/vectorsMatrices.hh"
#include "formula.hh"
#include "boundary.hh"

namespace concepts {

  // ************************************************************** Boundary **

  Boundary::Boundary() : type_(FREE), frm_(0) {}

  Boundary::Boundary(const enum boundaryTypes type) : type_(type), frm_(0) {
    conceptsAssert(type_ < MAX_TYPE, Assertion());
  }

  Boundary::Boundary(const enum boundaryTypes type, const Formula& frm) :
    type_(type), frm_(frm.clone()) {
    conceptsAssert(type_ < MAX_TYPE, Assertion());
  }

  Boundary::Boundary(const enum boundaryTypes type, const char* frm) :
    type_(type), frm_(new ParsedFormula(frm)) {
    conceptsAssert(type_ < MAX_TYPE, Assertion());
  }

  Boundary::Boundary(const Boundary& bnd) : type_(bnd.type_),
  frm_(bnd.frm_.get() ? bnd.frm_->clone() : 0) {
    conceptsAssert(type_ < MAX_TYPE, Assertion());
  }

  Boundary::~Boundary() {}

  Boundary& Boundary::operator=(const Boundary& bnd) {

    if (&bnd != this) {
      type_ = bnd.type_;
      if (bnd.frm_.get()) frm_.reset(bnd.frm_->clone());
      else frm_.reset(0);
    }

    return *this;
  }

  Real Boundary::operator()(const Real x, const Real t) const {

    return (frm_.get()) ? (*frm_)(x, t) : 0.0;
  }

  Real Boundary::operator()(const Real2d& x, const Real t) const {

    return (frm_.get()) ? (*frm_)(x, t) : 0.0;
  }

  Real Boundary::operator()(const Real3d& x, const Real t) const {

    return (frm_.get()) ? (*frm_)(x, t) : 0.0;
  }

  std::string Boundary::typeStr() const {
    switch(type_) {
    case 0: return "FREE";
    case 1: return "DIRICHLET";
    case 2: return "NEUMANN";
    default: return "UNDEFINED";
    }
  }

  std::ostream& Boundary::info(std::ostream& os) const {

    os << "Boundary(" << typeStr() << ", ";
    if (frm_.get()) os << *frm_;
    else os << "(0)";
    os << ")";
    return os;
  }

} // namespace concepts
