%option stack

%{
#include <stdlib.h>
#include <string.h>
// #include "input.tab.cc.h"
#include "input.tab.hh"
#define YY_DECL int yylex (YYSTYPE *lvalp)
%}

%x COMMENT
%x LINECOMMENT
%s INARRAY
%x INFORMULA

STRING  [_a-zA-Z][_a-zA-Z0-9]*
INT     [+-]?[0-9]+
DIGIT   [0-9]
NUM1    [-+]?{DIGIT}+\.?([eE][-+]?{DIGIT}+)?
NUM2    [-+]?{DIGIT}*\.{DIGIT}+([eE][-+]?{DIGIT}+)?
REAL  {NUM1}|{NUM2}
FORM    [^\"]

 char* YYstring = 0;
 char* YYform = 0;
 unsigned int YYstringleng = 0; /* length of array YYstring */
 unsigned int YYformleng = 0; /* length of array YYform */
 char* YYstringlenghten(char* old, unsigned int* oldlength, 
			unsigned int newlength);

%%

"//"                { yy_push_state(LINECOMMENT); }
<LINECOMMENT>.
<LINECOMMENT>\n     { yy_pop_state(); }

"/*"                { yy_push_state(COMMENT); }
<COMMENT>.
<COMMENT>\n
<COMMENT>"*/"       { yy_pop_state(); }

\"                  { yy_push_state(INFORMULA); }
<INFORMULA>{FORM}*  { if (yyleng >= YYformleng)
			YYform = YYstringlenghten(YYform, &YYformleng, 
						  yyleng+1);
		      else
			memset(YYform, 0, YYformleng);
		      strncpy(YYform, yytext, yyleng);
		      lvalp->name = YYform;
		      return ANYTHING; }
<INFORMULA>\"       { yy_pop_state(); }

array               { yy_push_state(INARRAY); return ARRAY; }
<INARRAY>"{"        { return '{'; }
<INARRAY>"}"        { yy_pop_state(); return '}'; }

 /* keywords */
end      { return 0; }
double   { return REAL; }
string   { return STRING; }
int      { return INT; }
bool     { return BOOL; }

 /* parsing numbers */
{INT}    { lvalp->ganz = atoi(yytext); return GANZ; }

{REAL}   { lvalp->value = atof(yytext); return NUM; }

true     { lvalp->ganz = 1; return TRUEFALSE; }
false    { lvalp->ganz = 0; return TRUEFALSE; }

 /* parsing names and strings */
{STRING} { if (yyleng >= YYstringleng)
	     YYstring = YYstringlenghten(YYstring, &YYstringleng, yyleng+1);
	   else
	     memset(YYstring, 0, YYstringleng);
	   strncpy(YYstring, yytext, yyleng);
	   lvalp->name = YYstring;
	   return NAME; }

\n       { return '\n'; }

. /* eat up any unmatched character */

%%

char* YYstringlenghten(char* old, unsigned int* oldlength, 
		       unsigned int newlength) {

  free(old);
  *oldlength = newlength;
  return (char*)calloc(newlength, sizeof(char));
}

