/* write table of results
 */

#ifndef resultsTable_hh
#define resultsTable_hh

#include <map>
#include <string>
#include "basics/outputOperator.hh"
#include "inputOutput.hh"

namespace concepts {

  // ********************************************************** ResultsTable **

  /** Organizes the results in the hashes from InOutParameters in
      a nice table.

      @author Philipp Frauenfelder, 2002
  */
  class ResultsTable : public OutputOperator {
  public:
    enum tableTypes { SCREEN, GNUPLOT, HTML };
    /** Table delimiters.
	@author Philipp Frauenfelder, 2002
    */
    template<enum tableTypes t>
    struct tableDelimTraits {};

    /// Types of maps in InOutParameters
    enum mapTypes { DOUBLE, STRING, INT, BOOL };

    /// Constructor
    ResultsTable();
    ~ResultsTable();

    /// 
    void addMap(enum mapTypes type, const char* name,
		const InOutParameters& holder);

    template<enum tableTypes t>
    void print(std::ostream& os) const;

    template<enum tableTypes t>
    void print(std::ostream& os, enum mapTypes typeL, const char* nameL) const;
  protected:
    virtual std::ostream& info(std::ostream& os) const;
  private:
    std::map<const char*, const std::map<int, double>*, ltstr> doubleMaps_;
    std::map<const char*,
	     const std::map<int, std::string>*, ltstr> stringMaps_;
    std::map<const char*, const std::map<int, int>*, ltstr> intMaps_;
    std::map<const char*, const std::map<int, bool>*, ltstr> boolMaps_;
    bool leader_;
    enum mapTypes typeL_;
    char* nameL_;
    template<enum tableTypes t>
    void headline_(std::ostream& os) const;
    template<enum tableTypes t>
    void tableline_(std::ostream& os, const int idx) const;
  };

} // namespace concepts

#endif // resultsTable_hh
