#include <sstream>
#include "blockOperator.hh"
#include "basics/exceptions.hh"

// debugging

#define BlockOperatorAppl_D 0

namespace vectorial {

  // ********************************************************* BlockOperator **

  template<class F>
  void BlockOperator<F>::operator()(const concepts::Function<F>& fncY,
                                    concepts::Function<F>& fncX) {
    const concepts::Vector<F>* fncYA;
    if ((fncYA = dynamic_cast<const concepts::Vector<F>*>(&fncY))) {
      concepts::Vector<F>* fncXA;
      if ((fncXA = dynamic_cast<concepts::Vector<F>*>(&fncX))) {
        (*this)(*fncYA, *fncXA); return;
      }
    }
    throw conceptsException(concepts::MissingFeature
                            ("function not supported"));
  }

  template<class F>
  void BlockOperator<F>::operator()(const concepts::Vector<F>& fncY,
                                    concepts::Vector<F>& fncX) {
    // subspaces
    const concepts::Space<F>& spcX = *(spc_.get(0));
    const concepts::Space<F>& spcY = *(spc_.get(1));

    // subvectors
    concepts::Vector<F> x(spcX, (F*)(fncX));
    concepts::Vector<F> y(spcY, (F*)(fncX)+spcX.dim());
    const concepts::Vector<F> f(spcX, (F*)(fncY));
    const concepts::Vector<F> g(spcY, (F*)(fncY)+spcX.dim());

    if (A_) {
      (*A_)(f, x);
      if (B_) {
        (*B_)(g, *tempX_);
        x += *tempX_;
      }
    } else {
      if(B_) {
        (*B_)(g, x);
      } else {
        x *= 0;
      }
    }
    if (C_) {
      (*C_)(f, y);
      if (D_) {
        (*D_)(g, *tempY_);
        y += *tempY_;
      }
    } else {
      if (D_) {
        (*D_)(g, y);
      } else {
        y *= 0;
      }
    }
  }

  template<class F>
  std::ostream& BlockOperator<F>::info(std::ostream& os) const {
    os << "BlockOperator(A = ";
    if (A_) os << *A_; else os << 0;
    os << ", B = ";
    if (B_) os << *B_; else os << 0;
    os << ", C = ";
    if (C_) os << *C_; else os << 0;
    os << ", D = ";
    if (D_) os << *D_; else os << 0;
    return os << ")";
  }

  template class BlockOperator<concepts::Real>;
  template class BlockOperator<concepts::Cmplx>;

} // namespace vectorial
