#include <map>

#include "elementPairListIdentic.hh"

namespace vectorial {

  template<class F>
  void ElementPairListIdentic<F>::constructor_
  (const concepts::ElementPairList<F>& sElemPairs,
   const Space<F>& vSpace1, const Space<F>& vSpace2) {

    const uint vdim1 = vSpace1.vdim();
    const uint vdim2 = vSpace2.vdim();

    // build up map concepts::Element -> vectorial::Element
    std::map<const concepts::Element<F>*, const Element<F>* > elements1;
    std::auto_ptr<typename Space<F>::Scanner> elScan(vSpace1.scan());
    while (*elScan) {
      const Element<F>& vElm = (*elScan)++;
      for(uint i=0; i<vdim1; ++i) {
        const concepts::Element<F>* sElm = vElm.get(i);
        elements1[sElm] = &vElm;
      }
    }

    std::map<const concepts::Element<F>*, const Element<F>* >* elements2;

    bool different = &vSpace1 != &vSpace2;

    if (different) {
      elements2 = new std::map<const concepts::Element<F>*, 
        const Element<F>* >;
      elScan.reset(vSpace2.scan());
      while (*elScan) {
        const Element<F>& vElm = (*elScan)++;
        for(uint i=0; i<vdim2; ++i) {
          const concepts::Element<F>* sElm = vElm.get(i);
          (*elements2)[sElm] = &vElm;
        }
      }
    } else {
      elements2 = &elements1;
    }
  
    // build up element pair list
    std::auto_ptr<typename concepts::ElementPairList<F>::Scanner> 
      epScan(sElemPairs.scan());
    while (*epScan) {
      concepts::ElementPair<F>& elmPair = (*epScan)++;
      ElementPair<F>* vElmPair =
        new ElementPair<F>(*elements1[&elmPair.elm1()],
                           *(*elements2)[&elmPair.elm2()]);
      for (uint iV=0; iV<vdim1; ++iV) {
        for (uint iU=0; iU<vdim2; ++iU) {
          vElmPair->put(elmPair, iV, iU);
        }
      }
      add(vElmPair);
    }

    if (different) {
      delete elements2;
    }

  }

  template class ElementPairListIdentic<concepts::Real>;

} // namespace vectorial
