#include "timeLinearForm.hh"
#include "basics/exceptions.hh"
#include "function/vector.hh"
#include "linearFormHelper.hh"

// debugging
#include "basics/debug.hh"

namespace vectorial {

  // ******************************************************** TimeLinearForm **

  TimeLinearForm::~TimeLinearForm(){
    DEBUGL(TimeLinearFormConstr_D, "deletion done.");
  }

  void TimeLinearForm::put(timestepping::TimeLinearForm &lf,
                           const int i, const int j) {
    vdata_[i] = &lf;
    DEBUGL(TimeLinearFormConstr_D,"idx = " << idx_ << ", i = " << i);
  }

  void TimeLinearForm::operator ()(const concepts::Element<Real>& elm,
                                   concepts::ElementMatrix<Real>& em){
    const Element<Real>* ev = dynamic_cast<const Element<Real>*>(&elm);
    conceptsAssert(ev, concepts::Assertion());
    timestepping::TimeLinearForm** array = vdata_;
    LinearFormHelper::assembleComponents
      (*ev, em, reinterpret_cast<concepts::LinearForm<Real>**>(array), vdim_);
  }

  void TimeLinearForm::time(Real time) {
    timestepping::TimeLinearForm::time(time);
    for (uint i=0;i<vdim_;++i) {
      if (vdata_[i]!=0) {
        vdata_[i]->time(time);
      }
    }
  }

  std::ostream& TimeLinearForm::info(std::ostream& os) const {
    os << "TimeLinearForm(" << idx_ << ", " << vdim_ << ")";
    return os;
  }

} // namespace vectorial
